import inspect



class ASTNode (object):

    cur_pos = False

    _res = False

    priority = 1000

    def text(self):
        pass

    def par(self, s):
        return "(" + s + ")"

    def accept(self, visitor):
        # Searching for the most specific visitor
        # Important: getmro must list classes from descendants up!
        found = False
        cn = self.__class__
        for c in [x.__name__ for x in inspect.getmro(cn)]:
            if c != "object" and hasattr(visitor, c + "_visit"):
                found = True
                eval("visitor." + c + "_visit")(self)
                break
        if not found:
            # Generic method
            visitor.visit(self)

    def __eq__(self, other):
        return self.text() == other.text()

    def __hash__(self):
        return hash(self.text())
    


class Expression (ASTNode):
    pass



class ConditionalExpression (Expression):

    condition = None
    then_exp = None
    else_exp = None
    priority = 10

    def text(self):
        return self.par(self.condition.text() + " ? " + self.then_exp.text() + " : " + self.else_exp.text())



class BinaryExpression (Expression):

    left = None
    right = None
    op = None
    priority = 20
    associative = False

    def text(self):
        lt = self.left.text() if self.left.priority > self.priority or \
             (self.left.priority == self.priority and self.associative) \
             else self.par(self.left.text())
        rt = self.right.text() if self.right.priority > self.priority or \
             (self.right.priority == self.priority and self.associative) \
             else self.par(self.right.text())
        return lt + " " + self.op + " " + rt

    def aggregate(self, ls):
        if len(ls) == 2:
            self.left = ls[0]
            self.right = ls[1]
        else:
            ni = eval(self.__class__.__name__)()
            self.left = ls[0]
            ni.aggregate(ls[1:])
            self.right = ni

    def mix_aggregate(self, ls):
        if len(ls) == 2:
            res = ls[1][0]
            res.left = ls[0][1] if len(ls[0]) > 1 else ls[0][0]
            res.right = ls[1][1]
        else:
            pr = self.mix_aggregate(ls[1:])
            res = ls[1][0]
            res.left = ls[0][1] if len(ls[0]) > 1 else ls[0][0]
            res.right = pr
        return res



class EquivalenceExpression (BinaryExpression):

    IFF = "<==>"
    NIFF = "<=!=>"
    priority = 30



class IffExpression (EquivalenceExpression):

    def __init__(self):
        self.op = self.IFF



class NIffExpression (EquivalenceExpression):

    def __init__(self):
        self.op = self.NIFF



class ImplicationExpression (BinaryExpression):

    IMPL = "==>"
    RIMPL = "<=="
    priority = 50



class ImpliesExpression (ImplicationExpression):

    def __init__(self):
        self.op = self.IMPL



class ImpliedExpression (ImplicationExpression):

    def __init__(self):
        self.op = self.RIMPL



class OrExpression (BinaryExpression):

    priority = 60

    def __init__(self):
        self.op = '||'
        self.associative = True



class AndExpression (BinaryExpression):

    priority = 70

    def __init__(self):
        self.op = '&&'
        self.associative = True



class EqualityExpression (BinaryExpression):

    EQUALS = "=="
    NEQUALS = "!="
    priority = 100



class EqualsExpression (EqualityExpression):

    def __init__(self):
        self.op = self.EQUALS



class NEqualsExpression (EqualityExpression):

    def __init__(self):
        self.op = self.NEQUALS



class RelationalExpression (BinaryExpression):

    LESS = "<"
    GREATER = ">"
    LESSEQUAL = "<="
    GREATEREQUAL = ">="
    priority = 120



class LessExpression (RelationalExpression):

    def __init__(self):
        self.op = self.LESS



class LessEqualExpression (RelationalExpression):

    def __init__(self):
        self.op = self.LESSEQUAL
    


class GreaterExpression (RelationalExpression):

    def __init__(self):
        self.op = self.GREATER
    


class GreaterEqualExpression (RelationalExpression):

    def __init__(self):
        self.op = self.GREATEREQUAL



class AdditiveExpression (BinaryExpression):

    PLUS = "+"
    MINUS = '-'
    priority = 150
    associative = True



class AddExpression (AdditiveExpression):

    def __init__(self):
        self.op = self.PLUS



class SubExpression (AdditiveExpression):

    def __init__(self):
        self.op = self.MINUS



class MultiplicativeExpression (BinaryExpression):

    TIMES = "*"
    DIV = "/"
    MOD = "%"
    priority = 200



class MultExpression (MultiplicativeExpression):

    def __init__(self):
        self.op = self.TIMES



class DivExpression (MultiplicativeExpression):

    def __init__(self):
        self.op = self.DIV



class ModExpression (MultiplicativeExpression):

    def __init__(self):
        self.op = self.MOD



class UnaryExpression (Expression):

    PLUSUN = '+'
    MINUSUN = '-'
    NOT = '!'
    item = None
    op = None
    priority = 220

    def text(self):
        t = self.item.text() if self.item.priority > self.priority else self.par(self.item.text())
        return self.op + t



class PlusUnaryExpression (UnaryExpression):

    def __init__(self):
        self.op = self.PLUSUN



class MinusUnaryExpression (UnaryExpression):

    def __init__(self):
        self.op = self.MINUSUN



class NotExpression (UnaryExpression):

    def __init__(self):
        self.op = self.NOT



class PrimitiveExpression (Expression):

    val = None
    priority = 500

    def text(self):
        return str(self.val)



class Identifier (PrimitiveExpression):

    THIS = "this"
    SUPER = "super"
    VOID = "void"
    RESULT = "\\result"
    OLD = "\\old"



class FeatureCall (Expression):

    target = None  # Is None for unqualified calls
    call = None  # List of dotted expressions
    priority = 300

    def call_text(self):
        return '.'.join([c.text() for c in self.call])

    def text(self):
        c = self.call_text()
        dot = "." if len(c) > 0 else ""
        if self.target is None:
            t = ""
        else:
            t = (self.target.text() if self.target.priority >= self.priority else \
                 self.par(self.target.text())) + dot
        return t + c



class AttributeCall (FeatureCall):
    pass



class MethodCall (FeatureCall):

    args = []
    def text(self):
        return super(MethodCall, self).text() + '(' + ', '.join([a.text() for a in self.args]) + ')'



class ArrayReference (FeatureCall):

    target = None
    args = []
    priority = 300

    def text(self):
        return super(ArrayReference, self).text() + ''.join(['[' + x.text() + ']' for x in self.args])



class Type (ASTNode):

    idt = ""

    def text(self):
        return str(self.idt)



class PrimitiveType (Type):

    BOOLEAN = "boolean"
    CHAR = "char"
    BYTE = "byte"
    SHORT = "short"
    INT = "int"
    LONG = "long"
    FLOAT = "float"
    DOUBLE = "double"



class ArrayType (Type):

    basetyp = None
    dim = 1  # How many []'s

    def text(self):
        return self.basetyp.text() + ''.join(self.dim*['[]'])



class Literal (PrimitiveExpression):

    typ = None
    val = 0



class ResultExpression (Identifier):

    def __init__(self):
        self.val = self.RESULT
    


class OldExpression (MethodCall):

    target = None  # As if it were an unqualified call

    def __init__(self):
        c = Identifier()
        c.val = c.OLD
        self.call = [c]



class QuantifiedExpression (Expression):

    FORALL = "\\forall"
    EXISTS = "\\exists"
    quantifier = None
    variables = []
    scope = None   # An expression
        
    def text(self):
        types = set([v.typ.text() for v in self.variables])
        if len(types) == 1:
            vt = self.variables[0].typ.text() + ' ' + ', '.join([v.val.text() for v in self.variables])
        else:
            vt = 'FIX SYNTAX' + ', '.join([v.text() for v in self.variables])
        s = '(' + self.quantifier + " " + vt + "; " + self.scope.text() +')'
        return s
    


class QuantifiedVariable (ASTNode):

    typ = None
    val = ""

    def text(self):
        return self.typ.text() + " " + self.val.text()
