import copy
import JML_ast



class ReachLocation(object):

    # Node where the search starts
    root = None
    # Current node, when found
    cur_node = None
    # Has the current node been found?
    found = False

    def __init__(self, root):
        self.root = root

    def getto_cur_node(self):
        self.root.accept(self)
        return self.cur_node

    def visit(self, node):
        if node.cur_pos:
            self.cur_node = node
            self.found = True

    def ConditionalExpression_visit(self, node):
        self.visit(node)
        node.condition.accept(self)
        node.then_exp.accept(self)
        node.else_exp.accept(self)

    def BinaryExpression_visit(self, node):
        self.visit(node)
        node.left.accept(self)
        node.right.accept(self)

    def UnaryExpression_visit(self, node):
        self.visit(node)
        node.item.accept(self)

    def FeatureCall_visit(self, node):
        self.visit(node)
        if node.target is not None:
            node.target.accept(self)
        for c in node.call:
            c.accept(self)

    def MethodCall_visit(self, node):
        self.FeatureCall_visit(node)
        for a in node.args:
            a.accept(self)

    def ArrayReference_visit(self, node):
        self.FeatureCall_visit(node)
        for a in node.args:
            a.accept(self)

    def QuantifiedExpression_visit(self, node):
        self.visit(node)
        for v in node.variables:
            v.accept(self)
        if node.scope is not None:
            node.scope.accept(self)
        

class Substitutions(object):

    # Expression to use as replacement
    nexp = None
    # Root node in the AST
    root = None
    # Current substitution
    current = None

    _all_subs = None
    _p = None

    itpos = None

    def __init__(self, nexp, root):
        self.nexp = nexp
        self.root = root
        # The first mutation is the indentity
        # self._all_subs = [root]
        # self._p = 1
        self._all_subs = list()
        self._p = 0

    def __iter__(self):
        self.itpos = 0
        self.root.accept(self)
        return self

    def next(self):
        if self.itpos >= self._p:
            raise StopIteration
        else:
            self.itpos += 1
            return self._all_subs[self.itpos - 1]

    def recopy(self):
        self.current = copy.deepcopy(self.root)
        self._all_subs.append(self.current)
        self._p += 1
        r = ReachLocation(self.current)
        n = r.getto_cur_node()
        assert r.found
        n.cur_pos = False  # For further substitutions
        return n

    def visit(self, node):
        pass

    def ConditionalExpression_visit(self, node):
        node.condition.accept(self)
        node.then_exp.accept(self)
        node.else_exp.accept(self)

    def BinaryExpression_visit(self, node):
        node.left.accept(self)
        node.right.accept(self)

    def UnaryExpression_visit(self, node):
        node.item.accept(self)

    def FeatureCall_visit(self, node):
        if node.target is not None:
            node.target.accept(self)
        for c in node.call:
            if c is not None:
                c.accept(self)

    def MethodCall_visit(self, node):
        self.FeatureCall_visit(node)
        for a in node.args:
            if a is not None:
                a.accept(self)

    def ArrayReference_visit(self, node):
        # Do not directly call MethodCall_visit because behavior may differ in some classes
        # E.g.: Boolean substitutions
        self.FeatureCall_visit(node)
        for a in node.args:
            if a is not None:
                a.accept(self)

    def QuantifiedExpression_visit(self, node):
        node.scope.accept(self)



class IntSubstitutions(Substitutions):

    def ConditionalExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.then_exp = self.nexp
        r = self.recopy()
        r.else_exp = self.nexp
        node.cur_pos = False
        super(IntSubstitutions, self).ConditionalExpression_visit(node)

    def EqualityExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.left = self.nexp
        r = self.recopy()
        r.right = self.nexp
        node.cur_pos = False
        super(IntSubstitutions, self).BinaryExpression_visit(node)

    def RelationalExpression_visit(self, node):
        self.EqualityExpression_visit(node)

    def AdditiveExpression_visit(self, node):
        self.EqualityExpression_visit(node)

    def MultiplicativeExpression_visit(self, node):
        self.EqualityExpression_visit(node)
            
    def PlusUnaryExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.item = self.nexp
        node.cur_pos = False
        super(IntSubstitutions, self).UnaryExpression_visit(node)

    def MinusUnaryExpression_visit(self, node):
        self.PlusUnaryExpression_visit(node)

    def MethodCall_visit(self, node):
        node.cur_pos = True
        for x in range(len(node.args)):
            r = self.recopy()
            r.args[x] = self.nexp
        node.cur_pos = False
        super(IntSubstitutions, self).MethodCall_visit(node)

    def ArrayReference_visit(self, node):
        self.MethodCall_visit(node)



class BooleanSubstitutions(Substitutions):

    def __init__(self, nexp, root):
        super(BooleanSubstitutions, self).__init__(nexp, root)
        # The first mutation is nexp replacing the whole root
        self._all_subs.append(nexp)
        self._p += 1


    # Assuming that conditional expressions are not used as booleans (use an implication instead)
    def ConditionalExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.condition = self.nexp
        node.cur_pos = False
        super(BooleanSubstitutions, self).ConditionalExpression_visit(node)

    def EquivalenceExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.left = self.nexp
        r = self.recopy()
        r.right = self.nexp
        node.cur_pos = False
        super(BooleanSubstitutions, self).BinaryExpression_visit(node)

    def ImplicationExpression_visit(self, node):
        self.EquivalenceExpression_visit(node)

    def OrExpression_visit(self, node):
        self.EquivalenceExpression_visit(node)

    def AndExpression_visit(self, node):
        self.EquivalenceExpression_visit(node)

    # Assuming that == and != expressions are not used for booleans (use <==>, <=!=> instead)

    def NotExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.item = self.nexp
        node.cur_pos = False
        super(BooleanSubstitutions, self).UnaryExpression_visit(node)
            
    def MethodCall_visit(self, node):
        node.cur_pos = True
        for x in range(len(node.args)):
            r = self.recopy()
            r.args[x] = self.nexp
        node.cur_pos = False
        super(BooleanSubstitutions, self).MethodCall_visit(node)

    # Note that array indexes must be integer expressions

    def QuantifiedExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.scope = self.nexp
        node.cur_pos = False
        super(BooleanSubstitutions, self).QuantifiedExpression_visit(node)



class ArraySubstitutions(Substitutions):
    """In the constructor: nexp is an array expression of the form a[k], 
       possibly dotted as a.b.c[k], and k is a placeholder"""

    def ArrayReference_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.target = self.nexp.target
        r.call = self.nexp.call
        node.cur_pos = False
        super(ArraySubstitutions, self).ArrayReference_visit(node)



class CallableSubstitutions(Substitutions):
    """In the constructor: nexp is a method call of the form a(k1, ..., kn), 
       possibly dotted as a.b.c(k1, ..., kn) and k1, ..., kn are placeholders"""

    def MethodCall_visit(self, node):
        """Does replacement only if number of arguments coincide"""
        if count(node.args) == count(self.nexp.args):
            node.cur_pos = True
            r = self.recopy()
            r.target = self.nexp.target
            r.call = self.nexp.call
            node.cur_pos = False
        super(CallableSubstitutions, self).MethodCall_visit(node)



class AgingSubstitutions(Substitutions):
    """In the constructor: nexp is an integer expression (typically, + 1 or - 1) 
       added to other integer expressions"""

    def add_combine(self, old, new):
        res = JML_ast.AddExpression()
        res.left = old
        res.right = new
        return res

    def ConditionalExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.then_exp = self.add_combine(r.then_exp, self.nexp)
        r = self.recopy()
        r.else_exp = self.add_combine(r.else_exp, self.nexp)
        node.cur_pos = False
        super(AgingSubstitutions, self).ConditionalExpression_visit(node)

    def EqualityExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.left = self.add_combine(r.left, self.nexp)
        r = self.recopy()
        r.right = self.add_combine(r.right, self.nexp)
        node.cur_pos = False
        super(AgingSubstitutions, self).BinaryExpression_visit(node)

    def RelationalExpression_visit(self, node):
        self.EqualityExpression_visit(node)

    def AdditiveExpression_visit(self, node):
        self.EqualityExpression_visit(node)

    def MultiplicativeExpression_visit(self, node):
        self.EqualityExpression_visit(node)
            
    def PlusUnaryExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.item = self.add_combine(r.item, self.nexp)
        node.cur_pos = False
        super(AgingSubstitutions, self).UnaryExpression_visit(node)

    def MinusUnaryExpression_visit(self, node):
        self.PlusUnaryExpression_visit(node)

    def MethodCall_visit(self, node):
        node.cur_pos = True
        for x in range(len(node.args)):
            r = self.recopy()
            r.args[x] = self.add_combine(r.args[x], self.nexp)
        node.cur_pos = False
        super(AgingSubstitutions, self).MethodCall_visit(node)

    def ArrayReference_visit(self, node):
        self.MethodCall_visit(node)



class WeakeningSubstitutions(Substitutions):
    """In the constructor: nexp is a boolean expression 
       added as antecedent to other boolean expressions"""

    def __init__(self, nexp, root):
        super(WeakeningSubstitutions, self).__init__(nexp, root)
        # The first mutation is nexp weakening the whole root
        self._all_subs.append(self.impl_combine(nexp, root))
        self._p += 1

    def impl_combine(self, ant, cons):
        res = JML_ast.ImpliesExpression()
        res.left = ant
        res.right = cons
        return res

    def or_combine(self, left, right):
        res = JML_ast.OrExpression()
        res.left = ant
        res.right = cons
        return res

    def ConditionalExpression_visit(self, node):
        # Not sure it's useful in practice; commenting out for now
        # node.cur_pos = True
        # r = self.recopy()
        # r.condition = self.or_combine(r.condition, self.nexp)
        # node.cur_pos = False
        super(WeakeningSubstitutions, self).ConditionalExpression_visit(node)

    def EquivalenceExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.left = self.impl_combine(self.nexp, r.left)
        r = self.recopy()
        r.right = self.impl_combine(self.nexp, r.right)
        node.cur_pos = False
        super(WeakeningSubstitutions, self).BinaryExpression_visit(node)

    def ImplicationExpression_visit(self, node):
        self.EquivalenceExpression_visit(node)

    def OrExpression_visit(self, node):
        self.EquivalenceExpression_visit(node)

    def AndExpression_visit(self, node):
        self.EquivalenceExpression_visit(node)

    def NotExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.item = self.impl_combine(self.nexp, r.item)
        node.cur_pos = False
        super(WeakeningSubstitutions, self).UnaryExpression_visit(node)
            
    def MethodCall_visit(self, node):
        node.cur_pos = True
        for x in range(len(node.args)):
            r = self.recopy()
            r.args[x] = self.impl_combine(self.nexp, r.args[x])
        node.cur_pos = False
        super(WeakeningSubstitutions, self).MethodCall_visit(node)

    def QuantifiedExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        r.scope = self.impl_combine(self.nexp, r.scope)
        node.cur_pos = False
        super(WeakeningSubstitutions, self).QuantifiedExpression_visit(node)



class PredicateExtractor(Substitutions):
    """All atomic boolean expressions in root"""

    def __init__(self, root):
        super(PredicateExtractor, self).__init__(None, root)

    def MethodCall_visit(self, node):
        if len(node.args) > 0:
            self._all_subs.append(node)
            self._p += 1

    # Don't consider array references, assuming they're integer arrays, not booleans

    # Skip \old expressions
    def OldExpression_visit(self, node):
        super(PredicateExtractor, self).MethodCall_visit(node)



class ResultRemover(Substitutions):
    """Remove expressions involving \\result"""

    def __init__(self, root):
        super(ResultRemover, self).__init__(None, root)

    def ConditionalExpression_visit(self, node):
        super(ResultRemover, self).ConditionalExpression_visit(node)
        node._res = condition._res or then_exp._res or else_exp._res

    def BinaryExpression_visit(self, node):
        super(ResultRemover, self).BinaryExpression_visit(node)
        node._res = node.left._res or node.right._res
        node.cur_pos = True
        if node.left._res:
            r = self.recopy()
            r.left = copy.deepcopy(r.right)
            r = self.recopy()
            ne = JML_ast.NotExpression()
            ne.item = copy.deepcopy(r.right)
            r.left = ne
        if node.right._res:
            node.cur_pos = True
            r = self.recopy()
            r.right = copy.deepcopy(r.left)
            r = self.recopy()
            ne = JML_ast.NotExpression()
            ne.item = copy.deepcopy(r.left)
            r.right = ne
        node.cur_pos = False

    def UnaryExpression_visit(self, node):
        super(ResultRemover, self).UnaryExpression_visit(node)
        node._res = node.item._res

    def FeatureCall_visit(self, node):
        super(ResultRemover, self).FeatureCall_visit(node)
        if node.target is not None:
            node._res = node.target._res
        if node.call is not None:
            for c in node.call:
                node._res = node._res or c._res
        
    def ResultExpression_visit(self, node):
        node._res = True



class OldAdder(Substitutions):
    """Add \\old expressions to operands of binary and unary expressions"""

    def BinaryExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        oe = JML_ast.OldExpression()
        oe.args = [r.left]
        r.left = oe
        r = self.recopy()
        oe = JML_ast.OldExpression()
        oe.args = [r.right]
        r.right = oe
        node.cur_pos = False
        super(OldAdder, self).BinaryExpression_visit(node)

    def UnaryExpression_visit(self, node):
        node.cur_pos = True
        r = self.recopy()
        oe = JML_ast.OldExpression()
        oe.args = [r.item]
        r.item = oe
        node.cur_pos = False
        super(OldAdder, self).UnaryExpression_visit(node)
