package java.io;

import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.apache.commons.vfs2.FileSystemException;
import org.apache.commons.vfs2.FileSystemManager;
import org.apache.commons.vfs2.VFS;

/**
 * Provides an interface between the evosuite main project and the evosuite-io subproject. (The separation was necessary, because evosuite-io would
 * not compile in some Java Virtual Machines.) Cannot be instantiated. (One may consider to transform this class into a singleton but I see no gain
 * for that atm. Quite the contrary, it would probably slightly degrade performance.)
 * 
 * @author Daniel Muth
 */
public class EvoSuiteIO {

	/**
	 * flag to turn on and off VFS functionality; if activeVFS is false, the original implementation will be used for all files, even virtual ones!
	 * 
	 * @see #enableVFS()
	 * @see #disableVFS()
	 * @see #isActiveVFS()
	 */
	static boolean activeVFS = false;

	/**
	 * NEW IN VERSION 0.5 to prevent bad things to happen in the real file system if evosuite-io or in particular {@link #activeVFS} is accessed
	 * multi-threaded (although evosuite-io does still clearly NOT support multi-threading!)
	 */
	static ReentrantReadWriteLock rwlock = new ReentrantReadWriteLock();

	/**
	 * tracks all files, identified by their canonical path, that were accessed by code under test; 'original' files are filtered out
	 */
	static Set<String> filesAccessedByCUT = new HashSet<String>();

	/**
	 * the FileSystemManager of Commons VFS; basically needed for every access to some virtual file
	 */
	private static FileSystemManager fsm = null;

	/**
	 * flag which indicates if all necessary information was passed from evosuite to evosuite-io
	 */
	private static boolean initialized = false;

	/** the package under test */
	static String testPackage;

	/**
	 * indicates if evosuite-io shall copy the state of an original file that does NOT reside in the sandbox read folder into the RAM VFS (or not);
	 * setting this to true and having an empty sandbox read folder produces test cases which assume that no accessed file by the CUT exists what can
	 * be desirable
	 */
	static boolean readOnlyFromSandboxFolder = false; // TODO hand it over from master to client? (if not yet done)

	/**
	 * the sandbox read folder as an absolute file
	 */
	static File sandboxReadFolder;

	/**
	 * just a simple switch to enable/disable a handful of <code>IOException.printStackTrace()</code> calls - used by File, FileInputStream and
	 * FileOutputStream
	 */
	static boolean debug = true;

	/**
	 * this flag determines:<br>
	 * 1.) ( = <code>true</code> ) if undeclared IOExceptions (sometimes necessary since I must not change the API of public methods and constructors
	 * of existing java.io classes) for which no error handling exist shall be catched in a 'silent' manner (at the utmost with an
	 * <code>exception.printStackTrace())</code> leading to methods returning normally even if the program state <b>may</b> be corrupt or <br>
	 * 2.) ( = <code>false</code> ) wrap such an exception into an IllegalStateException and throw it (leading to interruption of test case execution
	 * by EvoSuite) <br>
	 * - used by File, FileInputStream and FileOutputStream -
	 */
	static boolean tolerantExceptionHandling = false;

	/**
	 * private constructor to prevent instantiation by other classes
	 */
	private EvoSuiteIO() {
	}

	/**
	 * Creates and caches the Commons VFS FileSystemManager instance, that is basically needed for every access to a virtual file. The
	 * FileSystemManager instance has to be created when the VFS used by EvoSuite I/O is inactive what is ensured in this method, therefore direct use
	 * of {@link VFS#getManager()} is discouraged.
	 * 
	 * @return the Commons VFS FileSystemManager instance
	 * @throws FileSystemException
	 *             if an error occurs creating the manager
	 * @see VFS#getManager()
	 */
	static FileSystemManager getVFSManager() throws FileSystemException {
		if (fsm == null) {
			final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
			final boolean VFSwasActive = activeVFS;
			try {
				activeVFS = false;
				fsm = VFS.getManager();
			} finally {
				activeVFS = VFSwasActive;
				reacquireReadLocksAndReleaseWriteLock(holdCount);
			}
		}
		return fsm;
	}

	/**
	 * Resets the VFS to a clean, initial state in which all files in the VFS are deleted and every collection is cleared (also the set of file paths
	 * accessed by the CUT). After calling this method every newly accessed virtual file will get synchronized with its correspondent original file
	 * (if it exists and reading it is allowed). The activation state of the VFS after returnation of this method is the same as when entering this
	 * method.
	 * 
	 * @throws IOException
	 *             if resetting the VFS failed
	 */
	public static void resetVFS() throws IOException {
		final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
		final boolean VFSwasActive = activeVFS; // save the current activation state of the VFS (for the finally-block)
		try {
			// this deletes all files that have been created by EvoSuiteIO in Commons VFS RAM
			activeVFS = true;
			for (File root : File.listRoots()) { // FIXME does not work correctly for Unix (root cannot be deleted)
				setOriginal(root, false);
				setShallowCopy(root, true);
				try {
					forceReinitialization(root);
				} catch (FileSystemException e) {
					System.gc(); // closes unreferenced open streams that may have caused the FileSystemException
					forceReinitialization(root); // try again
				}
			}
		} finally {
			activeVFS = VFSwasActive; // set the activation state to the observed state when entering this method
			reacquireReadLocksAndReleaseWriteLock(holdCount);
			filesAccessedByCUT.clear();
			File.deletedFiles.clear(); // forces all newly created virtual files to be reinitialized again with their correspondent original file
			File.permissions.clear();
			File.shallowCopiedDirectories.clear();
			File.shallowCopiedFiles.clear();
		}
	}

	/**
	 * Determines if EvoSuite I/O was initialized with all necessary information to work properly with EvoSuite
	 * 
	 * @return <code>true</code>, if {@link EvoSuiteIO#initialize(String, boolean, File)} was already called, <code>false</code>
	 */
	public static boolean isInitialized() {
		return initialized;
	}

	/**
	 * must be called before the first use of the VFS in order to get it work properly
	 * 
	 * @param testPackage
	 *            the name of the package under test
	 * @param readOnlyFromSandboxFolder
	 *            <code>true</code> if EvoSuite I/O shall only read from the EvoSuite sandbox read folder, <code>false</code> if it also may read from
	 *            the original path if a file was not found in the sandbox read folder
	 * @param sandboxReadFolder
	 *            an absolute File instance denoting the absolute pathname of the sandbox read folder of EvoSuite
	 */
	public static void initialize(String testPackage, boolean readOnlyFromSandboxFolder, File sandboxReadFolder) {
		EvoSuiteIO.testPackage = testPackage;
		EvoSuiteIO.readOnlyFromSandboxFolder = readOnlyFromSandboxFolder;
		EvoSuiteIO.sandboxReadFolder = sandboxReadFolder;
		EvoSuiteIO.initialized = true;
	}

	/**
	 * Determines if files were accessed by the code under test. Internally: Checks if the EvoSuiteIO.filesAccessedByCUT set, that is only filled if
	 * EvoSuite I/O has been initialized and if the VFS is active, is empty or not
	 * 
	 * @return <code>true</code> if files were accessed by the code under test, <code>false</code> otherwise
	 * @see #isActiveVFS()
	 * @see #isInitialized()
	 */
	public static boolean filesWereAccessedByCUT() {
		return !filesAccessedByCUT.isEmpty();
	}

	/**
	 * Note that the set returned by this method is deleted when {@link #resetVFS()} is called
	 * 
	 * @return a set of canonical paths denoting the files that were accessed by the code under test
	 */
	public static Set<String> getFilesAccessedByCUT() {
		return filesAccessedByCUT;
	}

	/**
	 * Enables a handful of <code>IOException.printStackTrace()</code> calls for IOExceptions that may occur when using virtual files and are catched
	 * internally in order to preserve the public API of already existing classes in java.io.<br>
	 * Only meaningful if tolerant exception handling is enabled!
	 * 
	 * @see #disableTolerantExceptionHandling()
	 * @see #disablePrintStackTraceCalls()
	 * @see #enableTolerantExceptionHandling()
	 */
	public static void enablePrintStackTraceCalls() {
		debug = true;
	}

	/**
	 * Disables a handful of <code>IOException.printStackTrace()</code> calls for IOExceptions that may occur when using virtual files and are catched
	 * internally in order to preserve the public API of already existing classes in java.io.
	 * 
	 * @see #enablePrintStackTraceCalls()
	 * @see #enableTolerantExceptionHandling()
	 * @see #disableTolerantExceptionHandling()
	 */
	public static void disablePrintStackTraceCalls() {
		debug = false;
	}

	/**
	 * Enables tolerant exception handling for IOExceptions that may occur when using virtual files and <b>may</b> lead to a corrupt state of the VFS
	 * (and its virtual files). This means, that these exceptions are catched 'silently' (at the utmost with an
	 * <code>exception.printStackTrace()</code> call) leading to the method/constructor returning normally.
	 * 
	 * @see #enablePrintStackTraceCalls()
	 * @see #disablePrintStackTraceCalls()
	 * @see #disableTolerantExceptionHandling()
	 */
	public static void enableTolerantExceptionHandling() {
		tolerantExceptionHandling = true;
	}

	/**
	 * Disables tolerant exception handling for IOExceptions, that may occur when using virtual files and <b>may</b> lead to a corrupt state of the
	 * VFS (and its virtual files). This means, instead of silently catching these exceptions they are wrapped into an IllegalStateException and are
	 * thrown leading to EvoSuite interrupting the current test case execution.
	 * 
	 * @see #enableTolerantExceptionHandling()
	 * @see #enablePrintStackTraceCalls()
	 * @see #disablePrintStackTraceCalls()
	 */
	public static void disableTolerantExceptionHandling() {
		tolerantExceptionHandling = false;
	}

	/**
	 * Enables VFS functionality.
	 */
	public static void enableVFS() {
		rwlock.writeLock().lock();
		activeVFS = true;
		rwlock.writeLock().unlock();
	}

	/**
	 * Disables VFS functionality. Even files labled as virtual will be treated with the original implementation.
	 */
	public static void disableVFS() {
		rwlock.writeLock().lock();
		activeVFS = false;
		rwlock.writeLock().unlock();
	}

	/**
	 * @return <code>true</code> if VFS functionality is active, <code>false</code> otherwise
	 */
	public static boolean isActiveVFS() {
		rwlock.readLock().lock();
		try {
			return activeVFS;
		} finally {
			rwlock.readLock().unlock();
		}
	}

	/**
	 * Any virtual file whose correspondent original file (if existing) is greater in size than this threshold (in bytes) is not copied into memory
	 * until its content is accessed via a FileInputStream, FileOutputStream or File.renameTo. This threshold was introduced to save performance and
	 * is now set to zero bytes by default, but it may still be useful, so I leave it implemented.
	 * 
	 * @param bytes
	 *            the file size in bytes
	 */
	public static void setShallowCopyThreshold(long bytes) {
		File.shallowCopyThreshold = bytes;
	}

	/**
	 * determines, if the given file should be used within the orignal file system
	 * 
	 * @param file
	 *            the file to be analyzed
	 * @return true, if the given file should be treated within the original "real" file system
	 */
	static boolean shouldBeOriginal(File file) {
		if (!EvoSuiteIO.isInitialized())
			return true;

		// certain stack trace elements which indicate the file to be (not) original // TODO this is actually bad style - find better solution
		StackTraceElement[] stackTraceElements = Thread.currentThread().getStackTrace(); // TODO possible issue with security manager?

		boolean calledByClassUnderTest = false;
		for (int elementCounter = 0; elementCounter < stackTraceElements.length; elementCounter++) {
			StackTraceElement e = stackTraceElements[elementCounter];

			// if we want to access the original file with Commons VFS in order to copy it into RAM VFS we of course NEED the original file!
			if (e.getClassName().equals("org.apache.commons.vfs2.provider.local.LocalFile"))
				return true;

			// we also stay with the original implementation if a class loader is involved, otherwise we may get a chicken-and-egg problem (think of
			// the case a Commons VFS class shall be loaded from a class file!)
			if (e.getClassName().contains("ClassLoader"))
				return true;

			// determine the full package name of the class returned by e.getClassName()
			String[] fullyQualifiedClassName = e.getClassName().split("\\.");
			String packageName = "";
			if (fullyQualifiedClassName.length > 1) {
				packageName = fullyQualifiedClassName[0]; // root package
				for (int i = 1; i < fullyQualifiedClassName.length - 1; i++)
					// possible sub-packages
					packageName += "." + fullyQualifiedClassName[i];
			}
			 
			if (packageName.equals(EvoSuiteIO.testPackage) && !EvoSuiteIO.testPackage.equals("")) {
				calledByClassUnderTest = true;
			}
		}

		if (calledByClassUnderTest) {
			try {
				EvoSuiteIO.filesAccessedByCUT.add(file.getCanonicalPath());
			} catch (IOException e) {
				if (debug) {
					e.printStackTrace();
				}
			}
		}
		
		return false;
	}

	/**
	 * delegate method for {@link File#forceReinitialization}
	 */
	public static void forceReinitialization(File file) throws IOException {
		file.forceReinitialization();
	}

	/**
	 * delegate method for {@link File#deepDelete}
	 */
	public static boolean deepDelete(File file) throws IOException {
		return file.deepDelete();
	}

	/**
	 * delegate method for {@link File#isOriginal}
	 * 
	 * @return false, if the file should be handled by the vfs, true otherwise
	 */
	public static boolean isOriginal(File file) {
		return file.isOriginal();
	}

	/**
	 * delegate method for {@link File#setOriginal}; labels the file as original file, so that it won't be handled by the vfs
	 */
	public static void setOriginal(File file, boolean original) {
		file.setOriginal(original);
	}

	/**
	 * delegate method for {@link File#isShallowCopy()}
	 */
	public static boolean isShallowCopy(File file) {
		return file.isShallowCopy();
	}

	/**
	 * delegate method for {@link File#setShallowCopy(boolean)}
	 */
	public static void setShallowCopy(File file, boolean shallowCopy) {
		file.setShallowCopy(shallowCopy);
	}

	/**
	 * The calling thread releases all read locks holding for {@link EvoSuiteIO#rwlock} so that it can acquire the write lock (Idea taken from <a
	 * href="http://stackoverflow.com/a/464824/1236781">Stackoverflow.com</a>)
	 * 
	 * @return the number of read locks that were released by the calling thread
	 */
	static int releaseAllReadLocksAndAcquireWriteLock() {
		final int holdCount = rwlock.getReadHoldCount();
		for (int i = 0; i < holdCount; i++)
			rwlock.readLock().unlock();
		rwlock.writeLock().lock();
		return holdCount;
	}

	/**
	 * Before releasing the write lock, the calling thread reacquires all read locks again that it released in
	 * {@link #releaseAllReadLocksAndAcquireWriteLock()} (Idea taken from <a href="http://stackoverflow.com/a/464824/1236781">Stackoverflow.com</a>)
	 * 
	 * @param holdCount
	 *            the number of read locks to reacquire
	 */
	static void reacquireReadLocksAndReleaseWriteLock(int holdCount) {
		for (int i = 0; i < holdCount; i++)
			rwlock.readLock().lock();
		rwlock.writeLock().unlock();
	}

}
