/*
 * %W% %E% Copyright (c) 2006, Oracle and/or its affiliates. All rights reserved. ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package java.io;

import static java.io.EvoSuiteIO.rwlock;
import static java.io.EvoSuiteIO.releaseAllReadLocksAndAcquireWriteLock;
import static java.io.EvoSuiteIO.reacquireReadLocksAndReleaseWriteLock;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.security.AccessControlException;
import java.security.AccessController;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.vfs2.FileObject;
import org.apache.commons.vfs2.FileSystemException;
import org.apache.commons.vfs2.FileSystemManager;
import org.apache.commons.vfs2.FileType;
import org.apache.commons.vfs2.Selectors;
import org.apache.commons.vfs2.provider.ram.RamFileObject;

import sun.security.action.GetPropertyAction;

/**
 * ++++++ modified version of the 'original' <code>File</code> class; additionally allows complete in-memory I/O ++++++
 * <p>
 * An abstract representation of file and directory pathnames.
 * <p>
 * User interfaces and operating systems use system-dependent <em>pathname
 * strings</em> to name files and directories. This class presents an abstract, system-independent view of hierarchical pathnames. An
 * <em>abstract pathname</em> has two components:
 * <ol>
 * <li>An optional system-dependent <em>prefix</em> string, such as a disk-drive specifier, <code>"/"</code>&nbsp;for the UNIX root directory, or
 * <code>"\\\\"</code>&nbsp;for a Microsoft Windows UNC pathname, and
 * <li>A sequence of zero or more string <em>names</em>.
 * </ol>
 * The first name in an abstract pathname may be a directory name or, in the case of Microsoft Windows UNC pathnames, a hostname. Each subsequent name
 * in an abstract pathname denotes a directory; the last name may denote either a directory or a file. The <em>empty</em> abstract pathname has no
 * prefix and an empty name sequence.
 * <p>
 * The conversion of a pathname string to or from an abstract pathname is inherently system-dependent. When an abstract pathname is converted into a
 * pathname string, each name is separated from the next by a single copy of the default <em>separator character</em>. The default name-separator
 * character is defined by the system property <code>file.separator</code>, and is made available in the public static fields <code>{@link
 * #separator}</code> and <code>{@link #separatorChar}</code> of this class. When a pathname string is converted into an abstract pathname, the names
 * within it may be separated by the default name-separator character or by any other name-separator character that is supported by the underlying
 * system.
 * <p>
 * A pathname, whether abstract or in string form, may be either <em>absolute</em> or <em>relative</em>. An absolute pathname is complete in that no
 * other information is required in order to locate the file that it denotes. A relative pathname, in contrast, must be interpreted in terms of
 * information taken from some other pathname. By default the classes in the <code>java.io</code> package always resolve relative pathnames against
 * the current user directory. This directory is named by the system property <code>user.dir</code>, and is typically the directory in which the Java
 * virtual machine was invoked.
 * <p>
 * The <em>parent</em> of an abstract pathname may be obtained by invoking the {@link #getParent} method of this class and consists of the pathname's
 * prefix and each name in the pathname's name sequence except for the last. Each directory's absolute pathname is an ancestor of any <tt>File</tt>
 * object with an absolute abstract pathname which begins with the directory's absolute pathname. For example, the directory denoted by the abstract
 * pathname <tt>"/usr"</tt> is an ancestor of the directory denoted by the pathname <tt>"/usr/local/bin"</tt>.
 * <p>
 * The prefix concept is used to handle root directories on UNIX platforms, and drive specifiers, root directories and UNC pathnames on Microsoft
 * Windows platforms, as follows:
 * <ul>
 * <li>For UNIX platforms, the prefix of an absolute pathname is always <code>"/"</code>. Relative pathnames have no prefix. The abstract pathname
 * denoting the root directory has the prefix <code>"/"</code> and an empty name sequence.
 * <li>For Microsoft Windows platforms, the prefix of a pathname that contains a drive specifier consists of the drive letter followed by
 * <code>":"</code> and possibly followed by <code>"\\"</code> if the pathname is absolute. The prefix of a UNC pathname is <code>"\\\\"</code>; the
 * hostname and the share name are the first two names in the name sequence. A relative pathname that does not specify a drive has no prefix.
 * </ul>
 * <p>
 * Instances of this class may or may not denote an actual file-system object such as a file or a directory. If it does denote such an object then
 * that object resides in a <i>partition</i>. A partition is an operating system-specific portion of storage for a file system. A single storage
 * device (e.g. a physical disk-drive, flash memory, CD-ROM) may contain multiple partitions. The object, if any, will reside on the partition <a
 * name="partName">named</a> by some ancestor of the absolute form of this pathname.
 * <p>
 * A file system may implement restrictions to certain operations on the actual file-system object, such as reading, writing, and executing. These
 * restrictions are collectively known as <i>access permissions</i>. The file system may have multiple sets of access permissions on a single object.
 * For example, one set may apply to the object's <i>owner</i>, and another may apply to all other users. The access permissions on an object may
 * cause some methods in this class to fail.
 * <p>
 * Instances of the <code>File</code> class are immutable; that is, once created, the abstract pathname represented by a <code>File</code> object will
 * never change.
 * 
 * @version %I%, %G%
 * @author unascribed
 * @author Daniel Muth
 * @since JDK1.0
 */
public class File implements Serializable, Comparable<File> {
	/**
	 * indicates, if this file shall be handled like an 'original' file (in its original implementation)
	 */
	private boolean isOriginal = true; // MEMO TO MYSELF: The default value MUST stay 'true'!

	/**
	 * indicates, if this file shall be shallow copied from disk to memory during initialization of the ramFile, that means:
	 * <ul>
	 * <li>for a file: no content is copied</li>
	 * <li>for a directory: no descendants are copied</li>
	 * </ul>
	 */
	private boolean shallowCopy = false;

	/**
	 * the mapped original file (either in sandbox folder or in original folder) (may be <code>null</code>)
	 */
	private File fileToCopy;

	/**
	 * the correspondent memory representation of this file (may be <code>null</code>)
	 */
	private FileObject ramFile;

	/**
	 * a global map to keep track of file permissions; the mapping consists of canonicalPath -> permission (binary flag)
	 */
	static Map<String, Integer> permissions = new HashMap<String, Integer>();

	/**
	 * saves canonical paths of virtual files, that have been deleted and shall not be reinitialized again
	 */
	static Set<String> deletedFiles = new HashSet<String>();

	/**
	 * all files with size greater than this threshold (in bytes) are to be shallow copied into memory
	 */
	static long shallowCopyThreshold = 0L; // 10485760 Bytes = 10 MB

	/**
	 * a set of canonical paths denoting directories that have only been shallow copied into memory
	 */
	static Set<String> shallowCopiedDirectories = new HashSet<String>();

	/**
	 * a set of canonical paths denoting files that have only been shallow copied into memory
	 */
	static Set<String> shallowCopiedFiles = new HashSet<String>();

	/**
	 * The FileSystem object representing the platform's local file system.
	 */
	static private FileSystem fs = FileSystem.getFileSystem();

	/**
	 * This abstract pathname's normalized pathname string. A normalized pathname string uses the default name-separator character and does not
	 * contain any duplicate or redundant separators.
	 * 
	 * @serial
	 */
	private String path;

	/**
	 * The length of this abstract pathname's prefix, or zero if it has no prefix.
	 */
	private transient int prefixLength;

	/**
	 * Returns the length of this abstract pathname's prefix. For use by FileSystem classes.
	 */
	int getPrefixLength() {
		return prefixLength;
	}

	/**
	 * The system-dependent default name-separator character. This field is initialized to contain the first character of the value of the system
	 * property <code>file.separator</code>. On UNIX systems the value of this field is <code>'/'</code>; on Microsoft Windows systems it is
	 * <code>'\\'</code>.
	 * 
	 * @see java.lang.System#getProperty(java.lang.String)
	 */
	public static final char separatorChar = fs.getSeparator();

	/**
	 * The system-dependent default name-separator character, represented as a string for convenience. This string contains a single character, namely
	 * <code>{@link #separatorChar}</code>.
	 */
	public static final String separator = "" + separatorChar;

	/**
	 * The system-dependent path-separator character. This field is initialized to contain the first character of the value of the system property
	 * <code>path.separator</code>. This character is used to separate filenames in a sequence of files given as a <em>path list</em>. On UNIX
	 * systems, this character is <code>':'</code>; on Microsoft Windows systems it is <code>';'</code>.
	 * 
	 * @see java.lang.System#getProperty(java.lang.String)
	 */
	public static final char pathSeparatorChar = fs.getPathSeparator();

	/**
	 * The system-dependent path-separator character, represented as a string for convenience. This string contains a single character, namely
	 * <code>{@link #pathSeparatorChar}</code>.
	 */
	public static final String pathSeparator = "" + pathSeparatorChar;

	/* -- Constructors -- */

	/**
	 * Internal constructor for already-normalized pathname strings.
	 */
	private File(String pathname, int prefixLength) {
		this.path = pathname;
		this.prefixLength = prefixLength;

		rwlock.readLock().lock();
		try {
			if (EvoSuiteIO.activeVFS && !EvoSuiteIO.shouldBeOriginal(this)) {
				setOriginal(false);
			}
		} finally {
			rwlock.readLock().unlock();
		}
	}

	/**
	 * Internal constructor for already-normalized pathname strings. The parameter order is used to disambiguate this method from the public(File,
	 * String) constructor.
	 */
	private File(String child, File parent) {
		assert parent.path != null;
		assert (!parent.path.equals(""));
		this.path = fs.resolve(parent.path, child); // should show no other behaviour when simulation is used
		this.prefixLength = parent.prefixLength;

		rwlock.readLock().lock();
		try {
			if (EvoSuiteIO.activeVFS && !EvoSuiteIO.shouldBeOriginal(this)) {
				setOriginal(false);
			}
		} finally {
			rwlock.readLock().unlock();
		}
	}

	/**
	 * Creates a new <code>File</code> instance by converting the given pathname string into an abstract pathname. If the given string is the empty
	 * string, then the result is the empty abstract pathname.
	 * 
	 * @param pathname
	 *            A pathname string
	 * @throws NullPointerException
	 *             If the <code>pathname</code> argument is <code>null</code>
	 */
	public File(String pathname) {
		if (pathname == null) {
			throw new NullPointerException();
		}
		this.path = fs.normalize(pathname); // should show no other behaviour when simulation is used
		this.prefixLength = fs.prefixLength(this.path); // should show no other behaviour when simulation is used

		rwlock.readLock().lock();
		try {
			if (EvoSuiteIO.activeVFS && !EvoSuiteIO.shouldBeOriginal(this)) {
				setOriginal(false);
			}
		} finally {
			rwlock.readLock().unlock();
		}
	}

	/*
	 * Note: The two-argument File constructors do not interpret an empty parent abstract pathname as the current user directory. An empty parent
	 * instead causes the child to be resolved against the system-dependent directory defined by the FileSystem.getDefaultParent method. On Unix this
	 * default is "/", while on Microsoft Windows it is "\\". This is required for compatibility with the original behavior of this class.
	 */

	/**
	 * Creates a new <code>File</code> instance from a parent pathname string and a child pathname string.
	 * <p>
	 * If <code>parent</code> is <code>null</code> then the new <code>File</code> instance is created as if by invoking the single-argument
	 * <code>File</code> constructor on the given <code>child</code> pathname string.
	 * <p>
	 * Otherwise the <code>parent</code> pathname string is taken to denote a directory, and the <code>child</code> pathname string is taken to denote
	 * either a directory or a file. If the <code>child</code> pathname string is absolute then it is converted into a relative pathname in a
	 * system-dependent way. If <code>parent</code> is the empty string then the new <code>File</code> instance is created by converting
	 * <code>child</code> into an abstract pathname and resolving the result against a system-dependent default directory. Otherwise each pathname
	 * string is converted into an abstract pathname and the child abstract pathname is resolved against the parent.
	 * 
	 * @param parent
	 *            The parent pathname string
	 * @param child
	 *            The child pathname string
	 * @throws NullPointerException
	 *             If <code>child</code> is <code>null</code>
	 */
	public File(String parent, String child) {
		if (child == null) {
			throw new NullPointerException();
		}
		if (parent != null) {
			if (parent.equals("")) {
				this.path = fs.resolve(fs.getDefaultParent(), fs.normalize(child));
			} else {
				this.path = fs.resolve(fs.normalize(parent), fs.normalize(child));
			}
		} else {
			this.path = fs.normalize(child);
		}
		this.prefixLength = fs.prefixLength(this.path);

		rwlock.readLock().lock();
		try {
			if (EvoSuiteIO.activeVFS && !EvoSuiteIO.shouldBeOriginal(this)) {
				setOriginal(false);
			}
		} finally {
			rwlock.readLock().unlock();
		}
	}

	/**
	 * Creates a new <code>File</code> instance from a parent abstract pathname and a child pathname string.
	 * <p>
	 * If <code>parent</code> is <code>null</code> then the new <code>File</code> instance is created as if by invoking the single-argument
	 * <code>File</code> constructor on the given <code>child</code> pathname string.
	 * <p>
	 * Otherwise the <code>parent</code> abstract pathname is taken to denote a directory, and the <code>child</code> pathname string is taken to
	 * denote either a directory or a file. If the <code>child</code> pathname string is absolute then it is converted into a relative pathname in a
	 * system-dependent way. If <code>parent</code> is the empty abstract pathname then the new <code>File</code> instance is created by converting
	 * <code>child</code> into an abstract pathname and resolving the result against a system-dependent default directory. Otherwise each pathname
	 * string is converted into an abstract pathname and the child abstract pathname is resolved against the parent.
	 * 
	 * @param parent
	 *            The parent abstract pathname
	 * @param child
	 *            The child pathname string
	 * @throws NullPointerException
	 *             If <code>child</code> is <code>null</code>
	 */
	public File(File parent, String child) {
		if (child == null) {
			throw new NullPointerException();
		}
		if (parent != null) {
			if (parent.path.equals("")) {
				this.path = fs.resolve(fs.getDefaultParent(), fs.normalize(child));
			} else {
				this.path = fs.resolve(parent.path, fs.normalize(child));
			}
		} else {
			this.path = fs.normalize(child);
		}
		this.prefixLength = fs.prefixLength(this.path);

		rwlock.readLock().lock();
		try {
			if (EvoSuiteIO.activeVFS && !EvoSuiteIO.shouldBeOriginal(this)) {
				setOriginal(false);
			}
		} finally {
			rwlock.readLock().unlock();
		}
	}

	/**
	 * Creates a new <tt>File</tt> instance by converting the given <tt>file:</tt> URI into an abstract pathname.
	 * <p>
	 * The exact form of a <tt>file:</tt> URI is system-dependent, hence the transformation performed by this constructor is also system-dependent.
	 * <p>
	 * For a given abstract pathname <i>f</i> it is guaranteed that <blockquote><tt>
	 * new File(</tt><i>&nbsp;f</i><tt>.{@link #toURI() toURI}()).equals(</tt> <i>&nbsp;f</i><tt>.{@link #getAbsoluteFile() getAbsoluteFile}())
	 * </tt></blockquote> so long as the original abstract pathname, the URI, and the new abstract pathname are all created in (possibly different
	 * invocations of) the same Java virtual machine. This relationship typically does not hold, however, when a <tt>file:</tt> URI that is created in
	 * a virtual machine on one operating system is converted into an abstract pathname in a virtual machine on a different operating system.
	 * 
	 * @param uri
	 *            An absolute, hierarchical URI with a scheme equal to <tt>"file"</tt>, a non-empty path component, and undefined authority, query,
	 *            and fragment components
	 * @throws NullPointerException
	 *             If <tt>uri</tt> is <tt>null</tt>
	 * @throws IllegalArgumentException
	 *             If the preconditions on the parameter do not hold
	 * @see #toURI()
	 * @see java.net.URI
	 * @since 1.4
	 */
	public File(URI uri) { // should show no other behaviour when simulation is used

		// Check our many preconditions
		if (!uri.isAbsolute())
			throw new IllegalArgumentException("URI is not absolute");
		if (uri.isOpaque())
			throw new IllegalArgumentException("URI is not hierarchical");
		String scheme = uri.getScheme();
		if ((scheme == null) || !scheme.equalsIgnoreCase("file"))
			throw new IllegalArgumentException("URI scheme is not \"file\"");
		if (uri.getAuthority() != null)
			throw new IllegalArgumentException("URI has an authority component");
		if (uri.getFragment() != null)
			throw new IllegalArgumentException("URI has a fragment component");
		if (uri.getQuery() != null)
			throw new IllegalArgumentException("URI has a query component");
		String p = uri.getPath();
		if (p.equals(""))
			throw new IllegalArgumentException("URI path component is empty");

		// Okay, now initialize
		p = fs.fromURIPath(p);
		if (File.separatorChar != '/')
			p = p.replace('/', File.separatorChar);
		this.path = fs.normalize(p);
		this.prefixLength = fs.prefixLength(this.path);

		rwlock.readLock().lock();
		try {
			if (EvoSuiteIO.activeVFS && !EvoSuiteIO.shouldBeOriginal(this)) {
				setOriginal(false);
			}
		} finally {
			rwlock.readLock().unlock();
		}
	}

	/* -- Path-component accessors -- */

	/**
	 * Returns the name of the file or directory denoted by this abstract pathname. This is just the last name in the pathname's name sequence. If the
	 * pathname's name sequence is empty, then the empty string is returned.
	 * 
	 * @return The name of the file or directory denoted by this abstract pathname, or the empty string if this pathname's name sequence is empty
	 */
	public String getName() {
		int index = path.lastIndexOf(separatorChar);
		if (index < prefixLength)
			return path.substring(prefixLength);
		return path.substring(index + 1);
	}

	/**
	 * Returns the pathname string of this abstract pathname's parent, or <code>null</code> if this pathname does not name a parent directory.
	 * <p>
	 * The <em>parent</em> of an abstract pathname consists of the pathname's prefix, if any, and each name in the pathname's name sequence except for
	 * the last. If the name sequence is empty then the pathname does not name a parent directory.
	 * 
	 * @return The pathname string of the parent directory named by this abstract pathname, or <code>null</code> if this pathname does not name a
	 *         parent
	 */
	public String getParent() {
		int index = path.lastIndexOf(separatorChar);
		if (index < prefixLength) {
			if ((prefixLength > 0) && (path.length() > prefixLength))
				return path.substring(0, prefixLength);
			return null;
		}
		return path.substring(0, index);
	}

	/**
	 * Returns the abstract pathname of this abstract pathname's parent, or <code>null</code> if this pathname does not name a parent directory.
	 * <p>
	 * The <em>parent</em> of an abstract pathname consists of the pathname's prefix, if any, and each name in the pathname's name sequence except for
	 * the last. If the name sequence is empty then the pathname does not name a parent directory.
	 * 
	 * @return The abstract pathname of the parent directory named by this abstract pathname, or <code>null</code> if this pathname does not name a
	 *         parent
	 * @since 1.2
	 */
	public File getParentFile() {
		String p = this.getParent();
		if (p == null)
			return null;
		return new File(p, this.prefixLength);
	}

	/**
	 * Converts this abstract pathname into a pathname string. The resulting string uses the {@link #separator default name-separator character} to
	 * separate the names in the name sequence.
	 * 
	 * @return The string form of this abstract pathname
	 */
	public String getPath() {
		return path;
	}

	/* -- Path operations -- */

	/**
	 * Tests whether this abstract pathname is absolute. The definition of absolute pathname is system dependent. On UNIX systems, a pathname is
	 * absolute if its prefix is <code>"/"</code>. On Microsoft Windows systems, a pathname is absolute if its prefix is a drive specifier followed by
	 * <code>"\\"</code>, or if its prefix is <code>"\\\\"</code>.
	 * 
	 * @return <code>true</code> if this abstract pathname is absolute, <code>false</code> otherwise
	 */
	public boolean isAbsolute() {
		return fs.isAbsolute(this); // should show no other behaviour when simulation is used
	}

	/**
	 * Returns the absolute pathname string of this abstract pathname.
	 * <p>
	 * If this abstract pathname is already absolute, then the pathname string is simply returned as if by the <code>{@link #getPath}</code> method.
	 * If this abstract pathname is the empty abstract pathname then the pathname string of the current user directory, which is named by the system
	 * property <code>user.dir</code>, is returned. Otherwise this pathname is resolved in a system-dependent way. On UNIX systems, a relative
	 * pathname is made absolute by resolving it against the current user directory. On Microsoft Windows systems, a relative pathname is made
	 * absolute by resolving it against the current directory of the drive named by the pathname, if any; if not, it is resolved against the current
	 * user directory.
	 * 
	 * @return The absolute pathname string denoting the same file or directory as this abstract pathname
	 * @throws SecurityException
	 *             If a required system property value cannot be accessed.
	 * @see java.io.File#isAbsolute()
	 */
	public String getAbsolutePath() {
		return fs.resolve(this); // should show no other behaviour when simulation is used
	}

	/**
	 * Returns the absolute form of this abstract pathname. Equivalent to <code>new&nbsp;File(this.{@link #getAbsolutePath})</code>.
	 * 
	 * @return The absolute abstract pathname denoting the same file or directory as this abstract pathname
	 * @throws SecurityException
	 *             If a required system property value cannot be accessed.
	 * @since 1.2
	 */
	public File getAbsoluteFile() {
		String absPath = getAbsolutePath();
		return new File(absPath, fs.prefixLength(absPath));
	}

	/**
	 * Returns the canonical pathname string of this abstract pathname.
	 * <p>
	 * A canonical pathname is both absolute and unique. The precise definition of canonical form is system-dependent. This method first converts this
	 * pathname to absolute form if necessary, as if by invoking the {@link #getAbsolutePath} method, and then maps it to its unique form in a
	 * system-dependent way. This typically involves removing redundant names such as <tt>"."</tt> and <tt>".."</tt> from the pathname, resolving
	 * symbolic links (on UNIX platforms), and converting drive letters to a standard case (on Microsoft Windows platforms).
	 * <p>
	 * Every pathname that denotes an existing file or directory has a unique canonical form. Every pathname that denotes a nonexistent file or
	 * directory also has a unique canonical form. The canonical form of the pathname of a nonexistent file or directory may be different from the
	 * canonical form of the same pathname after the file or directory is created. Similarly, the canonical form of the pathname of an existing file
	 * or directory may be different from the canonical form of the same pathname after the file or directory is deleted.
	 * 
	 * @return The canonical pathname string denoting the same file or directory as this abstract pathname
	 * @throws IOException
	 *             If an I/O error occurs, which is possible because the construction of the canonical pathname may require filesystem queries
	 * @throws SecurityException
	 *             If a required system property value cannot be accessed, or if a security manager exists and its <code>
	 *             {@link java.lang.SecurityManager#checkRead}</code> method denies read access to the file
	 * @since JDK1.1
	 */
	public String getCanonicalPath() throws IOException {
		// Here, IO takes place, but read only, so we allow it
		// We need this method to uniquely map a real file to its memory correspondent one and vice versa!

		rwlock.readLock().lock();
		try {
			if (EvoSuiteIO.activeVFS) {
				// In the implementation of fs.canonicalize a file object is created that may be labeled and initialized as virtual file. While this
				// principally works, it wastes a lot of unnecessary runtime, so I deactivate the VFS temporarily.
				String canonicalPath;
				final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
				try {
					EvoSuiteIO.activeVFS = false;
					canonicalPath = fs.canonicalize(fs.resolve(this));
				} finally {
					EvoSuiteIO.activeVFS = true; // assures that it is enabled again even if an exception was thrown!
					reacquireReadLocksAndReleaseWriteLock(holdCount);
				}
				return canonicalPath;
			} else {
				return fs.canonicalize(fs.resolve(this));
			}
		} finally {
			rwlock.readLock().unlock();
		}
	}

	/**
	 * Returns the canonical form of this abstract pathname. Equivalent to <code>new&nbsp;File(this.{@link #getCanonicalPath})</code>.
	 * 
	 * @return The canonical pathname string denoting the same file or directory as this abstract pathname
	 * @throws IOException
	 *             If an I/O error occurs, which is possible because the construction of the canonical pathname may require filesystem queries
	 * @throws SecurityException
	 *             If a required system property value cannot be accessed, or if a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead}</code> method denies read access to the file
	 * @since 1.2
	 */
	public File getCanonicalFile() throws IOException {
		String canonPath = getCanonicalPath();
		return new File(canonPath, fs.prefixLength(canonPath));
	}

	private static String slashify(String path, boolean isDirectory) {
		String p = path;
		if (File.separatorChar != '/')
			p = p.replace(File.separatorChar, '/');
		if (!p.startsWith("/"))
			p = "/" + p;
		if (!p.endsWith("/") && isDirectory)
			p = p + "/";
		return p;
	}

	/**
	 * Converts this abstract pathname into a <code>file:</code> URL. The exact form of the URL is system-dependent. If it can be determined that the
	 * file denoted by this abstract pathname is a directory, then the resulting URL will end with a slash.
	 * 
	 * @return A URL object representing the equivalent file URL
	 * @throws MalformedURLException
	 *             If the path cannot be parsed as a URL
	 * @see #toURI()
	 * @see java.net.URI
	 * @see java.net.URI#toURL()
	 * @see java.net.URL
	 * @since 1.2
	 * @deprecated This method does not automatically escape characters that are illegal in URLs. It is recommended that new code convert an abstract
	 *             pathname into a URL by first converting it into a URI, via the {@link #toURI() toURI} method, and then converting the URI into a
	 *             URL via the {@link java.net.URI#toURL() URI.toURL} method.
	 */
	@Deprecated
	public URL toURL() throws MalformedURLException {
		/*
		 * isDirectory() checks for file existence and is guarded by the Security Manager! problematic, but deprecated anyway if isDirectory() is
		 * implemented accordingly, this should cause no problems
		 */
		return new URL("file", "", slashify(getAbsolutePath(), isDirectory()));
	}

	/**
	 * Constructs a <tt>file:</tt> URI that represents this abstract pathname.
	 * <p>
	 * The exact form of the URI is system-dependent. If it can be determined that the file denoted by this abstract pathname is a directory, then the
	 * resulting URI will end with a slash.
	 * <p>
	 * For a given abstract pathname <i>f</i>, it is guaranteed that <blockquote><tt>
	 * new {@link #File(java.net.URI) File}(</tt><i>&nbsp;f</i> <tt>.toURI()).equals(</tt><i>&nbsp;f</i>
	 * <tt>.{@link #getAbsoluteFile() getAbsoluteFile}())
	 * </tt></blockquote> so long as the original abstract pathname, the URI, and the new abstract pathname are all created in (possibly different
	 * invocations of) the same Java virtual machine. Due to the system-dependent nature of abstract pathnames, however, this relationship typically
	 * does not hold when a <tt>file:</tt> URI that is created in a virtual machine on one operating system is converted into an abstract pathname in
	 * a virtual machine on a different operating system.
	 * 
	 * @return An absolute, hierarchical URI with a scheme equal to <tt>"file"</tt>, a path representing this abstract pathname, and undefined
	 *         authority, query, and fragment components
	 * @throws SecurityException
	 *             If a required system property value cannot be accessed.
	 * @see #File(java.net.URI)
	 * @see java.net.URI
	 * @see java.net.URI#toURL()
	 * @since 1.4
	 */
	public URI toURI() {
		try {
			File f = getAbsoluteFile();
			String sp = slashify(f.getPath(), f.isDirectory()); // isDirectory has to be implemented
			if (sp.startsWith("//"))
				sp = "//" + sp;
			return new URI("file", null, sp, null);
		} catch (URISyntaxException x) {
			throw new Error(x); // Can't happen
		}
	}

	/* -- Attribute accessors -- */

	/**
	 * Tests whether the application can read the file denoted by this abstract pathname.
	 * 
	 * @return <code>true</code> if and only if the file specified by this abstract pathname exists <em>and</em> can be read by the application;
	 *         <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 */
	public boolean canRead() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					if (ramFile.exists()) {
						String canonicalPath = getCanonicalPath();
						if (permissions.containsKey(canonicalPath)) {
							return (permissions.get(canonicalPath).intValue() & FileSystem.ACCESS_READ) != 0;
						} else {
							return true; // default value
						}
					} else {
						return false;
					}
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return canReadOriginal();
	}

	/**
	 * Tests whether the application can read the file denoted by this abstract pathname.
	 * 
	 * @return <code>true</code> if and only if the file specified by this abstract pathname exists <em>and</em> can be read by the application;
	 *         <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 */
	private boolean canReadOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkRead(path);
		}
		return fs.checkAccess(this, FileSystem.ACCESS_READ);
	}

	/**
	 * Tests whether the application can modify the file denoted by this abstract pathname.
	 * 
	 * @return <code>true</code> if and only if the file system actually contains a file denoted by this abstract pathname <em>and</em> the
	 *         application is allowed to write to the file; <code>false</code> otherwise.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method denies write access to the file
	 */
	public boolean canWrite() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					if (ramFile.exists()) {
						String canonicalPath = getCanonicalPath();
						if (permissions.containsKey(canonicalPath)) {
							return (permissions.get(canonicalPath).intValue() & FileSystem.ACCESS_WRITE) != 0;
						} else {
							return true; // default value
						}
					} else {
						return false;
					}
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return canWriteOriginal();
	}

	/**
	 * Tests whether the application can modify the file denoted by this abstract pathname.
	 * 
	 * @return <code>true</code> if and only if the file system actually contains a file denoted by this abstract pathname <em>and</em> the
	 *         application is allowed to write to the file; <code>false</code> otherwise.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method denies write access to the file
	 */
	private boolean canWriteOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkWrite(path);
		}
		return fs.checkAccess(this, FileSystem.ACCESS_WRITE);
	}

	/**
	 * Same as {@link #canWriteOriginal()}, except that the security manager check is omitted; only intended for use in {@link #initRamFile()}!
	 */
	private boolean canWriteOriginalWithoutSecurityManagerCheck() {
		// leave out security manager check here //
		return fs.checkAccess(this, FileSystem.ACCESS_WRITE);
	}

	/**
	 * Tests whether the application can execute the file denoted by this abstract pathname.
	 * 
	 * @return <code>true</code> if and only if the abstract pathname exists <em>and</em> the application is allowed to execute the file
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkExec(java.lang.String)}</code> method denies execute access to the file
	 * @since 1.6
	 */
	public boolean canExecute() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					if (ramFile.exists()) {
						String canonicalPath = getCanonicalPath();
						if (permissions.containsKey(canonicalPath)) {
							return (permissions.get(canonicalPath).intValue() & FileSystem.ACCESS_EXECUTE) != 0;
						} else {
							return true; // default value
						}
					} else {
						return false;
					}
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return canExecuteOriginal();
	}

	/**
	 * Tests whether the application can execute the file denoted by this abstract pathname.
	 * 
	 * @return <code>true</code> if and only if the abstract pathname exists <em>and</em> the application is allowed to execute the file
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkExec(java.lang.String)}</code> method denies execute access to the file
	 * @since 1.6
	 */
	private boolean canExecuteOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkExec(path);
		}
		return fs.checkAccess(this, FileSystem.ACCESS_EXECUTE);
	}

	/**
	 * Same as {@link #canExecuteOriginal()}, except that the security manager check is omitted; only intended for use in {@link #initRamFile()}!
	 */
	private boolean canExecuteOriginalWithoutSecurityManagerCheck() {
		// leave out security manager check here //
		return fs.checkAccess(this, FileSystem.ACCESS_EXECUTE);
	}

	/**
	 * Tests whether the file or directory denoted by this abstract pathname exists.
	 * 
	 * @return <code>true</code> if and only if the file or directory denoted by this abstract pathname exists; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file or directory
	 */
	public boolean exists() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					return ramFile.exists();
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return existsOriginal();

	}

	/**
	 * Tests whether the file or directory denoted by this abstract pathname exists.
	 * 
	 * @return <code>true</code> if and only if the file or directory denoted by this abstract pathname exists; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file or directory
	 */
	private boolean existsOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkRead(path);
		}
		return ((fs.getBooleanAttributes(this) & FileSystem.BA_EXISTS) != 0);
	}

	/**
	 * Tests whether the file denoted by this abstract pathname is a directory.
	 * 
	 * @return <code>true</code> if and only if the file denoted by this abstract pathname exists <em>and</em> is a directory; <code>false</code>
	 *         otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 */
	public boolean isDirectory() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					return (ramFile.getType() == FileType.FOLDER);
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return isDirectoryOriginal();
	}

	/**
	 * Tests whether the file denoted by this abstract pathname is a directory.
	 * 
	 * @return <code>true</code> if and only if the file denoted by this abstract pathname exists <em>and</em> is a directory; <code>false</code>
	 *         otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 */
	private boolean isDirectoryOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkRead(path);
		}
		return ((fs.getBooleanAttributes(this) & FileSystem.BA_DIRECTORY) != 0);
	}

	/**
	 * Tests whether the file denoted by this abstract pathname is a normal file. A file is <em>normal</em> if it is not a directory and, in addition,
	 * satisfies other system-dependent criteria. Any non-directory file created by a Java application is guaranteed to be a normal file.
	 * 
	 * @return <code>true</code> if and only if the file denoted by this abstract pathname exists <em>and</em> is a normal file; <code>false</code>
	 *         otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 */
	public boolean isFile() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					return (ramFile.getType() == FileType.FILE);
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return isFileOriginal();

	}

	/**
	 * Tests whether the file denoted by this abstract pathname is a normal file. A file is <em>normal</em> if it is not a directory and, in addition,
	 * satisfies other system-dependent criteria. Any non-directory file created by a Java application is guaranteed to be a normal file.
	 * 
	 * @return <code>true</code> if and only if the file denoted by this abstract pathname exists <em>and</em> is a normal file; <code>false</code>
	 *         otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 */
	private boolean isFileOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkRead(path);
		}
		return ((fs.getBooleanAttributes(this) & FileSystem.BA_REGULAR) != 0);
	}

	/**
	 * Tests whether the file named by this abstract pathname is a hidden file. The exact definition of <em>hidden</em> is system-dependent. On UNIX
	 * systems, a file is considered to be hidden if its name begins with a period character (<code>'.'</code>). On Microsoft Windows systems, a file
	 * is considered to be hidden if it has been marked as such in the filesystem.
	 * 
	 * @return <code>true</code> if and only if the file denoted by this abstract pathname is hidden according to the conventions of the underlying
	 *         platform
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 * @since 1.2
	 */
	public boolean isHidden() {
		/*
		 * In Commons VFS RAM, files and folders cannot be hidden. Furthermore, in Java 6 the 'hidden' attribute cannot be set. Because of that, it
		 * makes no sense to simulate this method on the virtual file - I simply check for existence and if the file/folder exists, I invoke the
		 * original method!
		 */
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					if (!ramFile.exists()) {
						return false;
					}
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				// fallthrough to isHiddenOriginal()
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return isHiddenOriginal();
	}

	/**
	 * Tests whether the file named by this abstract pathname is a hidden file. The exact definition of <em>hidden</em> is system-dependent. On UNIX
	 * systems, a file is considered to be hidden if its name begins with a period character (<code>'.'</code>). On Microsoft Windows systems, a file
	 * is considered to be hidden if it has been marked as such in the filesystem.
	 * 
	 * @return <code>true</code> if and only if the file denoted by this abstract pathname is hidden according to the conventions of the underlying
	 *         platform
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 * @since 1.2
	 */
	private boolean isHiddenOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkRead(path);
		}
		return ((fs.getBooleanAttributes(this) & FileSystem.BA_HIDDEN) != 0);
	}

	/**
	 * Returns the time that the file denoted by this abstract pathname was last modified.
	 * 
	 * @return A <code>long</code> value representing the time the file was last modified, measured in milliseconds since the epoch (00:00:00 GMT,
	 *         January 1, 1970), or <code>0L</code> if the file does not exist or if an I/O error occurs
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 */
	public long lastModified() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					if (!ramFile.exists())
						return 0L;
					return ramFile.getContent().getLastModifiedTime();
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return 0L;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return lastModifiedOriginal();
	}

	/**
	 * Returns the time that the file denoted by this abstract pathname was last modified.
	 * 
	 * @return A <code>long</code> value representing the time the file was last modified, measured in milliseconds since the epoch (00:00:00 GMT,
	 *         January 1, 1970), or <code>0L</code> if the file does not exist or if an I/O error occurs
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 */
	private long lastModifiedOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkRead(path);
		}
		return fs.getLastModifiedTime(this);
	}

	/**
	 * Returns the length of the file denoted by this abstract pathname. The return value is unspecified if this pathname denotes a directory.
	 * 
	 * @return The length, in bytes, of the file denoted by this abstract pathname, or <code>0L</code> if the file does not exist. Some operating
	 *         systems may return <code>0L</code> for pathnames denoting system-dependent entities such as devices or pipes.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 */
	public long length() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					if (isDirectory() || !exists()) {
						return 0L; // the return value for folders is unspecified - on Windows I observed '0L'
					}
					long length = ramFile.getContent().getSize();
					// if the file was shallow copied and not deleted so far, check the original file
					if (length == 0L) {
						String canonPath = getCanonicalPath();
						if (shallowCopiedFiles.contains(canonPath) && !containsFileOrParent(this, deletedFiles)) {
							return lengthOriginal(); // returns 0L if the file does not exist (which is fine)
						}
					}
					return length;
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return 0L;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return lengthOriginal();
	}

	/**
	 * Returns the length of the file denoted by this abstract pathname. The return value is unspecified if this pathname denotes a directory.
	 * 
	 * @return The length, in bytes, of the file denoted by this abstract pathname, or <code>0L</code> if the file does not exist. Some operating
	 *         systems may return <code>0L</code> for pathnames denoting system-dependent entities such as devices or pipes.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the file
	 */
	private long lengthOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkRead(path);
		}
		return fs.getLength(this);
	}

	/* -- File operations -- */

	/**
	 * Atomically creates a new, empty file named by this abstract pathname if and only if a file with this name does not yet exist. The check for the
	 * existence of the file and the creation of the file if it does not exist are a single operation that is atomic with respect to all other
	 * filesystem activities that might affect the file.
	 * <P>
	 * Note: this method should <i>not</i> be used for file-locking, as the resulting protocol cannot be made to work reliably. The
	 * {@link java.nio.channels.FileLock FileLock} facility should be used instead.
	 * 
	 * @return <code>true</code> if the named file does not exist and was successfully created; <code>false</code> if the named file already exists
	 * @throws IOException
	 *             If an I/O error occurred
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method denies
	 *             write access to the file
	 * @since 1.2
	 */
	public boolean createNewFile() throws IOException {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					// fail if target folder does not exist to stay in line with specification!
					File parentFile;
					final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
					try {
						EvoSuiteIO.activeVFS = false; // prevents labeling of the parent with EvoSuiteIO.shouldBeOriginal
						parentFile = getCanonicalFile().getParentFile();
					} finally {
						EvoSuiteIO.activeVFS = true; // assures that VFS is enabled again even if an exception was thrown!
						reacquireReadLocksAndReleaseWriteLock(holdCount);
					}
					parentFile.setOriginal(false);
					parentFile.setShallowCopy(true); // NEW IN VERSION 0.3
					if (parentFile != null) {
						if (!parentFile.exists()) { // the parent directory will only be shallow-copied (if it does not yet reside in the VFS)
							throw new FileNotFoundException("Folder does not exist: " + parentFile.getAbsolutePath());
						}
					}
					if (ramFile == null)
						initRamFile();
					if (!ramFile.exists() && parentFile != null && parentFile.canWrite()) {
						ramFile.createFile(); // TODO should be atomically to meet specification, however, this is not possible with Commons VFS
						return true;
					} else {
						return false;
					}
					// I removed the catch-blocks at this place since I decided to pass every possible exception onwards anyway
					// Just as a reminder, there may be:
					// 1.) a FileNotFoundException (extends IOException) if the parent folder does not exist
					// 2.) a FileSystemException (extends IOException) caused by the various Commons VFS methods
					// 3.) an IOException that originates from getCanonicalFile()
				}
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return createNewFileOriginal();
	}

	/**
	 * Atomically creates a new, empty file named by this abstract pathname if and only if a file with this name does not yet exist. The check for the
	 * existence of the file and the creation of the file if it does not exist are a single operation that is atomic with respect to all other
	 * filesystem activities that might affect the file.
	 * <P>
	 * Note: this method should <i>not</i> be used for file-locking, as the resulting protocol cannot be made to work reliably. The
	 * {@link java.nio.channels.FileLock FileLock} facility should be used instead.
	 * 
	 * @return <code>true</code> if the named file does not exist and was successfully created; <code>false</code> if the named file already exists
	 * @throws IOException
	 *             If an I/O error occurred
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method denies
	 *             write access to the file
	 * @since 1.2
	 */
	private boolean createNewFileOriginal() throws IOException {
		SecurityManager security = System.getSecurityManager();
		if (security != null)
			security.checkWrite(path);

		// the following quickfix was introduced due to a different method signature of FileSystem.createFileExclusively in the Oracle JDK for Mac
		Method m = null;
		// check for WinNT or UNIX version
		try {
			m = fs.getClass().getMethod("createFileExclusively", String.class);
		} catch (NoSuchMethodException e) {
			// do nothing here (it still could be the Mac version)
		}

		if (m == null) {
			// check for Mac version
			try {
				m = fs.getClass().getMethod("createFileExclusively", String.class, boolean.class);
			} catch (NoSuchMethodException e) {
				System.err
						.println("The used JRE is unknown and uncompatible to the evosuite-io library - please contact the developer of evosuite-io for further help!");
				e.printStackTrace();
			}
		}

		Boolean b = null;
		try {
			b = (Boolean) m.invoke(fs, path);
		} catch (IllegalAccessException e) {
			if (!EvoSuiteIO.tolerantExceptionHandling)
				throw new IllegalStateException(e);
			if (EvoSuiteIO.debug)
				e.printStackTrace();
		} catch (InvocationTargetException e) {
			if (e.getCause() != null) {
				if (e.getCause() instanceof IOException) { // if it is an IOException
					throw (IOException) e.getCause(); // pass it on
				} else {
					if (!EvoSuiteIO.tolerantExceptionHandling)
						throw new IllegalStateException(e.getCause());
					if (EvoSuiteIO.debug)
						e.getCause().printStackTrace();
				}
			} else {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
			}
		}
		return b.booleanValue();
		// return fs.createFileExclusively(path); // does not work with Oracle JDK for Mac
	}

	/**
	 * Deletes the file or directory denoted by this abstract pathname. If this pathname denotes a directory, then the directory must be empty in
	 * order to be deleted. <b>+++ NOTE: Deleting a virtual file will NOT cause it to sychnronize with the real file system at the next access! To
	 * synchronize it, use the newly implemented forceReinitialization() method! +++</b>
	 * 
	 * @return <code>true</code> if and only if the file or directory is successfully deleted; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkDelete}</code> method denies delete access to the file
	 */
	public boolean delete() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					File parent;
					final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
					try {
						EvoSuiteIO.activeVFS = false; // prevents labeling of the parent with EvoSuiteIO.shouldBeOriginal
						parent = getCanonicalFile().getParentFile();
					} finally {
						EvoSuiteIO.activeVFS = true; // assures that VFS is enabled again even if an exception was thrown!
						reacquireReadLocksAndReleaseWriteLock(holdCount);
					}
					parent.setOriginal(false);
					parent.setShallowCopy(true); // introduced in version 0.4
					if (parent == null || !parent.canWrite()) // the parent directory will be shallow copied (if it does not yet reside in the VFS)
						return false;
					if (ramFile.delete()) {
						// forces the vfs to not reinitialize the virtual file with its correspondent original file
						deletedFiles.add(getCanonicalPath());

						// delete permission bitstring
						permissions.remove(getCanonicalPath());

						// ramFile = null; // forces the ramFile to be initialized again (so that possible changes in the real fs are recognized)!
						// it seems more useful if we do NOT have this behaviour; deleting a virtual file should really effect in evaluating
						// virtualFile.exists() to false even if the original file still exists!
						// to reinitialize the ramFile anyway I implemented the new method forceReinitialization()
						return true;
					} else {
						return false;
					}
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return deleteOriginal();
	}

	/**
	 * Deletes the file or directory denoted by this abstract pathname. If this pathname denotes a directory, then the directory must be empty in
	 * order to be deleted.
	 * 
	 * @return <code>true</code> if and only if the file or directory is successfully deleted; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkDelete}</code> method denies delete access to the file
	 */
	private boolean deleteOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkDelete(path);
		}
		return fs.delete(this);
	}

	/**
	 * Requests that the file or directory denoted by this abstract pathname be deleted when the virtual machine terminates. Files (or directories)
	 * are deleted in the reverse order that they are registered. Invoking this method to delete a file or directory that is already registered for
	 * deletion has no effect. Deletion will be attempted only for normal termination of the virtual machine, as defined by the Java Language
	 * Specification.
	 * <p>
	 * Once deletion has been requested, it is not possible to cancel the request. This method should therefore be used with care.
	 * <P>
	 * Note: this method should <i>not</i> be used for file-locking, as the resulting protocol cannot be made to work reliably. The
	 * {@link java.nio.channels.FileLock FileLock} facility should be used instead.
	 * 
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkDelete}</code> method denies delete access to the
	 *             file
	 * @see #delete
	 * @since 1.2
	 */
	public void deleteOnExit() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					/*
					 * Since we store our files in memory, we apparently do not need to delete anything. The allocated memory will be freed, when the
					 * JVM terminates. So we simply do nothing here.
					 */
					return;
				}
			} finally {
				rwlock.readLock().unlock();
			}
		}
		deleteOnExitOriginal();
	}

	/**
	 * Requests that the file or directory denoted by this abstract pathname be deleted when the virtual machine terminates. Files (or directories)
	 * are deleted in the reverse order that they are registered. Invoking this method to delete a file or directory that is already registered for
	 * deletion has no effect. Deletion will be attempted only for normal termination of the virtual machine, as defined by the Java Language
	 * Specification.
	 * <p>
	 * Once deletion has been requested, it is not possible to cancel the request. This method should therefore be used with care.
	 * <P>
	 * Note: this method should <i>not</i> be used for file-locking, as the resulting protocol cannot be made to work reliably. The
	 * {@link java.nio.channels.FileLock FileLock} facility should be used instead.
	 * 
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkDelete}</code> method denies delete access to the
	 *             file
	 * @see #delete
	 * @since 1.2
	 */
	private void deleteOnExitOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkDelete(path);
		}
		DeleteOnExitHook.add(path);
	}

	/**
	 * Returns an array of strings naming the files and directories in the directory denoted by this abstract pathname.
	 * <p>
	 * If this abstract pathname does not denote a directory, then this method returns <code>null</code>. Otherwise an array of strings is returned,
	 * one for each file or directory in the directory. Names denoting the directory itself and the directory's parent directory are not included in
	 * the result. Each string is a file name rather than a complete path.
	 * <p>
	 * There is no guarantee that the name strings in the resulting array will appear in any specific order; they are not, in particular, guaranteed
	 * to appear in alphabetical order.
	 * 
	 * @return An array of strings naming the files and directories in the directory denoted by this abstract pathname. The array will be empty if the
	 *         directory is empty. Returns <code>null</code> if this abstract pathname does not denote a directory, or if an I/O error occurs.
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkRead(java.lang.String)} </code> method denies
	 *             read access to the directory
	 */
	public String[] list() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					// enforce extended shallow copy of the directory
					if (ramFile == null || shallowCopiedDirectories.contains(getCanonicalPath())) {
						setShallowCopy(false);
						initRamFile();
					}
					if ((ramFile.getType() != FileType.FOLDER) || !canRead()) {
						return null; // this is in line with the specification
					}
					FileObject[] children = ramFile.getChildren();
					String[] output = new String[children.length];
					for (int i = 0; i < children.length; i++) {
						output[i] = children[i].getName().getBaseName();
					}
					return output;
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return null;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return listOriginal();
	}

	/**
	 * Returns an array of strings naming the files and directories in the directory denoted by this abstract pathname.
	 * <p>
	 * If this abstract pathname does not denote a directory, then this method returns <code>null</code>. Otherwise an array of strings is returned,
	 * one for each file or directory in the directory. Names denoting the directory itself and the directory's parent directory are not included in
	 * the result. Each string is a file name rather than a complete path.
	 * <p>
	 * There is no guarantee that the name strings in the resulting array will appear in any specific order; they are not, in particular, guaranteed
	 * to appear in alphabetical order.
	 * 
	 * @return An array of strings naming the files and directories in the directory denoted by this abstract pathname. The array will be empty if the
	 *         directory is empty. Returns <code>null</code> if this abstract pathname does not denote a directory, or if an I/O error occurs.
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkRead(java.lang.String)} </code> method denies
	 *             read access to the directory
	 */
	private String[] listOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkRead(path);
		}
		return fs.list(this);
	}

	/**
	 * Returns an array of strings naming the files and directories in the directory denoted by this abstract pathname that satisfy the specified
	 * filter. The behavior of this method is the same as that of the <code>{@link #list()}</code> method, except that the strings in the returned
	 * array must satisfy the filter. If the given <code>filter</code> is <code>null</code> then all names are accepted. Otherwise, a name satisfies
	 * the filter if and only if the value <code>true</code> results when the <code>{@link
	 * FilenameFilter#accept}</code> method of the filter is invoked on this abstract pathname and the name of a file or directory in the directory
	 * that it denotes.
	 * 
	 * @param filter
	 *            A filename filter
	 * @return An array of strings naming the files and directories in the directory denoted by this abstract pathname that were accepted by the given
	 *         <code>filter</code>. The array will be empty if the directory is empty or if no names were accepted by the filter. Returns
	 *         <code>null</code> if this abstract pathname does not denote a directory, or if an I/O error occurs.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the directory
	 */
	public String[] list(FilenameFilter filter) { // should also work with simulation
		String names[] = list();
		if ((names == null) || (filter == null)) {
			return names;
		}
		// ArrayList instances were already unparameterized in original code! - I leave it unchanged
		ArrayList v = new ArrayList();
		for (int i = 0; i < names.length; i++) {
			if (filter.accept(this, names[i])) {
				v.add(names[i]);
			}
		}
		return (String[]) (v.toArray(new String[v.size()]));
	}

	/**
	 * Returns an array of abstract pathnames denoting the files in the directory denoted by this abstract pathname.
	 * <p>
	 * If this abstract pathname does not denote a directory, then this method returns <code>null</code>. Otherwise an array of <code>File</code>
	 * objects is returned, one for each file or directory in the directory. Pathnames denoting the directory itself and the directory's parent
	 * directory are not included in the result. Each resulting abstract pathname is constructed from this abstract pathname using the
	 * <code>{@link #File(java.io.File, java.lang.String)
	 * File(File,&nbsp;String)}</code> constructor. Therefore if this pathname is absolute then each resulting pathname is absolute; if this pathname
	 * is relative then each resulting pathname will be relative to the same directory.
	 * <p>
	 * There is no guarantee that the name strings in the resulting array will appear in any specific order; they are not, in particular, guaranteed
	 * to appear in alphabetical order.
	 * 
	 * @return An array of abstract pathnames denoting the files and directories in the directory denoted by this abstract pathname. The array will be
	 *         empty if the directory is empty. Returns <code>null</code> if this abstract pathname does not denote a directory, or if an I/O error
	 *         occurs.
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkRead(java.lang.String)} </code> method denies
	 *             read access to the directory
	 * @since 1.2
	 */
	public File[] listFiles() { // should also work with simulation
		String[] ss = list();
		if (ss == null)
			return null;
		int n = ss.length;
		File[] fs = new File[n];
		for (int i = 0; i < n; i++) {
			fs[i] = new File(ss[i], this);
		}
		return fs;
	}

	/**
	 * Returns an array of abstract pathnames denoting the files and directories in the directory denoted by this abstract pathname that satisfy the
	 * specified filter. The behavior of this method is the same as that of the <code>{@link #listFiles()}</code> method, except that the pathnames in
	 * the returned array must satisfy the filter. If the given <code>filter</code> is <code>null</code> then all pathnames are accepted. Otherwise, a
	 * pathname satisfies the filter if and only if the value <code>true</code> results when the <code>{@link FilenameFilter#accept}</code> method of
	 * the filter is invoked on this abstract pathname and the name of a file or directory in the directory that it denotes.
	 * 
	 * @param filter
	 *            A filename filter
	 * @return An array of abstract pathnames denoting the files and directories in the directory denoted by this abstract pathname. The array will be
	 *         empty if the directory is empty. Returns <code>null</code> if this abstract pathname does not denote a directory, or if an I/O error
	 *         occurs.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the directory
	 * @since 1.2
	 */
	public File[] listFiles(FilenameFilter filter) { // unchanged - should also work with simulation
		String ss[] = list();
		if (ss == null)
			return null;
		ArrayList v = new ArrayList();
		for (int i = 0; i < ss.length; i++) {
			if ((filter == null) || filter.accept(this, ss[i])) {
				v.add(new File(ss[i], this));
			}
		}
		return (File[]) (v.toArray(new File[v.size()]));
	}

	/**
	 * Returns an array of abstract pathnames denoting the files and directories in the directory denoted by this abstract pathname that satisfy the
	 * specified filter. The behavior of this method is the same as that of the <code>{@link #listFiles()}</code> method, except that the pathnames in
	 * the returned array must satisfy the filter. If the given <code>filter</code> is <code>null</code> then all pathnames are accepted. Otherwise, a
	 * pathname satisfies the filter if and only if the value <code>true</code> results when the <code>{@link FileFilter#accept(java.io.File)}</code>
	 * method of the filter is invoked on the pathname.
	 * 
	 * @param filter
	 *            A file filter
	 * @return An array of abstract pathnames denoting the files and directories in the directory denoted by this abstract pathname. The array will be
	 *         empty if the directory is empty. Returns <code>null</code> if this abstract pathname does not denote a directory, or if an I/O error
	 *         occurs.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to the directory
	 * @since 1.2
	 */
	public File[] listFiles(FileFilter filter) { // unchanged - should also work with simulation
		String ss[] = list();
		if (ss == null)
			return null;
		ArrayList v = new ArrayList();
		for (int i = 0; i < ss.length; i++) {
			File f = new File(ss[i], this);
			if ((filter == null) || filter.accept(f)) {
				v.add(f);
			}
		}
		return (File[]) (v.toArray(new File[v.size()]));
	}

	/**
	 * Creates the directory named by this abstract pathname.
	 * 
	 * @return <code>true</code> if and only if the directory was created; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method does not permit the named directory to be created
	 */
	public boolean mkdir() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					if (ramFile.exists())
						return false;
					File parent;
					final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
					try {
						EvoSuiteIO.activeVFS = false; // prevents labeling of the parent with EvoSuiteIO.shouldBeOriginal
						parent = getCanonicalFile().getParentFile();
					} finally {
						EvoSuiteIO.activeVFS = true; // assures that VFS is enabled again even if an exception was thrown!
						reacquireReadLocksAndReleaseWriteLock(holdCount);
					}
					parent.setOriginal(false);
					parent.setShallowCopy(true); // introduced in version 0.4
					if (parent == null || !parent.exists() || !parent.canWrite()) // parent will only be shallow copied (if it does not yet reside in
																					// VFS)
						return false; // mkdir() does not create any parent directories! if they do not exist, mkdir() fails!
					try {
						ramFile.createFolder(); // TODO this should actually be performed atomically, however, I cannot assure that with Commons VFS
					} catch (FileSystemException e) {
						// ramFile.createFolder may throw a FileSystemException in cases the original implementation does not, so we fail 'silently'
						return false;
					}
					return true;
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return mkdirOriginal();
	}

	/**
	 * Creates the directory named by this abstract pathname.
	 * 
	 * @return <code>true</code> if and only if the directory was created; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method does not permit the named directory to be created
	 */
	private boolean mkdirOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkWrite(path);
		}
		return fs.createDirectory(this);
	}

	/**
	 * Creates the directory named by this abstract pathname, including any necessary but nonexistent parent directories. Note that if this operation
	 * fails it may have succeeded in creating some of the necessary parent directories.
	 * 
	 * @return <code>true</code> if and only if the directory was created, along with all necessary parent directories; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code> method does not permit verification of the existence of the named
	 *             directory and all necessary parent directories; or if the <code>{@link 
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method does not permit the named directory and all necessary parent
	 *             directories to be created
	 */
	public boolean mkdirs() { // uses mkdir() - should therefore also work with VFS
		if (exists()) {
			return false;
		}
		if (mkdir()) {
			return true;
		}
		File canonFile = null;
		try {
			canonFile = getCanonicalFile();
			if (!isOriginal) {
				rwlock.readLock().lock();
				try {
					if (EvoSuiteIO.activeVFS) {
						canonFile.setOriginal(false); // new
					}
				} finally {
					rwlock.readLock().unlock();
				}
			}
		} catch (IOException e) {
			return false;
		}

		File parent = canonFile.getParentFile();
		if (!isOriginal) { // new
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					parent.setOriginal(false); // new
					parent.setShallowCopy(true); // new in version 0.4
				}
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return (parent != null && (parent.mkdirs() || parent.exists()) && canonFile.mkdir());
	}

	// public boolean mkdirs() { // alternative implementation with a Commons VFS method
	// if (!isOriginal && EvoSuiteIO.activeVFS) {
	// try {
	// if (ramFile == null)
	// initRamFile();
	//
	// if (!ramFile.exists()) { // this should actually be performed atomically, however, I cannot assure that with Commons VFS
	// ramFile.createFolder(); // creates also all ancestor folders
	// return true;
	// } else {
	// return false;
	// }
	// } catch (IOException e) {
	// if (e instanceof FileSystemException) {
	// if (debug) {
	// e.printStackTrace();
	// }
	// } else {
	// e.printStackTrace();
	// }
	// }
	// } else {
	// return mkdirsOriginal();
	// }
	// return false;
	// }

	/**
	 * <b>+++ CAUTION: WHEN USED DURING ACTIVE VFS, THIS METHOD MAY HAVE VERY BAD PERFORMANCE SINCE THIS FILE/DIRECTORY MAY BE DEEP COPIED FROM DISK
	 * INTO MEMORY! (regardless of its size) +++</b>
	 * <p>
	 * Renames the file denoted by this abstract pathname.
	 * <p>
	 * Many aspects of the behavior of this method are inherently platform-dependent: The rename operation might not be able to move a file from one
	 * filesystem to another, it might not be atomic, and it might not succeed if a file with the destination abstract pathname already exists. The
	 * return value should always be checked to make sure that the rename operation was successful. <b>NOTE: Virtual files will be treated with the
	 * semantics of Commons VFS which in turn is similiar to Unix behaviour. That means, the rename operation will:
	 * <ul>
	 * <li>NOT be able to move a file from one filesystem to another - like in Unix</li>
	 * <li>be NOT atomic</li>
	 * <li>succeed if a file with the destination abstract pathname already exists (which in turn will be deleted) - like in Unix</li>
	 * </ul>
	 * </b>
	 * 
	 * @param dest
	 *            The new abstract pathname for the named file
	 * @return <code>true</code> if and only if the renaming succeeded; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method denies
	 *             write access to either the old or new pathnames
	 * @throws NullPointerException
	 *             If parameter <code>dest</code> is <code>null</code>
	 */
	public boolean renameTo(File dest) {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (dest == null)
						throw new NullPointerException("parameter 'dest' must not be null!"); // do this to stay in line with specification
					if (!exists() || !canRead() || !canWrite())
						return false;
					// permission to delete this file?
					File parent;
					final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
					try {
						EvoSuiteIO.activeVFS = false; // prevents labeling of the parent with EvoSuiteIO.shouldBeOriginal
						parent = getCanonicalFile().getParentFile();
					} finally {
						EvoSuiteIO.activeVFS = true; // assures that VFS is enabled again even if an exception was thrown!
						reacquireReadLocksAndReleaseWriteLock(holdCount);
					}
					parent.setOriginal(false);
					parent.setShallowCopy(true); // introduced in version 0.4
					if (parent == null || !parent.canWrite()) // parent directory will be shallow copied (if it does not yet reside in the VFS)
						return false;
					// permission to create files in the destination folder?
					File destParent = dest.getCanonicalFile().getParentFile();
					boolean fail = false;
					while (destParent != null) {
						destParent.setOriginal(false);
						destParent.setShallowCopy(true); // introduced in version 0.4
						if (destParent.exists() && !destParent.canWrite()) { // parent directory will be shallow copied (if it does not yet reside in
																				// the VFS)
							fail = true;
							break;
						}
						destParent = destParent.getCanonicalFile().getParentFile();
					}
					if (fail)
						return false;
					if (ramFile.canRenameTo(dest.getInitializedRamFile())) { // this just checks if dest is on another filesystem!
						long lastModified = lastModified(); // memorize lastModified of file to be renamed

						// TODO fix performance issue for this (won't be simple)
						if (!forceDeepCopy())
							return false;

						/*
						 * deletes the destination ramFile and all its possible descendants - this has to be done since FileObject.moveTo is buggy
						 * (throws an FileSystemException) although it shouldn't if dest exists
						 */
						String destCanonicalPath = dest.getCanonicalPath();
						if (dest.getInitializedRamFile().exists()) {
							deletedFiles.addAll(dest.getCanonicalPathsOfThisAndAllDescendantsInVFS());
							dest.getInitializedRamFile().delete(Selectors.SELECT_ALL);
							filterCanonicalPath(permissions.keySet(), destCanonicalPath);
							// filterCanonicalPath(shallowCopiedDirectories, destCanonicalPath); // not sure if this is necessary
							// filterCanonicalPath(shallowCopiedFiles, destCanonicalPath); // not sure if this is necessary
						}

						// deep copy to destination - FileObject.moveTo is apparently too stupid to do it itself >:(
						dest.getInitializedRamFile().copyFrom(getInitializedRamFile(), Selectors.SELECT_ALL);

						// ramFile.moveTo(dest.getRamFile()); // sorry, but this Commons VFS method is a mess

						dest.setLastModified(lastModified); // lastModified is not copied reliably by FileObject.moveTo so we have to do it ourselves
						// TODO also do this for possible children?

						// move access permissions (also for descendants)
						String canonicalPath = getCanonicalPath();
						for (String srcPath : getMatchesForCanonicalPath(permissions.keySet(), canonicalPath)) {
							String destPath = srcPath.replace(canonicalPath, destCanonicalPath);
							permissions.put(destPath, permissions.get(srcPath));
							permissions.remove(srcPath);
						}
						// not sure if shallowCopyDirectories and shallowCopiedFiles entries should be copied either

						// deep deletion of the source path (also adds all deleted files to the deletedFiles set)
						deepDelete();

						return true;
					} else {
						return false;
					}
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return renameToOriginal(dest);
	}

	/**
	 * Renames the file denoted by this abstract pathname.
	 * <p>
	 * Many aspects of the behavior of this method are inherently platform-dependent: The rename operation might not be able to move a file from one
	 * filesystem to another, it might not be atomic, and it might not succeed if a file with the destination abstract pathname already exists. The
	 * return value should always be checked to make sure that the rename operation was successful.
	 * 
	 * @param dest
	 *            The new abstract pathname for the named file
	 * @return <code>true</code> if and only if the renaming succeeded; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method denies
	 *             write access to either the old or new pathnames
	 * @throws NullPointerException
	 *             If parameter <code>dest</code> is <code>null</code>
	 */
	private boolean renameToOriginal(File dest) {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkWrite(path);
			security.checkWrite(dest.path);
		}
		return fs.rename(this, dest);
	}

	/**
	 * Sets the last-modified time of the file or directory named by this abstract pathname.
	 * <p>
	 * All platforms support file-modification times to the nearest second, but some provide more precision. The argument will be truncated to fit the
	 * supported precision. If the operation succeeds and no intervening operations on the file take place, then the next invocation of the
	 * <code>{@link #lastModified}</code> method will return the (possibly truncated) <code>time</code> argument that was passed to this method.
	 * 
	 * @param time
	 *            The new last-modified time, measured in milliseconds since the epoch (00:00:00 GMT, January 1, 1970)
	 * @return <code>true</code> if and only if the operation succeeded; <code>false</code> otherwise
	 * @throws IllegalArgumentException
	 *             If the argument is negative
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method denies
	 *             write access to the named file
	 * @since 1.2
	 */
	public boolean setLastModified(long time) {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (time < 0)
						throw new IllegalArgumentException("Negative time"); // must not be catched here!
					if (ramFile == null)
						initRamFile();
					if (!ramFile.exists() || !canWrite())
						return false;
					try {
						ramFile.getContent().setLastModifiedTime(time); // creates the file, if it does not exist! -> it should never come so far!
					} catch (FileSystemException e) {
						// ramFile.getContent().setLastModifiedTime(time) may throw some FileSystemExceptions in conditions where the original method
						// does not cause IOExceptions -> hence, they should be thrown 'silently' and just 'false' should be returned
						return false;
					}
					return true;
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return setLastModifiedOriginal(time);
	}

	/**
	 * Sets the last-modified time of the file or directory named by this abstract pathname.
	 * <p>
	 * All platforms support file-modification times to the nearest second, but some provide more precision. The argument will be truncated to fit the
	 * supported precision. If the operation succeeds and no intervening operations on the file take place, then the next invocation of the
	 * <code>{@link #lastModified}</code> method will return the (possibly truncated) <code>time</code> argument that was passed to this method.
	 * 
	 * @param time
	 *            The new last-modified time, measured in milliseconds since the epoch (00:00:00 GMT, January 1, 1970)
	 * @return <code>true</code> if and only if the operation succeeded; <code>false</code> otherwise
	 * @throws IllegalArgumentException
	 *             If the argument is negative
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method denies
	 *             write access to the named file
	 * @since 1.2
	 */
	private boolean setLastModifiedOriginal(long time) {
		if (time < 0)
			throw new IllegalArgumentException("Negative time");
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkWrite(path);
		}
		return fs.setLastModifiedTime(this, time);
	}

	/**
	 * Marks the file or directory named by this abstract pathname so that only read operations are allowed. After invoking this method the file or
	 * directory is guaranteed not to change until it is either deleted or marked to allow write access. Whether or not a read-only file or directory
	 * may be deleted depends upon the underlying system.
	 * 
	 * @return <code>true</code> if and only if the operation succeeded; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method denies write access to the named file
	 * @since 1.2
	 */
	public boolean setReadOnly() {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					return setWritable(false, false);
				}
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return setReadOnlyOriginal();
	}

	/**
	 * Marks the file or directory named by this abstract pathname so that only read operations are allowed. After invoking this method the file or
	 * directory is guaranteed not to change until it is either deleted or marked to allow write access. Whether or not a read-only file or directory
	 * may be deleted depends upon the underlying system.
	 * 
	 * @return <code>true</code> if and only if the operation succeeded; <code>false</code> otherwise
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method denies write access to the named file
	 * @since 1.2
	 */
	private boolean setReadOnlyOriginal() {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkWrite(path);
		}
		return fs.setReadOnly(this);
	}

	/*
	 * the setReadable(), setWritable(), setExecutable() methods differ in their behaviour on different file systems; I will simulate the behaviour of
	 * UNIX file systems; see http://java.sun.com/developer/technicalArticles/J2SE/Desktop/javase6/enhancements/ for a more precise explanation on the
	 * difference besides that, my implementation of access permissions for the VFS does currently not support multi-user mode; therefore the
	 * 'ownerOnly' flag is ignored in these methods
	 */

	/**
	 * Sets the owner's or everybody's write permission for this abstract pathname.
	 * 
	 * @param writable
	 *            If <code>true</code>, sets the access permission to allow write operations; if <code>false</code> to disallow write operations
	 * @param ownerOnly
	 *            If <code>true</code>, the write permission applies only to the owner's write permission; otherwise, it applies to everybody. If the
	 *            underlying file system can not distinguish the owner's write permission from that of others, then the permission will apply to
	 *            everybody, regardless of this value.
	 * @return <code>true</code> if and only if the operation succeeded. The operation will fail if the user does not have permission to change the
	 *         access permissions of this abstract pathname.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method denies write access to the named file
	 * @since 1.6
	 */
	public boolean setWritable(boolean writable, boolean ownerOnly) {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					if (!ramFile.exists())
						return false;
					String canonicalPath = getCanonicalPath();
					int permissionBitstring = getDefaultPermissionBitstring();
					if (permissions.containsKey(canonicalPath)) {
						permissionBitstring = permissions.get(canonicalPath).intValue();
					}
					if (writable) { // we do not care about the ownerOnly flag atm due to missing support for multi-user access permissions
						permissionBitstring = permissionBitstring | FileSystem.ACCESS_WRITE; // set the write bit
					} else {
						permissionBitstring = permissionBitstring & ~FileSystem.ACCESS_WRITE; // unset the write bit - in Windows same effect as:
																								// setReadOnly()
					}
					permissions.put(canonicalPath, permissionBitstring); // store the permission
					return true; // operation successful
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return setWritableOriginal(writable, ownerOnly);
	}

	/**
	 * Sets the owner's or everybody's write permission for this abstract pathname.
	 * 
	 * @param writable
	 *            If <code>true</code>, sets the access permission to allow write operations; if <code>false</code> to disallow write operations
	 * @param ownerOnly
	 *            If <code>true</code>, the write permission applies only to the owner's write permission; otherwise, it applies to everybody. If the
	 *            underlying file system can not distinguish the owner's write permission from that of others, then the permission will apply to
	 *            everybody, regardless of this value.
	 * @return <code>true</code> if and only if the operation succeeded. The operation will fail if the user does not have permission to change the
	 *         access permissions of this abstract pathname.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method denies write access to the named file
	 * @since 1.6
	 */
	private boolean setWritableOriginal(boolean writable, boolean ownerOnly) {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkWrite(path);
		}
		return fs.setPermission(this, FileSystem.ACCESS_WRITE, writable, ownerOnly);
	}

	/**
	 * A convenience method to set the owner's write permission for this abstract pathname.
	 * <p>
	 * An invocation of this method of the form <tt>file.setWritable(arg)</tt> behaves in exactly the same way as the invocation
	 * 
	 * <pre>
	 * file.setWritable(arg, true)
	 * </pre>
	 * 
	 * @param writable
	 *            If <code>true</code>, sets the access permission to allow write operations; if <code>false</code> to disallow write operations
	 * @return <code>true</code> if and only if the operation succeeded. The operation will fail if the user does not have permission to change the
	 *         access permissions of this abstract pathname.
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method denies
	 *             write access to the file
	 * @since 1.6
	 */
	public boolean setWritable(boolean writable) {
		return setWritable(writable, true);
	}

	/**
	 * Sets the owner's or everybody's read permission for this abstract pathname.
	 * 
	 * @param readable
	 *            If <code>true</code>, sets the access permission to allow read operations; if <code>false</code> to disallow read operations
	 * @param ownerOnly
	 *            If <code>true</code>, the read permission applies only to the owner's read permission; otherwise, it applies to everybody. If the
	 *            underlying file system can not distinguish the owner's read permission from that of others, then the permission will apply to
	 *            everybody, regardless of this value.
	 * @return <code>true</code> if and only if the operation succeeded. The operation will fail if the user does not have permission to change the
	 *         access permissions of this abstract pathname. If <code>readable</code> is <code>false</code> and the underlying file system does not
	 *         implement a read permission, then the operation will fail. <b>+++ NOTICE: The VFS will simulate UNIX behaviour! +++</b>
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method denies write access to the file
	 * @since 1.6
	 */
	public boolean setReadable(boolean readable, boolean ownerOnly) {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					if (!ramFile.exists())
						return false;
					String canonicalPath = getCanonicalPath();
					int permissionBitstring = getDefaultPermissionBitstring();
					if (permissions.containsKey(canonicalPath)) {
						permissionBitstring = permissions.get(canonicalPath).intValue();
					}
					if (readable) { // we do not care about the ownerOnly flag atm due to missing support for multi-user access permissions
						permissionBitstring = permissionBitstring | FileSystem.ACCESS_READ; // set the read bit
					} else {
						permissionBitstring = permissionBitstring & ~FileSystem.ACCESS_READ; // unset the read bit - here a difference to the real fs
																								// is possible: in the real fs setReadable(false) may
																								// fail (e.g. in Windows)!
					}
					permissions.put(canonicalPath, permissionBitstring); // store the permission
					return true; // operation successful
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return setReadableOriginal(readable, ownerOnly);
	}

	/**
	 * Sets the owner's or everybody's read permission for this abstract pathname.
	 * 
	 * @param readable
	 *            If <code>true</code>, sets the access permission to allow read operations; if <code>false</code> to disallow read operations
	 * @param ownerOnly
	 *            If <code>true</code>, the read permission applies only to the owner's read permission; otherwise, it applies to everybody. If the
	 *            underlying file system can not distinguish the owner's read permission from that of others, then the permission will apply to
	 *            everybody, regardless of this value.
	 * @return <code>true</code> if and only if the operation succeeded. The operation will fail if the user does not have permission to change the
	 *         access permissions of this abstract pathname. If <code>readable</code> is <code>false</code> and the underlying file system does not
	 *         implement a read permission, then the operation will fail.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method denies write access to the file
	 * @since 1.6
	 */
	private boolean setReadableOriginal(boolean readable, boolean ownerOnly) {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkWrite(path);
		}
		return fs.setPermission(this, FileSystem.ACCESS_READ, readable, ownerOnly);
	}

	/**
	 * A convenience method to set the owner's read permission for this abstract pathname.
	 * <p>
	 * An invocation of this method of the form <tt>file.setReadable(arg)</tt> behaves in exactly the same way as the invocation
	 * 
	 * <pre>
	 * file.setReadable(arg, true)
	 * </pre>
	 * 
	 * @param readable
	 *            If <code>true</code>, sets the access permission to allow read operations; if <code>false</code> to disallow read operations
	 * @return <code>true</code> if and only if the operation succeeded. The operation will fail if the user does not have permission to change the
	 *         access permissions of this abstract pathname. If <code>readable</code> is <code>false</code> and the underlying file system does not
	 *         implement a read permission, then the operation will fail.
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method denies
	 *             write access to the file
	 * @since 1.6
	 */
	public boolean setReadable(boolean readable) {
		return setReadable(readable, true);
	}

	/**
	 * Sets the owner's or everybody's execute permission for this abstract pathname.
	 * 
	 * @param executable
	 *            If <code>true</code>, sets the access permission to allow execute operations; if <code>false</code> to disallow execute operations
	 * @param ownerOnly
	 *            If <code>true</code>, the execute permission applies only to the owner's execute permission; otherwise, it applies to everybody. If
	 *            the underlying file system can not distinguish the owner's execute permission from that of others, then the permission will apply to
	 *            everybody, regardless of this value.
	 * @return <code>true</code> if and only if the operation succeeded. The operation will fail if the user does not have permission to change the
	 *         access permissions of this abstract pathname. If <code>executable</code> is <code>false</code> and the underlying file system does not
	 *         implement an execute permission, then the operation will fail. <b>+++ NOTICE: The VFS will simulate UNIX behaviour! +++</b>
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method denies write access to the file
	 * @since 1.6
	 */
	public boolean setExecutable(boolean executable, boolean ownerOnly) {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null)
						initRamFile();
					if (!ramFile.exists())
						return false;
					String canonicalPath = getCanonicalPath();
					int permissionBitstring = getDefaultPermissionBitstring();
					if (permissions.containsKey(canonicalPath)) {
						permissionBitstring = permissions.get(canonicalPath).intValue();
					}
					if (executable) { // we do not care about the ownerOnly flag atm due to missing support for multi-user access permissions
						permissionBitstring = permissionBitstring | FileSystem.ACCESS_EXECUTE; // set the execute bit
					} else {
						permissionBitstring = permissionBitstring & ~FileSystem.ACCESS_EXECUTE; // unset the execute bit - here a difference to the
																								// real fs is possible: e.g. in Windows
																								// setExecutable(false) is not possible and will fail!
						// TODO may-have: simulate different behaviours of different platforms?
					}
					permissions.put(canonicalPath, permissionBitstring);
					return true;
				}
			} catch (IOException e) {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
				return false;
			} finally {
				rwlock.readLock().unlock();
			}
		}
		return setExecutableOriginal(executable, ownerOnly);
	}

	/**
	 * Sets the owner's or everybody's execute permission for this abstract pathname.
	 * 
	 * @param executable
	 *            If <code>true</code>, sets the access permission to allow execute operations; if <code>false</code> to disallow execute operations
	 * @param ownerOnly
	 *            If <code>true</code>, the execute permission applies only to the owner's execute permission; otherwise, it applies to everybody. If
	 *            the underlying file system can not distinguish the owner's execute permission from that of others, then the permission will apply to
	 *            everybody, regardless of this value.
	 * @return <code>true</code> if and only if the operation succeeded. The operation will fail if the user does not have permission to change the
	 *         access permissions of this abstract pathname. If <code>executable</code> is <code>false</code> and the underlying file system does not
	 *         implement an execute permission, then the operation will fail.
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method denies write access to the file
	 * @since 1.6
	 */
	private boolean setExecutableOriginal(boolean executable, boolean ownerOnly) {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkWrite(path);
		}
		return fs.setPermission(this, FileSystem.ACCESS_EXECUTE, executable, ownerOnly);
	}

	/**
	 * A convenience method to set the owner's execute permission for this abstract pathname.
	 * <p>
	 * An invocation of this method of the form <tt>file.setExcutable(arg)</tt> behaves in exactly the same way as the invocation
	 * 
	 * <pre>
	 * file.setExecutable(arg, true)
	 * </pre>
	 * 
	 * @param executable
	 *            If <code>true</code>, sets the access permission to allow execute operations; if <code>false</code> to disallow execute operations
	 * @return <code>true</code> if and only if the operation succeeded. The operation will fail if the user does not have permission to change the
	 *         access permissions of this abstract pathname. If <code>executable</code> is <code>false</code> and the underlying file system does not
	 *         implement an excute permission, then the operation will fail.
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method denies
	 *             write access to the file
	 * @since 1.6
	 */
	public boolean setExecutable(boolean executable) {
		return setExecutable(executable, true);
	}

	/* -- Filesystem interface -- */

	/**
	 * List the available filesystem roots.
	 * <p>
	 * A particular Java platform may support zero or more hierarchically-organized file systems. Each file system has a <code>root</code> directory
	 * from which all other files in that file system can be reached. Windows platforms, for example, have a root directory for each active drive;
	 * UNIX platforms have a single root directory, namely <code>"/"</code>. The set of available filesystem roots is affected by various system-level
	 * operations such as the insertion or ejection of removable media and the disconnecting or unmounting of physical or virtual disk drives.
	 * <p>
	 * This method returns an array of <code>File</code> objects that denote the root directories of the available filesystem roots. It is guaranteed
	 * that the canonical pathname of any file physically present on the local machine will begin with one of the roots returned by this method.
	 * <p>
	 * The canonical pathname of a file that resides on some other machine and is accessed via a remote-filesystem protocol such as SMB or NFS may or
	 * may not begin with one of the roots returned by this method. If the pathname of a remote file is syntactically indistinguishable from the
	 * pathname of a local file then it will begin with one of the roots returned by this method. Thus, for example, <code>File</code> objects
	 * denoting the root directories of the mapped network drives of a Windows platform will be returned by this method, while <code>File</code>
	 * objects containing UNC pathnames will not be returned by this method.
	 * <p>
	 * Unlike most methods in this class, this method does not throw security exceptions. If a security manager exists and its <code>{@link
	 * java.lang.SecurityManager#checkRead(java.lang.String)}</code> method denies read access to a particular root directory, then that directory
	 * will not appear in the result.
	 * 
	 * @return An array of <code>File</code> objects denoting the available filesystem roots, or <code>null</code> if the set of roots could not be
	 *         determined. The array will be empty if there are no filesystem roots.
	 * @since 1.2
	 */
	public static File[] listRoots() { // very hard to simulate and should not cause problems - so I leave it unchanged
		return fs.listRoots();
	}

	/* -- Disk usage -- */

	/**
	 * Returns the size of the partition <a href="#partName">named</a> by this abstract pathname.
	 * 
	 * @return The size, in bytes, of the partition or <tt>0L</tt> if this abstract pathname does not name a partition
	 * @throws SecurityException
	 *             If a security manager has been installed and it denies {@link RuntimePermission}<tt>("getFileSystemAttributes")</tt> or its
	 *             {@link SecurityManager#checkRead(String)} method denies read access to the file named by this abstract pathname
	 * @since 1.6
	 */
	public long getTotalSpace() { // very hard to simulate and should not cause problems - so I leave it unchanged
		SecurityManager sm = System.getSecurityManager();
		if (sm != null) {
			sm.checkPermission(new RuntimePermission("getFileSystemAttributes"));
			sm.checkRead(path);
		}
		return fs.getSpace(this, FileSystem.SPACE_TOTAL);
	}

	/**
	 * Returns the number of unallocated bytes in the partition <a href="#partName">named</a> by this abstract path name.
	 * <p>
	 * The returned number of unallocated bytes is a hint, but not a guarantee, that it is possible to use most or any of these bytes. The number of
	 * unallocated bytes is most likely to be accurate immediately after this call. It is likely to be made inaccurate by any external I/O operations
	 * including those made on the system outside of this virtual machine. This method makes no guarantee that write operations to this file system
	 * will succeed.
	 * 
	 * @return The number of unallocated bytes on the partition <tt>0L</tt> if the abstract pathname does not name a partition. This value will be
	 *         less than or equal to the total file system size returned by {@link #getTotalSpace}.
	 * @throws SecurityException
	 *             If a security manager has been installed and it denies {@link RuntimePermission}<tt>("getFileSystemAttributes")</tt> or its
	 *             {@link SecurityManager#checkRead(String)} method denies read access to the file named by this abstract pathname
	 * @since 1.6
	 */
	public long getFreeSpace() { // very hard to simulate and should not cause problems - so I leave it unchanged
		SecurityManager sm = System.getSecurityManager();
		if (sm != null) {
			sm.checkPermission(new RuntimePermission("getFileSystemAttributes"));
			sm.checkRead(path);
		}
		return fs.getSpace(this, FileSystem.SPACE_FREE);
	}

	/**
	 * Returns the number of bytes available to this virtual machine on the partition <a href="#partName">named</a> by this abstract pathname. When
	 * possible, this method checks for write permissions and other operating system restrictions and will therefore usually provide a more accurate
	 * estimate of how much new data can actually be written than {@link #getFreeSpace}.
	 * <p>
	 * The returned number of available bytes is a hint, but not a guarantee, that it is possible to use most or any of these bytes. The number of
	 * unallocated bytes is most likely to be accurate immediately after this call. It is likely to be made inaccurate by any external I/O operations
	 * including those made on the system outside of this virtual machine. This method makes no guarantee that write operations to this file system
	 * will succeed.
	 * 
	 * @return The number of available bytes on the partition or <tt>0L</tt> if the abstract pathname does not name a partition. On systems where this
	 *         information is not available, this method will be equivalent to a call to {@link #getFreeSpace}.
	 * @throws SecurityException
	 *             If a security manager has been installed and it denies {@link RuntimePermission}<tt>("getFileSystemAttributes")</tt> or its
	 *             {@link SecurityManager#checkRead(String)} method denies read access to the file named by this abstract pathname
	 * @since 1.6
	 */
	public long getUsableSpace() { // very hard to simulate and should not cause problems - so I leave it unchanged
		SecurityManager sm = System.getSecurityManager();
		if (sm != null) {
			sm.checkPermission(new RuntimePermission("getFileSystemAttributes"));
			sm.checkRead(path);
		}
		return fs.getSpace(this, FileSystem.SPACE_USABLE);
	}

	/* -- Temporary files -- */

	// lazy initialization of SecureRandom and temporary file directory
	private static class LazyInitialization {
		static final SecureRandom random = new SecureRandom();

		static final String temporaryDirectory = temporaryDirectory();

		static String temporaryDirectory() {
			return fs.normalize(AccessController.doPrivileged(new GetPropertyAction("java.io.tmpdir")));
		}
	}

	private static File generateFile(String prefix, String suffix, File dir) // should cause no problems
			throws IOException {
		long n = LazyInitialization.random.nextLong();
		if (n == Long.MIN_VALUE) {
			n = 0; // corner case
		} else {
			n = Math.abs(n);
		}
		return new File(dir, prefix + Long.toString(n) + suffix);
	}

	private static boolean checkAndCreate(String filename, SecurityManager sm) // only used by createTempFileOriginal()
			throws IOException {
		if (sm != null) {
			try {
				sm.checkWrite(filename);
			} catch (AccessControlException x) {
				/*
				 * Throwing the original AccessControlException could disclose the location of the default temporary directory, so we re-throw a more
				 * innocuous SecurityException
				 */
				throw new SecurityException("Unable to create temporary file");
			}
		}

		// the following quickfix was introduced due to a different method signature of FileSystem.createFileExclusively in the Oracle JDK for Mac
		Method m = null;
		// check for WinNT or UNIX version
		try {
			m = fs.getClass().getMethod("createFileExclusively", String.class);
		} catch (NoSuchMethodException e) {
			// do nothing here - it still could be the Mac version
		}

		if (m == null) {
			// check for Mac version
			try {
				m = fs.getClass().getMethod("createFileExclusively", String.class, boolean.class);
			} catch (NoSuchMethodException e) {
				System.err
						.println("The used JRE is unknown and uncompatible to the evosuite-io library - please contact the developer of evosuite-io for further help!");
				e.printStackTrace();
			}
		}

		Boolean b = null;
		try {
			b = (Boolean) m.invoke(fs, filename);
		} catch (IllegalAccessException e) {
			if (!EvoSuiteIO.tolerantExceptionHandling)
				throw new IllegalStateException(e);
			if (EvoSuiteIO.debug)
				e.printStackTrace();
		} catch (InvocationTargetException e) {
			if (e.getCause() != null) {
				if (e.getCause() instanceof IOException) { // if it is an IOException
					throw (IOException) e.getCause(); // pass it on
				} else {
					if (!EvoSuiteIO.tolerantExceptionHandling)
						throw new IllegalStateException(e.getCause());
					if (EvoSuiteIO.debug)
						e.getCause().printStackTrace();
				}
			} else {
				if (!EvoSuiteIO.tolerantExceptionHandling)
					throw new IllegalStateException(e);
				if (EvoSuiteIO.debug)
					e.printStackTrace();
			}
		}
		return b.booleanValue();
		// return fs.createFileExclusively(filename); // does not work with Oracle JDK for Mac
	}

	/**
	 * <p>
	 * Creates a new empty file in the specified directory, using the given prefix and suffix strings to generate its name. If this method returns
	 * successfully then it is guaranteed that:
	 * <ol>
	 * <li>The file denoted by the returned abstract pathname did not exist before this method was invoked, and
	 * <li>Neither this method nor any of its variants will return the same abstract pathname again in the current invocation of the virtual machine.
	 * </ol>
	 * This method provides only part of a temporary-file facility. To arrange for a file created by this method to be deleted automatically, use the
	 * <code>{@link #deleteOnExit}</code> method.
	 * <p>
	 * The <code>prefix</code> argument must be at least three characters long. It is recommended that the prefix be a short, meaningful string such
	 * as <code>"hjb"</code> or <code>"mail"</code>. The <code>suffix</code> argument may be <code>null</code>, in which case the suffix
	 * <code>".tmp"</code> will be used.
	 * <p>
	 * To create the new file, the prefix and the suffix may first be adjusted to fit the limitations of the underlying platform. If the prefix is too
	 * long then it will be truncated, but its first three characters will always be preserved. If the suffix is too long then it too will be
	 * truncated, but if it begins with a period character (<code>'.'</code>) then the period and the first three characters following it will always
	 * be preserved. Once these adjustments have been made the name of the new file will be generated by concatenating the prefix, five or more
	 * internally-generated characters, and the suffix.
	 * <p>
	 * If the <code>directory</code> argument is <code>null</code> then the system-dependent default temporary-file directory will be used. The
	 * default temporary-file directory is specified by the system property <code>java.io.tmpdir</code>. On UNIX systems the default value of this
	 * property is typically <code>"/tmp"</code> or <code>"/var/tmp"</code>; on Microsoft Windows systems it is typically
	 * <code>"C:\\WINNT\\TEMP"</code>. A different value may be given to this system property when the Java virtual machine is invoked, but
	 * programmatic changes to this property are not guaranteed to have any effect upon the temporary directory used by this method.
	 * 
	 * @param prefix
	 *            The prefix string to be used in generating the file's name; must be at least three characters long
	 * @param suffix
	 *            The suffix string to be used in generating the file's name; may be <code>null</code>, in which case the suffix <code>".tmp"</code>
	 *            will be used
	 * @param directory
	 *            The directory in which the file is to be created, or <code>null</code> if the default temporary-file directory is to be used
	 * @return An abstract pathname denoting a newly-created empty file
	 * @throws IllegalArgumentException
	 *             If the <code>prefix</code> argument contains fewer than three characters
	 * @throws IOException
	 *             If a file could not be created
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method does not
	 *             allow a file to be created
	 * @since 1.2
	 */
	public static File createTempFile(String prefix, String suffix, File directory) throws IOException {
		rwlock.readLock().lock();
		try {
			if (EvoSuiteIO.activeVFS) {
				if (prefix == null)
					throw new NullPointerException();
				if (prefix.length() < 3)
					throw new IllegalArgumentException("Prefix string too short");
				String s = (suffix == null) ? ".tmp" : suffix;
				if (directory == null) {
					String tmpDir = LazyInitialization.temporaryDirectory();
					directory = new File(tmpDir, fs.prefixLength(tmpDir));
				}
				// SecurityManager sm = System.getSecurityManager(); // we don't ask the security manager
				File f;
				do {
					f = generateFile(prefix, s, directory);
					f.setOriginal(false);
				} while (!f.createNewFile());
				return f;
			} else {
				File f = createTempFileOriginal(prefix, suffix, directory);
				f.setOriginal(true);
				return f;
			}
		} finally {
			rwlock.readLock().unlock();
		}
	}

	/**
	 * <p>
	 * Creates a new empty file in the specified directory, using the given prefix and suffix strings to generate its name. If this method returns
	 * successfully then it is guaranteed that:
	 * <ol>
	 * <li>The file denoted by the returned abstract pathname did not exist before this method was invoked, and
	 * <li>Neither this method nor any of its variants will return the same abstract pathname again in the current invocation of the virtual machine.
	 * </ol>
	 * This method provides only part of a temporary-file facility. To arrange for a file created by this method to be deleted automatically, use the
	 * <code>{@link #deleteOnExit}</code> method.
	 * <p>
	 * The <code>prefix</code> argument must be at least three characters long. It is recommended that the prefix be a short, meaningful string such
	 * as <code>"hjb"</code> or <code>"mail"</code>. The <code>suffix</code> argument may be <code>null</code>, in which case the suffix
	 * <code>".tmp"</code> will be used.
	 * <p>
	 * To create the new file, the prefix and the suffix may first be adjusted to fit the limitations of the underlying platform. If the prefix is too
	 * long then it will be truncated, but its first three characters will always be preserved. If the suffix is too long then it too will be
	 * truncated, but if it begins with a period character (<code>'.'</code>) then the period and the first three characters following it will always
	 * be preserved. Once these adjustments have been made the name of the new file will be generated by concatenating the prefix, five or more
	 * internally-generated characters, and the suffix.
	 * <p>
	 * If the <code>directory</code> argument is <code>null</code> then the system-dependent default temporary-file directory will be used. The
	 * default temporary-file directory is specified by the system property <code>java.io.tmpdir</code>. On UNIX systems the default value of this
	 * property is typically <code>"/tmp"</code> or <code>"/var/tmp"</code>; on Microsoft Windows systems it is typically
	 * <code>"C:\\WINNT\\TEMP"</code>. A different value may be given to this system property when the Java virtual machine is invoked, but
	 * programmatic changes to this property are not guaranteed to have any effect upon the temporary directory used by this method.
	 * 
	 * @param prefix
	 *            The prefix string to be used in generating the file's name; must be at least three characters long
	 * @param suffix
	 *            The suffix string to be used in generating the file's name; may be <code>null</code>, in which case the suffix <code>".tmp"</code>
	 *            will be used
	 * @param directory
	 *            The directory in which the file is to be created, or <code>null</code> if the default temporary-file directory is to be used
	 * @return An abstract pathname denoting a newly-created empty file
	 * @throws IllegalArgumentException
	 *             If the <code>prefix</code> argument contains fewer than three characters
	 * @throws IOException
	 *             If a file could not be created
	 * @throws SecurityException
	 *             If a security manager exists and its <code> {@link java.lang.SecurityManager#checkWrite(java.lang.String)} </code> method does not
	 *             allow a file to be created
	 * @since 1.2
	 */
	private static File createTempFileOriginal(String prefix, String suffix, File directory) throws IOException {
		if (prefix == null)
			throw new NullPointerException();
		if (prefix.length() < 3)
			throw new IllegalArgumentException("Prefix string too short");
		String s = (suffix == null) ? ".tmp" : suffix;
		if (directory == null) {
			String tmpDir = LazyInitialization.temporaryDirectory();
			directory = new File(tmpDir, fs.prefixLength(tmpDir));
			directory.setOriginal(true); // new
		}
		SecurityManager sm = System.getSecurityManager();
		File f;
		do {
			f = generateFile(prefix, s, directory);
			f.setOriginal(true); // new
		} while (!checkAndCreate(f.getPath(), sm));
		return f;
	}

	/**
	 * Creates an empty file in the default temporary-file directory, using the given prefix and suffix to generate its name. Invoking this method is
	 * equivalent to invoking <code>{@link #createTempFile(java.lang.String,
	 * java.lang.String, java.io.File)
	 * createTempFile(prefix,&nbsp;suffix,&nbsp;null)}</code>.
	 * 
	 * @param prefix
	 *            The prefix string to be used in generating the file's name; must be at least three characters long
	 * @param suffix
	 *            The suffix string to be used in generating the file's name; may be <code>null</code>, in which case the suffix <code>".tmp"</code>
	 *            will be used
	 * @return An abstract pathname denoting a newly-created empty file
	 * @throws IllegalArgumentException
	 *             If the <code>prefix</code> argument contains fewer than three characters
	 * @throws IOException
	 *             If a file could not be created
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkWrite(java.lang.String)}</code> method does not allow a file to be created
	 * @since 1.2
	 */
	public static File createTempFile(String prefix, String suffix) throws IOException {
		return createTempFile(prefix, suffix, null);
	}

	/* -- Basic infrastructure -- */

	/**
	 * Compares two abstract pathnames lexicographically. The ordering defined by this method depends upon the underlying system. On UNIX systems,
	 * alphabetic case is significant in comparing pathnames; on Microsoft Windows systems it is not.
	 * 
	 * @param pathname
	 *            The abstract pathname to be compared to this abstract pathname
	 * @return Zero if the argument is equal to this abstract pathname, a value less than zero if this abstract pathname is lexicographically less
	 *         than the argument, or a value greater than zero if this abstract pathname is lexicographically greater than the argument
	 * @since 1.2
	 */
	public int compareTo(File pathname) { // should also work with simulation
		return fs.compare(this, pathname);
	}

	/**
	 * Tests this abstract pathname for equality with the given object. Returns <code>true</code> if and only if the argument is not <code>null</code>
	 * and is an abstract pathname that denotes the same file or directory as this abstract pathname. Whether or not two abstract pathnames are equal
	 * depends upon the underlying system. On UNIX systems, alphabetic case is significant in comparing pathnames; on Microsoft Windows systems it is
	 * not.
	 * 
	 * @param obj
	 *            The object to be compared with this abstract pathname
	 * @return <code>true</code> if and only if the objects are the same; <code>false</code> otherwise
	 */
	@Override
	public boolean equals(Object obj) { // should also work with simulation
		if ((obj != null) && (obj instanceof File)) {
			return compareTo((File) obj) == 0;
		}
		return false;
	}

	/**
	 * Computes a hash code for this abstract pathname. Because equality of abstract pathnames is inherently system-dependent, so is the computation
	 * of their hash codes. On UNIX systems, the hash code of an abstract pathname is equal to the exclusive <em>or</em> of the hash code of its
	 * pathname string and the decimal value <code>1234321</code>. On Microsoft Windows systems, the hash code is equal to the exclusive <em>or</em>
	 * of the hash code of its pathname string converted to lower case and the decimal value <code>1234321</code>. Locale is not taken into account on
	 * lowercasing the pathname string.
	 * 
	 * @return A hash code for this abstract pathname
	 */
	@Override
	public int hashCode() { // should also work with simulation
		return fs.hashCode(this);
	}

	/**
	 * Returns the pathname string of this abstract pathname. This is just the string returned by the <code>{@link #getPath}</code> method.
	 * 
	 * @return The string form of this abstract pathname
	 */
	@Override
	public String toString() {
		return getPath();
	}

	/**
	 * WriteObject is called to save this filename. The separator character is saved also so it can be replaced in case the path is reconstituted on a
	 * different host type.
	 * <p>
	 * 
	 * @serialData Default fields followed by separator character.
	 */
	private synchronized void writeObject(java.io.ObjectOutputStream s) // complicated to simulate; besides I estimatate it to not happen very often,
																		// that one wants to serialize a file object; furthermore it shouldn't destroy
																		// anything, so I leave it unchanged
			throws IOException {
		s.defaultWriteObject();
		s.writeChar(this.separatorChar); // Add the separator character
	}

	/**
	 * readObject is called to restore this filename. The original separator character is read. If it is different than the separator character on
	 * this system, then the old separator is replaced by the local separator.
	 */
	private synchronized void readObject(java.io.ObjectInputStream s) // complicated to simulate; besides I estimatate it to not happen very often,
																		// that one wants to serialize a file object; furthermore it shouldn't destroy
																		// anything, so I leave it unchanged
			throws IOException, ClassNotFoundException {
		ObjectInputStream.GetField fields = s.readFields();
		String pathField = (String) fields.get("path", null);
		char sep = s.readChar(); // read the previous separator char
		if (sep != separatorChar)
			pathField = pathField.replace(sep, separatorChar);
		this.path = fs.normalize(pathField);
		this.prefixLength = fs.prefixLength(this.path);
	}

	/** use serialVersionUID from JDK 1.0.2 for interoperability */
	private static final long serialVersionUID = 301077366599181567L;

	/* -- additional methods of EvoSuite I/O -- */

	/**
	 * checks if the given file or a parent directory of the file is in the given collection
	 * 
	 * @param file
	 *            the file to search with
	 * @param collection
	 *            the collection to search in
	 * @return <code>true</code> if the file or a parent is in the given collection, <code>false</code> otherwise
	 * @throws IOException
	 *             if getCanonicalFile() fails
	 */
	static boolean containsFileOrParent(File file, Collection<String> collection) throws IOException {
		File canonicalFile = file.getCanonicalFile();
		while (canonicalFile != null) {
			if (collection.contains(canonicalFile.getPath())) {
				return true;
			}
			final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
			final boolean VFSwasActive = EvoSuiteIO.activeVFS;
			try {
				EvoSuiteIO.activeVFS = false; // prevents labeling of the parent by EvoSuiteIO.shouldBeOriginal
				canonicalFile = canonicalFile.getParentFile(); // is already canonical
			} finally {
				EvoSuiteIO.activeVFS = VFSwasActive;
				reacquireReadLocksAndReleaseWriteLock(holdCount);
			}
		}
		return false;
	}

	/**
	 * helper method to remove canonical paths from the given collection that start with the given canonical path <br>
	 * this method was adapted from an example in the Oracle Docs on <a
	 * href="http://docs.oracle.com/javase/tutorial/collections/interfaces/collection.html">The Collection Interface</a>
	 * 
	 * @param collection
	 *            the collection to remove canonical paths from
	 * @param canonicalPath
	 *            the canonical path to filter with
	 */
	private static void filterCanonicalPath(Collection<String> collection, String canonicalPath) {
		for (Iterator<String> it = collection.iterator(); it.hasNext();)
			if (it.next().startsWith(canonicalPath))
				it.remove();
	}

	/**
	 * Applies {@link #filterCanonicalPath(Collection, String)} on all static collections in File.
	 * 
	 * @param canonicalPath
	 *            the canonical path to filter with
	 */
	private static void filterCollections(String canonicalPath) {
		// this method is very expensive wrt. runtime // TODO fix performance issue (e.g. use trees as datastructure)

		// deletes permissions of this file and all descendants
		filterCanonicalPath(permissions.keySet(), canonicalPath);

		// clears this file and all descendants from the deletedFiles set
		filterCanonicalPath(deletedFiles, canonicalPath);

		// clears this file and all descendants from the shallowCopiedDirectories set
		filterCanonicalPath(shallowCopiedDirectories, canonicalPath);

		// clears this file and all descendants from the shallowCopiedFiles set
		filterCanonicalPath(shallowCopiedFiles, canonicalPath);
	}

	/**
	 * Generates a permission bitstring that grants all access permissions. This default bitstring is used by {@link #setReadable},
	 * {@link #setWritable} and {@link #setExecutable} if no permission bitstring was set before.
	 * 
	 * @return a permission bitstring granting all access permissions
	 */
	private static int getDefaultPermissionBitstring() {
		int permissionBitstring = 0x00;
		permissionBitstring = permissionBitstring | FileSystem.ACCESS_READ;
		permissionBitstring = permissionBitstring | FileSystem.ACCESS_WRITE;
		permissionBitstring = permissionBitstring | FileSystem.ACCESS_EXECUTE;
		return permissionBitstring;
	}

	/**
	 * A helper method that scans the given collection for canonical paths that start with the given canonical path.
	 * 
	 * @param collection
	 *            the collection to search in
	 * @param canonicalPath
	 *            the canonical path to search with
	 * @return a set of canonical paths that start with the given canonical path
	 */
	private static Set<String> getMatchesForCanonicalPath(Collection<String> collection, String canonicalPath) {
		Set<String> matches = new HashSet<String>();
		for (String subPath : collection)
			if (subPath.startsWith(canonicalPath))
				matches.add(subPath);
		return matches;
	}

	/**
	 * Lists all canonical paths of all descendants of this file residing in the VFS plus the file itself. The list is ordered by depth, that is, a
	 * child appears in the list before its parent.
	 * 
	 * @return a list of canonical paths of all descendants of this file residing in the VFS plus the file itself
	 * @throws FileSystemException
	 *             on error traversing the descendants in VFS
	 * @throws IOException
	 *             if {@link #getCanonicalPath()} failes for a file
	 * @see FileObject#findFiles(org.apache.commons.vfs2.FileSelector)
	 */
	private List<String> getCanonicalPathsOfThisAndAllDescendantsInVFS() throws FileSystemException, IOException {
		if (ramFile == null)
			initRamFile();

		List<String> canonicalPathsOfDescendantFiles = new ArrayList<String>();
		final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
		try {
			EvoSuiteIO.activeVFS = false;
			for (FileObject descendant : ramFile.findFiles(Selectors.SELECT_ALL)) {
				String ramPath = descendant.getName().toString();
				File file = new File(ramPath.substring(6, ramPath.length())); // cuts away "ram://"
				canonicalPathsOfDescendantFiles.add(file.getCanonicalPath());
			}
			return canonicalPathsOfDescendantFiles;
		} finally {
			EvoSuiteIO.activeVFS = true;
			reacquireReadLocksAndReleaseWriteLock(holdCount);
		}
	}

	/**
	 * Initializes (if not yet done) and assigns the correspondent {@link RamFileObject} to this File instance. At its first access the RamFileObject
	 * is synchronized with its correspondent original file. If you want to understand in more detail how this synchronization is performend, I ask
	 * you to look into the source code of this method since it is well structured and well documented at all important places and probably more
	 * helpful than this JavaDoc method description could ever be. ;-)
	 * 
	 * @throws FileSystemException
	 *             thrown by one of the Commons VFS methods
	 * @throws IOException
	 *             thrown by {@link #getCanonicalPath()}
	 * @see #isShallowCopy()
	 * @see #shallowCopyThreshold
	 * @see #resolvedFileToCopyLiesInSandboxFolder()
	 * @see #shallowCopy(File)
	 */
	private void initRamFile() throws FileSystemException, IOException {
		final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
		try {
			EvoSuiteIO.activeVFS = false; // we want original behaviour during initialization
			
			String canonicalPath = getCanonicalPath();
			ramFile = EvoSuiteIO.getVFSManager().resolveFile("ram://" + canonicalPath);

			// note, that the following code is NOT thread-safe!
			// only proceed if the ramFile has never existed or this directory was shallow copied and not deleted so far
			if ((!ramFile.exists() || shallowCopiedDirectories.contains(canonicalPath) && !isShallowCopy())
					&& !containsFileOrParent(this, deletedFiles)) {

				fileToCopy = this; // either this original file or same file in sandbox read folder

				// copying the file or rather the state of the file into memory (if it exists and may be read)
				if ((resolvedFileToCopyLiesInSandboxFolder() || !EvoSuiteIO.readOnlyFromSandboxFolder) && fileToCopy.existsOriginal()) {

					FileObject localFile = EvoSuiteIO.getVFSManager().resolveFile(fileToCopy.getCanonicalPath());

					// before creating the file/directory determine which parent folders have to be shallow copied (and do it)
					// This is the case if the parent directory does not exist and never existed in the VFS.
					// The shallow copies are performed recursively in order from root to direct parent so that not every parent wants to shallow copy
					// its parents in turn!
					File parent = getCanonicalFile().getParentFile();
					if (parent != null && !parent.getRamFile().exists()) {
						parent.setOriginal(false);
						parent.setShallowCopy(true);
						parent.initRamFile(); // recursive initialization of parents
						EvoSuiteIO.activeVFS = false; // it was finally enabled in initRamFile() but it has to be disabled for fileToCopy.listFiles()
														// in the 'extended shallow copy' - section!
					}

					if (fileToCopy.canReadOriginal()) {
						if (fileToCopy.shallowCopy) {
							// NEW IN VERSION 0.3: only make a shallow copy - see shallowCopy(File) for more information; USE THIS WITH CAUTION:
							// THE BEHAVIOUR OF THIS FILE/DIRECTORY MAY DIFFER FROM THE ORIGINAL FILE/DIRECTORY!
							shallowCopy(fileToCopy);
						} else {
							if (fileToCopy.isDirectoryOriginal()) {
								ramFile.createFolder();
								try {
									// extended 'shallow' copy (only direct children)
									File[] subFiles = fileToCopy.listFiles();
									if (subFiles != null) {
										for (File subFile : fileToCopy.listFiles()) {
											if (subFile.isDirectoryOriginal()) {
												subFile.setShallowCopy(true);
												subFile.initRamFile();
											} else {
												subFile.initRamFile();
											}
										}
									} else {
										throw new FileSystemException("\"" + localFile + "\" could not be read!");
									}
									// extended shallow copy was successful at this point
									shallowCopiedDirectories.remove(canonicalPath);
								} catch (FileSystemException e) {
									// since this is not unusual to happen, I treat (and catch) it specially
									// even if copying partially fails the caller method may succeed anyhow
									System.err.println("WARNING: Copying of directory \"" + localFile + "\" to \"" + ramFile + "\" failed!");
									if (EvoSuiteIO.debug) {
										e.printStackTrace();
									}

									if (ramFile.getChildren().length == 0) {
										// shallowCopiedDirectories.add(canonicalPath); // while principally correct, this could again lead to a
										// FileSystemException since there may be a retry to init the folder and it still fails
										System.err.println(" \"" + ramFile + " was simply created (without children)!\n");
									} else {
										System.err.println("\n");
									}
								} finally {
									EvoSuiteIO.activeVFS = false; // (it is finally enabled in initRamFile so we need this)
								}
							} else {
								// do a shallow copy for files with size > SHALLOW_COPY_THRESHOLD bytes
								if (lengthOriginal() > shallowCopyThreshold) {
									shallowCopy(fileToCopy);
								} else { // deep copy
									try {
										ramFile.copyFrom(localFile, Selectors.SELECT_ALL); // works for files as well as for directories
										shallowCopiedFiles.remove(canonicalPath);
									} catch (FileSystemException e) {
										// since this is not unusual to happen, I treat (and catch) it specially
										System.err.print("WARNING: Copying of file \"" + localFile + "\" to \"" + ramFile + "\" failed!");
										if (EvoSuiteIO.debug) {
											e.printStackTrace();
										}

										// I simply create the file
										ramFile.createFile();
										shallowCopiedFiles.add(canonicalPath);
										System.err.print(" \"" + ramFile + "\" was simply created (without content)!\n");
									}
								}
							}
						}
					} else {
						shallowCopy(fileToCopy);
					}

					// now copy attributes

					ramFile.getContent().setLastModifiedTime(fileToCopy.lastModifiedOriginal());

					// file permissions
					int permissionBitstring = 0x00;
					if (fileToCopy.canReadOriginal())
						permissionBitstring = permissionBitstring | FileSystem.ACCESS_READ;
					if (fileToCopy.canWriteOriginalWithoutSecurityManagerCheck())
						permissionBitstring = permissionBitstring | FileSystem.ACCESS_WRITE;
					if (fileToCopy.canExecuteOriginalWithoutSecurityManagerCheck())
						permissionBitstring = permissionBitstring | FileSystem.ACCESS_EXECUTE;
					permissions.put(canonicalPath, permissionBitstring);
				}
			}
		} finally {
			EvoSuiteIO.activeVFS = true; // in any case enable the VFS again
			reacquireReadLocksAndReleaseWriteLock(holdCount);
		}
	}

	/**
	 * <b>NOTE: EvoSuiteIO.activeVFS should be false when entering this method!</b>
	 * <p>
	 * Assigns this.fileToCopy with either
	 * <ul>
	 * <li>a File instance, describing a path in the sandbox read folder, if it exists there</li>
	 * <li>or <code>this</code> File instance describing the original path</li>
	 * </ul>
	 * 
	 * @return <code>true</code> if and only if the local file describes a path in the sandbox read folder; <code>false</code> otherwise
	 * @throws IOException
	 *             thrown by {@link #getCanonicalPath()}
	 */
	private boolean resolvedFileToCopyLiesInSandboxFolder() throws IOException {
		if (EvoSuiteIO.sandboxReadFolder != null) {
			if (getCanonicalPath().startsWith(EvoSuiteIO.sandboxReadFolder.getCanonicalPath())) {
				return true;
			} else {
				File sandboxFile;
				if (isAbsolute()) {
					// if the file is absolutely addressed we cannot resolve it relatively to the sandbox read folder
					// thus, we only search for the name of the file exclusively in the sandbox read folder (and not in subdirectories)
					if (Arrays.asList(File.listRoots()).contains(this))
						// if this is a root we cannot find it in the sandbox folder
						// we need this special treatment since getName() of a root returns the empty pathname what would effectively lead to
						// setting the sandboxFile to the sandbox read folder path (which exists and therefore would be unintentionally
						// initialized)
						sandboxFile = null;
					else {
						sandboxFile = new File(EvoSuiteIO.sandboxReadFolder, getName());
					}
				} else {
					// if the file is not absolutely addressed, we resolve it relatively against the sandbox read folder
					// thus, it also can be found in subdirectories of the sandbox read folder
					sandboxFile = new File(EvoSuiteIO.sandboxReadFolder, getPath());
					// if the file has not been found yet it may be directly located in the sandbox read folder instead of a subfolder
					if (!sandboxFile.existsOriginal()) {
						sandboxFile = new File(EvoSuiteIO.sandboxReadFolder, getName());
					}
				}
				if (sandboxFile != null && sandboxFile.existsOriginal()) {
					fileToCopy = sandboxFile;
					return true;
				}

			}
		}
		return false;
	}

	/**
	 * Shallow copies a file or directory into the RAM VFS. If the file or directory already exists, this method does nothing. Shallow copied means:
	 * <ul>
	 * <li>for a file: no content is copied</li>
	 * <li>for a directory: no descendants are copied</li>
	 * </ul>
	 * <p>
	 * <b>USE THIS WITH CAUTION: THE BEHAVIOUR OF THIS FILE/DIRECTORY MAY DIFFER FROM THE ORIGINAL FILE/DIRECTORY!</b>
	 * <p>
	 * Note: No permissions are copied from the original file to the RAM file by this method. That is performed by {@link #initRamFile()} which in
	 * turn also calls {@link #shallowCopy(File)} if required!
	 * 
	 * @since version 0.3
	 * @param fileToCopy
	 *            the file or directory to be shallow copied into memory under the canonical path of <code>this</code> file
	 * @throws IOException
	 *             thrown by {@link #getCanonicalPath()}
	 */
	private void shallowCopy(File fileToCopy) throws IOException {
		if (fileToCopy.isFileOriginal()) {
			ramFile.createFile();
			shallowCopiedFiles.add(getCanonicalPath());
		} else if (fileToCopy.isDirectoryOriginal()) {
			ramFile.createFolder();
			shallowCopiedDirectories.add(getCanonicalPath());
		} else {
			throw new IOException(fileToCopy.toString() + "is neither a file nor a folder!");
		}
	}

	/**
	 * Forces this virtual file to be reinitialized at the <b>next</b> access, thus synchronizing it with its correspondent original file on the real
	 * file system. This is accomplished by deleting this file and possible descendants. Note, that simply deleting a virtual file will NOT cause
	 * synchronization with the real filesystem at the next access!
	 * <p>
	 * If this file is an original file, this method does nothing.
	 * 
	 * @throws IOException
	 *             may occur during deletion of files
	 * @see #isOriginal()
	 * @see #initRamFile()
	 */
	public void forceReinitialization() throws IOException {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramFile == null) {
						initRamFile();
					}
					ramFile.delete(Selectors.SELECT_ALL); // deletes the ramFile and all its descendants
					ramFile = null;
					filterCollections(getCanonicalPath());
				}
			} finally {
				rwlock.readLock().unlock();
			}
		}
	}

	/**
	 * Helper method for deep deletion of a directory. Deep deletion means that also all descendants of this directory are deleted. Works also for
	 * normal files (without descendants).
	 * 
	 * @return <code>true</code> if this file/directory existed before this method was called AND all descendants of this file/directory in VFS and
	 *         the file/directory itself have been successfully deleted; <code>false</code> otherwise
	 * @throws FileSystemException
	 *             thrown by {@link FileObject#delete(org.apache.commons.vfs2.FileSelector)}
	 * @throws IOException
	 *             thrown by {@link #getCanonicalPathsOfThisAndAllDescendantsInVFS()}
	 */
	public boolean deepDelete() throws FileSystemException, IOException {
		if (!exists())
			return false;

		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					List<String> newDeletedFiles = getCanonicalPathsOfThisAndAllDescendantsInVFS();
					ramFile.delete(Selectors.SELECT_ALL);
					deletedFiles.addAll(newDeletedFiles);
					filterCanonicalPath(permissions.keySet(), getCanonicalPath());
					return true;
				}
			} finally {
				rwlock.readLock().unlock();
			}
		}
		// implementation for an original file/directory
		boolean complete = true;
		if (isDirectoryOriginal()) {
			for (File f : listFiles()) {
				f.setOriginal(isOriginal);
				complete &= f.deepDelete();
			}
		}
		complete &= delete();
		return complete;
	}

	/**
	 * Package private method that forces the deep copy of an original file to its correspondent virtual file (even if the
	 * {@link #shallowCopyThreshold} is violated). The original file is first searched in the sandbox read folder. <b>NOTE: This method was
	 * specifically created for its use in FileInputStream, FileOutputStream and File.forceDeepDirectoryCopy! I therefore left out almost any
	 * unnecessary sanity check. If you plan on using it, BE SURE TO DOUBLE-CHECK ITS SOURCE CODE!</b>
	 * 
	 * @throws FileSystemException
	 *             thrown by one of the used Commons VFS methods
	 * @throws IOException
	 *             thrown by {@link #getCanonicalPath()}
	 */
	void forceDeepFileCopy() throws FileSystemException, IOException {
		if (File.shallowCopiedFiles.contains(getCanonicalPath()) && !File.containsFileOrParent(this, File.deletedFiles)) {

			final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
			try {
				EvoSuiteIO.activeVFS = false; // we want original behaviour during initialization
				
				fileToCopy = this; // either this original file or same file in sandbox read folder
				ramFile = EvoSuiteIO.getVFSManager().resolveFile("ram://" + getCanonicalPath());

				if ((resolvedFileToCopyLiesInSandboxFolder() || !EvoSuiteIO.readOnlyFromSandboxFolder) && fileToCopy.canReadOriginal()) {

					FileObject localFile = EvoSuiteIO.getVFSManager().resolveFile(fileToCopy.getCanonicalPath());
					ramFile.copyFrom(localFile, Selectors.SELECT_ALL);
					shallowCopiedFiles.remove(getCanonicalPath());
				}
			} finally {
				EvoSuiteIO.activeVFS = true; // in any case enable the VFS again
				reacquireReadLocksAndReleaseWriteLock(holdCount);
			}
		}

	}

	/**
	 * Forces deep copy of a directory. <b>Note: This private method was specifically created for its use in File.renameTo. If you plan on using it,
	 * be sure to double-check its source code!</b>
	 * 
	 * @return <code>true</code> if deep directory copying succeeded; <code>false</code> otherwise
	 * @throws FileSystemException
	 *             thrown by one of the used Commons VFS methods
	 * @throws IOException
	 *             thrown by {@link #getCanonicalPath()}
	 */
	private boolean forceDeepDirectoryCopy() throws FileSystemException, IOException {
		String canonicalPath = getCanonicalPath();
		if (shallowCopiedDirectories.contains(canonicalPath)) {
			setShallowCopy(false);
			initRamFile(); // now all direct children are in VFS
		}

		// if there are no more subfolders or subfiles to deep copy, we are done
		if (getMatchesForCanonicalPath(shallowCopiedDirectories, canonicalPath).isEmpty()
				|| getMatchesForCanonicalPath(shallowCopiedFiles, canonicalPath).isEmpty())
			return true;

		// else start the deep copying process
		final int holdCount = releaseAllReadLocksAndAcquireWriteLock();
		try {
			EvoSuiteIO.activeVFS = false; // we want original behaviour during initialization
			
			fileToCopy = this; // either this original file or same file in sandbox read folder
			ramFile = EvoSuiteIO.getVFSManager().resolveFile("ram://" + canonicalPath);

			if ((resolvedFileToCopyLiesInSandboxFolder() || !EvoSuiteIO.readOnlyFromSandboxFolder)) {

				if (!fileToCopy.canReadOriginal())
					return false;

				FileObject localFile = EvoSuiteIO.getVFSManager().resolveFile(fileToCopy.getCanonicalPath());

				try {
					// deep copy
					File[] subFiles = fileToCopy.listFiles();
					if (subFiles != null) {
						for (File subFile : fileToCopy.listFiles()) {
							if (subFile.isDirectoryOriginal()) {
								if (!subFile.forceDeepDirectoryCopy())
									return false;
							} else {
								subFile.forceDeepFileCopy(); // was already initialized (at least shallow)
							}
						}
					} else {
						throw new FileSystemException("\"" + localFile + "\" could not be read!");
					}
				} catch (FileSystemException e) {
					// since this is not unusual to happen, I treat (and catch) it specially
					// even if copying partially fails the caller method may succeed anyhow
					System.err.println("WARNING: Copying of directory \"" + localFile + "\" to \"" + ramFile + "\" failed!");
					if (EvoSuiteIO.debug) {
						e.printStackTrace();
					}

					if (ramFile.getChildren().length == 0) {
						// shallowCopiedDirectories.add(canonicalPath); // while principally correct, this could again lead to a
						// FileSystemException since there may be a retry to init the folder and it still fails
						System.err.println(" \"" + ramFile + " was simply created (without children)!\n");
					} else {
						System.err.println("\n");
					}
					return false;
				}
			}
		} finally {
			EvoSuiteIO.activeVFS = true; // in any case enable the VFS again
			reacquireReadLocksAndReleaseWriteLock(holdCount);
		}
		return true;
	}

	/**
	 * Either chooses {@link #forceDeepDirectoryCopy()} or {@link #forceDeepFileCopy()} depending on what this file instance is. Be sure to check the
	 * documentation of these methods before using this private method!
	 * 
	 * @see #forceDeepDirectoryCopy()
	 * @see #forceDeepFileCopy()
	 * @return <code>true</code> if successful; <code>false</code> otherwise
	 * @throws IOException
	 *             either thrown by {@link #forceDeepDirectoryCopy()} or by {@link #forceDeepFileCopy()}
	 */
	private boolean forceDeepCopy() throws IOException {
		if (isFile()) {
			forceDeepFileCopy();
			return true;
		} else {
			return forceDeepDirectoryCopy();
		}
	}

	/**
	 * Returns the correspondent memory representation of this file; creates it if it does not already exist
	 * 
	 * @return the virtual representation of this file in memory
	 * @throws IOException
	 *             if initializing the {@link RamFileObject} instance failed
	 * @see #getRamFile()
	 */
	public FileObject getInitializedRamFile() throws IOException {
		if (ramFile == null) {
			initRamFile();
		}
		return ramFile;
	}

	/**
	 * returns the correspondent memory path of this file; the {@link #ramFile} may or may not have been initialized
	 * 
	 * @return a {@link RamFileObject} instance denoting the memory path of this file
	 * @throws FileSystemException
	 *             thrown by {@link FileSystemManager#resolveFile(String)}
	 * @throws IOException
	 *             thrown by {@link #getCanonicalPath()}
	 * @see #getInitializedRamFile()
	 */
	public FileObject getRamFile() throws FileSystemException, IOException {
		FileObject ramFile = this.ramFile;
		if (ramFile == null) {
			ramFile = EvoSuiteIO.getVFSManager().resolveFile("ram://" + getCanonicalPath());
		}
		return ramFile;
	}

	/**
	 * If the file is not an original file, it is named 'virtual' file.
	 * 
	 * @return <code>true</code> if the file shall be treated with the original implementation; <code>false</code> otherwise
	 */
	public boolean isOriginal() {
		return isOriginal;
	}

	/**
	 * If the file is not an original file, it is named 'virtual' file.
	 * 
	 * @param isOriginal
	 *            <code>true</code> if the file shall be treated with the original implementation; <code>false</code> otherwise
	 */
	public void setOriginal(boolean isOriginal) {
		this.isOriginal = isOriginal;
	}

	/**
	 * This method is only meaningful for virtual files (that are to be stored in the VFS). It determines, if the file shall be shallow copied from
	 * disk to memory during initialization of the ramFile. Shallow copied means:
	 * <ul>
	 * <li>for a file: no content is copied</li>
	 * <li>for a directory: no descendants are copied</li>
	 * </ul>
	 * 
	 * @return <code>true</code> if this file shall be shallow copied from disk to memory, <code>false</code> otherwise
	 * @since version 0.3
	 */
	public boolean isShallowCopy() {
		return shallowCopy;
	}

	/**
	 * This method is only meaningful for virtual files (that are to be stored in the VFS). It determines, if the file shall be shallow copied from
	 * disk to memory during initialization of the ramFile. Shallow copied means:
	 * <ul>
	 * <li>for a file: no content is copied</li>
	 * <li>for a directory: no descendants are copied</li>
	 * </ul>
	 * <b>USE THIS WITH CAUTION: THE BEHAVIOUR OF THIS FILE/DIRECTORY MAY DIFFER FROM THE ORIGINAL FILE/DIRECTORY!</b>
	 * <p>
	 * 
	 * @param shallowCopy
	 *            <code>true</code> if the file shall be shallow copied from disk to memory during {@link #ramFile} initialization, <code>false</code>
	 *            otherwise
	 * @since version 0.3
	 */
	public void setShallowCopy(boolean shallowCopy) {
		this.shallowCopy = shallowCopy;
	}

}
