/*
 * %W% %E%
 * 
 * Copyright (c) 2006, Oracle and/or its affiliates. All rights reserved. ORACLE
 * PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package java.io;

import static java.io.EvoSuiteIO.rwlock;

import java.nio.channels.FileChannel;

import org.apache.commons.vfs2.FileSystemException;
import org.apache.commons.vfs2.provider.DefaultFileContent;

import sun.nio.ch.FileChannelImpl;

/**
 * ++++++ modified version of the 'original' <code>FileOutputStream</code> class; additionally allows complete in-memory I/O ++++++
 * 
 * <p>
 * A file output stream is an output stream for writing data to a <code>File</code> or to a <code>FileDescriptor</code>. Whether or not a file is
 * available or may be created depends upon the underlying platform. Some platforms, in particular, allow a file to be opened for writing by only one
 * <tt>FileOutputStream</tt> (or other file-writing object) at a time. In such situations the constructors in this class will fail if the file
 * involved is already open.
 * 
 * <p>
 * <code>FileOutputStream</code> is meant for writing streams of raw bytes such as image data. For writing streams of characters, consider using
 * <code>FileWriter</code>.
 * 
 * @author Arthur van Hoff
 * @author Daniel Muth
 * @version %I%, %G%
 * @see java.io.File
 * @see java.io.FileDescriptor
 * @see java.io.FileInputStream
 * @since JDK1.0
 */
public class FileOutputStream extends OutputStream {
	/**
	 * The system dependent file descriptor. The value is 1 more than actual file descriptor. This means that the default value 0 indicates that the
	 * file is not open.
	 */
	private FileDescriptor fd;

	private FileChannel channel = null;

	private boolean append = false;

	/**
	 * new; gets set in constructor; represents the output stream of a given virtual file; stays null for original files
	 */
	private OutputStream ramOutputStream;

	/**
	 * flag, which decides, if the FileOutputStream instance shall use new or original implementation
	 */
	private boolean isOriginal = true;

	private final Object closeLock = new Object();
	private volatile boolean closed = false;

	private static final ThreadLocal<Boolean> runningFinalize = new ThreadLocal<Boolean>();

	private static boolean isRunningFinalize() {
		Boolean val;
		if ((val = runningFinalize.get()) != null)
			return val.booleanValue();
		return false;
	}

	/**
	 * Creates an output file stream to write to the file with the specified name. A new <code>FileDescriptor</code> object is created to represent
	 * this file connection.
	 * <p>
	 * First, if there is a security manager, its <code>checkWrite</code> method is called with <code>name</code> as its argument.
	 * <p>
	 * If the file exists but is a directory rather than a regular file, does not exist but cannot be created, or cannot be opened for any other
	 * reason then a <code>FileNotFoundException</code> is thrown.
	 * 
	 * @param name
	 *            the system-dependent filename
	 * @exception FileNotFoundException
	 *                if the file exists but is a directory rather than a regular file, does not exist but cannot be created, or cannot be opened for
	 *                any other reason
	 * @exception SecurityException
	 *                if a security manager exists and its <code>checkWrite</code> method denies write access to the file.
	 * @see java.lang.SecurityManager#checkWrite(java.lang.String)
	 */
	public FileOutputStream(String name) throws FileNotFoundException {
		this(name != null ? new File(name) : null, false);
	}

	/**
	 * Creates an output file stream to write to the file with the specified <code>name</code>. If the second argument is <code>true</code>, then
	 * bytes will be written to the end of the file rather than the beginning. A new <code>FileDescriptor</code> object is created to represent this
	 * file connection.
	 * <p>
	 * First, if there is a security manager, its <code>checkWrite</code> method is called with <code>name</code> as its argument.
	 * <p>
	 * If the file exists but is a directory rather than a regular file, does not exist but cannot be created, or cannot be opened for any other
	 * reason then a <code>FileNotFoundException</code> is thrown.
	 * 
	 * @param name
	 *            the system-dependent file name
	 * @param append
	 *            if <code>true</code>, then bytes will be written to the end of the file rather than the beginning
	 * @exception FileNotFoundException
	 *                if the file exists but is a directory rather than a regular file, does not exist but cannot be created, or cannot be opened for
	 *                any other reason.
	 * @exception SecurityException
	 *                if a security manager exists and its <code>checkWrite</code> method denies write access to the file.
	 * @see java.lang.SecurityManager#checkWrite(java.lang.String)
	 * @since JDK1.1
	 */
	public FileOutputStream(String name, boolean append) throws FileNotFoundException {
		this(name != null ? new File(name) : null, append);
	}

	/**
	 * Creates a file output stream to write to the file represented by the specified <code>File</code> object. A new <code>FileDescriptor</code>
	 * object is created to represent this file connection.
	 * <p>
	 * First, if there is a security manager, its <code>checkWrite</code> method is called with the path represented by the <code>file</code> argument
	 * as its argument.
	 * <p>
	 * If the file exists but is a directory rather than a regular file, does not exist but cannot be created, or cannot be opened for any other
	 * reason then a <code>FileNotFoundException</code> is thrown.
	 * 
	 * @param file
	 *            the file to be opened for writing.
	 * @exception FileNotFoundException
	 *                if the file exists but is a directory rather than a regular file, does not exist but cannot be created, or cannot be opened for
	 *                any other reason
	 * @exception SecurityException
	 *                if a security manager exists and its <code>checkWrite</code> method denies write access to the file.
	 * @see java.io.File#getPath()
	 * @see java.lang.SecurityException
	 * @see java.lang.SecurityManager#checkWrite(java.lang.String)
	 */
	public FileOutputStream(File file) throws FileNotFoundException {
		this(file, false);
	}

	/**
	 * Creates a file output stream to write to the file represented by the specified <code>File</code> object. If the second argument is
	 * <code>true</code>, then bytes will be written to the end of the file rather than the beginning. A new <code>FileDescriptor</code> object is
	 * created to represent this file connection.
	 * <p>
	 * First, if there is a security manager, its <code>checkWrite</code> method is called with the path represented by the <code>file</code> argument
	 * as its argument.
	 * <p>
	 * If the file exists but is a directory rather than a regular file, does not exist but cannot be created, or cannot be opened for any other
	 * reason then a <code>FileNotFoundException</code> is thrown.
	 * 
	 * @param file
	 *            the file to be opened for writing.
	 * @param append
	 *            if <code>true</code>, then bytes will be written to the end of the file rather than the beginning
	 * @exception FileNotFoundException
	 *                if the file exists but is a directory rather than a regular file, does not exist but cannot be created, or cannot be opened for
	 *                any other reason
	 * @exception SecurityException
	 *                if a security manager exists and its <code>checkWrite</code> method denies write access to the file.
	 * @see java.io.File#getPath()
	 * @see java.lang.SecurityException
	 * @see java.lang.SecurityManager#checkWrite(java.lang.String)
	 * @since 1.4
	 */
	public FileOutputStream(File file, boolean append) throws FileNotFoundException {
		if (!file.isOriginal()) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					this.isOriginal = false;
					String name = (file != null ? file.getPath() : null);
					// leave out security manager check here //
					if (name == null) {
						throw new NullPointerException();
					}
					fd = new FileDescriptor(); // should also work with virtual file
					fd.incrementAndGetUseCount(); // same
					this.append = append;

					// throw FileNotFoundException in three cases in order to be in line with specification
					// 1.) If the file exists but is a directory rather than a regular file
					if (file.isDirectory())
						throw new FileNotFoundException(file.getAbsolutePath() + " is an existing directory!");

					// 2.) does not exist but cannot be created
					boolean fileExists = file.exists();
					boolean fileCouldBeCreated = false;
					// in the if-statement following this try-catch block, file.createNewFile() would not be called if file.exists() is true
					// since (false AND x = false) and the compiler will skip 'x'
					if (!fileExists) {
						try {
							fileCouldBeCreated = file.createNewFile();
						} catch (FileNotFoundException e) { // thrown if the parent folder does not exist
							throw e; // pass it on
						} catch (IOException e) { // may also be a FileSystemException
							if (!EvoSuiteIO.tolerantExceptionHandling)
								throw new IllegalStateException(e);
							if (EvoSuiteIO.debug)
								e.printStackTrace();
						}
					}
					if (!fileExists && !fileCouldBeCreated) // this possibly creates the file but that is okay since the original implementation also
															// creates the file when calling open(String) or openAppend(String) - please note that
															// file.createNewFile() may just shallow copy the parent directory; see
															// File.createNewFile() for more information
						throw new FileNotFoundException(file.getAbsolutePath()
								+ " cannot be created due to restricted write access to the target directory!");

					// 3.) cannot be opened for any other reason
					if (file.exists() && !file.canWrite())
						throw new FileNotFoundException(file.getAbsolutePath() + " is read-only!");
					try {
						if (append) {
							file.forceDeepFileCopy(); // if file was shallow copied and append is true, deep copy it
						} else {
							File.shallowCopiedFiles.remove(file.getCanonicalPath()); // in any case remove the file as a shallowCopiedFile now (since
																						// we will write to it)
						}

						/*
						 * PROBLEM: There may only be a single RamFileOutputStream open for a file at any time, but it is allowed to have multiple
						 * FileOutputStream instances to the same file open! So this will throw a FileSystemException, when the real implementation
						 * does not, and all further accesses to ramOutputStream will throw a NullpointerException! - To quickfix this issue, I
						 * 'transform' the FileSystemException into a FileNotFoundException, which can be thrown by this constructor (even though for
						 * other reasons) TODO find a better solution for this (won't be simple)
						 */
						ramOutputStream = file.getInitializedRamFile().getContent().getOutputStream(append); // is an instance of
																												// FileContentOutputStream (subclass
																												// of BufferedOutputStream); uses an
																												// RamFileOutputStream
						return;
					} catch (IOException e) {
						// see PROBLEM a few lines above this comment
						if (e instanceof FileSystemException) {
							FileSystemException casted = (FileSystemException) e;
							if (casted.getCode() != null && casted.getCode().equals("vfs.provider/write-in-use.error")) {
								throw new FileNotFoundException("There must not be more than one FileOutputStream open to the virtual file "
										+ file.getAbsolutePath() + " at the same time!");
							}
						}
						// if the exception is not caused by the 'multiple open output streams' error, apply default behaviour
						if (!EvoSuiteIO.tolerantExceptionHandling)
							throw new IllegalStateException(e);
						if (EvoSuiteIO.debug)
							e.printStackTrace();
						return;
					}
				}
			} finally {
				rwlock.readLock().unlock();
			}
		}
		// original code
		String name = (file != null ? file.getPath() : null);
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
			security.checkWrite(name);
		}
		if (name == null) {
			throw new NullPointerException();
		}
		fd = new FileDescriptor();
		fd.incrementAndGetUseCount();
		this.append = append;
		if (append) {
			openAppend(name); // this creates the file, if it does not already exist!
		} else {
			open(name); // this creates the file, if it does not already exist!
		}
	}

	/**
	 * Creates an output file stream to write to the specified file descriptor, which represents an existing connection to an actual file in the file
	 * system.
	 * <p>
	 * First, if there is a security manager, its <code>checkWrite</code> method is called with the file descriptor <code>fdObj</code> argument as its
	 * argument.
	 * <p>
	 * <b>NOTE: The FileDescriptor.sync() method will throw a SyncFailedException for virtual files and FIS/FOS instances, that were constructed with
	 * a FileDesciptor describing a virtual file, will throw IOExceptions on any operation. But I do not consider this as an issue, since according to
	 * the specification of FileDescriptor "Applications should not create their own file descriptors." In fact they are only used for the standard
	 * input stream, standard output stream and standard error stream, which are not treated by my thesis and should therefore use the original
	 * implementation anyway!</b>
	 * 
	 * @param fdObj
	 *            the file descriptor to be opened for writing
	 * @exception SecurityException
	 *                if a security manager exists and its <code>checkWrite</code> method denies write access to the file descriptor
	 * @see java.lang.SecurityManager#checkWrite(java.io.FileDescriptor)
	 */
	public FileOutputStream(FileDescriptor fdObj) {
		/*
		 * the FileDescriptor.sync() method will throw a SyncFailedException for virtual files and FIS/FOS instances, that were constructed with a
		 * FileDesciptor describing a virtual file, will throw IOExceptions on any operation. But I do not consider this as an issue, since according
		 * to the specification of FileDescriptor "Applications should not create their own file descriptors." In fact they are only used for the
		 * standard input stream, standard output stream and standard error stream, which are not treated by my thesis and should therefore use the
		 * original implementation anyway!
		 */
		SecurityManager security = System.getSecurityManager();
		if (fdObj == null) {
			throw new NullPointerException();
		}
		if (security != null) {
			security.checkWrite(fdObj);
		}
		fd = fdObj;

		/*
		 * FileDescriptor is being shared by streams. Ensure that it's GC'ed only when all the streams/channels are done using it.
		 */
		fd.incrementAndGetUseCount();
	}

	/**
	 * Opens a file, with the specified name, for writing.
	 * 
	 * @param name
	 *            name of file to be opened
	 */
	private native void open(String name) throws FileNotFoundException; // not used by simulation

	/**
	 * Opens a file, with the specified name, for appending.
	 * 
	 * @param name
	 *            name of file to be opened
	 */
	private native void openAppend(String name) throws FileNotFoundException; // not used by simulation

	/**
	 * Writes the specified byte to this file output stream. Implements the <code>write</code> method of <code>OutputStream</code>. <b>NOTE: If this
	 * is a virtual FileOutputStream, you may have to <code>flush</code> or <code>close</code> it before reading the accessed file, since it makes use
	 * of a {@link BufferedOutputStream} which may have to be forced to pass on buffered bytes the VFS!</b>
	 * 
	 * @param b
	 *            the byte to be written.
	 * @exception IOException
	 *                if an I/O error occurs.
	 */
	@Override
	public void write(int b) throws IOException {
		if (EvoSuiteIO.activeVFS && !isOriginal) {
			ramOutputStream.write(b);
		} else {
			// writeOriginal(b); // this will throw an UnsatisfiedLinkError (see writeOriginal()); the following code should solve that problem
			byte[] buffer = new byte[1];
			buffer[0] = (byte) b;
			write(buffer);
		}
	}

	// /**
	// * Writes the specified byte to this file output stream. Implements the <code>write</code> method of <code>OutputStream</code>.
	// *
	// * @param b
	// * the byte to be written.
	// * @exception IOException
	// * if an I/O error occurs.
	// */
	// private native void writeOriginal(int b) throws IOException; // this will throw an UnsatisfiedLinkError, since the C code expects a native
	// method 'write(int b)'

	/**
	 * Writes a sub array as a sequence of bytes.
	 * 
	 * @param b
	 *            the data to be written
	 * @param off
	 *            the start offset in the data
	 * @param len
	 *            the number of bytes that are written
	 * @exception IOException
	 *                If an I/O error has occurred.
	 */
	private native void writeBytes(byte b[], int off, int len) // private and implementation specific - does not need to be simulated
			throws IOException;

	/**
	 * Writes <code>b.length</code> bytes from the specified byte array to this file output stream. <b>NOTE: If this is a virtual FileOutputStream,
	 * you may have to <code>flush</code> or <code>close</code> it before reading the accessed file, since it makes use of a
	 * {@link BufferedOutputStream} which may have to be forced to pass on buffered bytes the VFS!</b>
	 * 
	 * @param b
	 *            the data.
	 * @exception IOException
	 *                if an I/O error occurs.
	 */
	@Override
	public void write(byte b[]) throws IOException {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					ramOutputStream.write(b);
					return;
				}
			} finally {
				rwlock.readLock().unlock();
			}
		}
		writeBytes(b, 0, b.length);
	}

	/**
	 * Writes <code>len</code> bytes from the specified byte array starting at offset <code>off</code> to this file output stream. <b>NOTE: If this is
	 * a virtual FileOutputStream, you may have to <code>flush</code> or <code>close</code> it before reading the accessed file, since it makes use of
	 * a {@link BufferedOutputStream} which may have to be forced to pass on buffered bytes the VFS!</b>
	 * 
	 * @param b
	 *            the data.
	 * @param off
	 *            the start offset in the data.
	 * @param len
	 *            the number of bytes to write.
	 * @exception IOException
	 *                if an I/O error occurs.
	 */
	@Override
	public void write(byte b[], int off, int len) throws IOException {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					ramOutputStream.write(b, off, len);
					return;
				}
			} finally {
				rwlock.readLock().unlock();
			}
		}
		writeBytes(b, off, len);
	}

	/**
	 * In the original implementation of <code>FileOutputStream</code> <code>flush()</code> is not overridden! However, we need to override it for a
	 * virtual FileOutputStream, which is an instance of {@link DefaultFileContent.FileContentOutputStream} (a subclass of
	 * {@link BufferedOutputStream}, that is known to override <code>flush()</code>)
	 * 
	 */
	@Override
	public void flush() throws IOException {
		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					ramOutputStream.flush();
				}
			} finally {
				rwlock.readLock().unlock();
			}
		}
		// the original implementation does not override flush and therefore uses the implementation of OutputStream which does nothing
	}

	/**
	 * Closes this file output stream and releases any system resources associated with this stream. This file output stream may no longer be used for
	 * writing bytes.
	 * 
	 * <p>
	 * If this stream has an associated channel then the channel is closed as well.
	 * 
	 * @exception IOException
	 *                if an I/O error occurs.
	 * 
	 * @revised 1.4
	 * @spec JSR-51
	 */
	@Override
	public void close() throws IOException {
		synchronized (closeLock) {
			if (closed) {
				return;
			}
			closed = true;
		}

		if (channel != null) {
			/*
			 * Decrement FD use count associated with the channel The use count is incremented whenever a new channel is obtained from this stream.
			 */
			fd.decrementAndGetUseCount();
			channel.close();
		}

		/*
		 * Decrement FD use count associated with this stream
		 */
		int useCount = fd.decrementAndGetUseCount();

		if (!isOriginal) {
			rwlock.readLock().lock();
			try {
				if (EvoSuiteIO.activeVFS) {
					if (ramOutputStream != null) {
						ramOutputStream.close();
					}
					return;
				}
			} finally {
				rwlock.readLock().unlock();
			}
		} else {
			/*
			 * If FileDescriptor is still in use by another stream, the finalizer will not close it.
			 */
			if ((useCount <= 0) || !isRunningFinalize()) {
				close0();
			}
		}
	}

	/**
	 * Returns the file descriptor associated with this stream.
	 * 
	 * @return the <code>FileDescriptor</code> object that represents the connection to the file in the file system being used by this
	 *         <code>FileOutputStream</code> object.
	 * 
	 * @exception IOException
	 *                if an I/O error occurs.
	 * @see java.io.FileDescriptor
	 */
	public final FileDescriptor getFD() throws IOException {
		if (fd != null)
			return fd;
		throw new IOException();
	}

	/**
	 * Returns the unique {@link java.nio.channels.FileChannel FileChannel} object associated with this file output stream. </p>
	 * 
	 * <p>
	 * The initial {@link java.nio.channels.FileChannel#position() </code>position<code>} of the returned channel will be equal to the number of bytes
	 * written to the file so far unless this stream is in append mode, in which case it will be equal to the size of the file. Writing bytes to this
	 * stream will increment the channel's position accordingly. Changing the channel's position, either explicitly or by writing, will change this
	 * stream's file position.
	 * 
	 * @return the file channel associated with this file output stream
	 * 
	 * @since 1.4
	 * @spec JSR-51
	 */
	public FileChannel getChannel() { // TODO simulate FileChannel
		synchronized (this) {
			if (channel == null) {
				channel = FileChannelImpl.open(fd, false, true, this, append);

				/*
				 * Increment fd's use count. Invoking the channel's close() method will result in decrementing the use count set for the channel.
				 */
				fd.incrementAndGetUseCount();
			}
			return channel;
		}
	}

	/**
	 * Cleans up the connection to the file, and ensures that the <code>close</code> method of this file output stream is called when there are no
	 * more references to this stream.
	 * 
	 * @exception IOException
	 *                if an I/O error occurs.
	 * @see java.io.FileInputStream#close()
	 */
	@Override
	protected void finalize() throws IOException { // should also work with simulation
		if (fd != null) {
			if (fd == FileDescriptor.out || fd == FileDescriptor.err) {
				flush();
			} else {

				/*
				 * Finalizer should not release the FileDescriptor if another stream is still using it. If the user directly invokes close() then the
				 * FileDescriptor is also released.
				 */
				runningFinalize.set(Boolean.TRUE);
				try {
					close();
				} finally {
					runningFinalize.set(Boolean.FALSE);
				}
			}
		}
	}

	private native void close0() throws IOException;

	// not sure, if this could compromise simulation // TODO check that
	private static native void initIDs();

	static {
		initIDs();
	}

}
