package org.ginpink;

import java.io.IOException;
import java.lang.reflect.Method;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.antlr.runtime.RecognitionException;
import org.ginpink.daikon.GinPink;
import org.ginpink.daikon.GinPinkProcessor;
import org.ginpink.eval.JavaHeapEvaluator;
import org.ginpink.expr.JExpr;
import org.ginpink.parser.InvariantFileReader;
import org.ginpink.parser.InvariantFileWriter;
import org.ginpink.parser.MutationsFileReader;
import org.ginpink.typecheck.ExprTypeChecker;
import org.ginpink.typecheck.ExprVisibilityChecker;

public final class GinPinkChecker {

	private static final String LOOP_INVARIANT = "loop_invariant";

	private final String className;
	private final String methodName;
	private final List<LoopInvariantMethod> loop_invariant_methods;

	private List<Invariant> parsed_invariants = new LinkedList<Invariant>();

	public GinPinkChecker(String className, String methodName)
			throws ClassNotFoundException, NoSuchMethodException {
		super();
		this.className = className;
		this.methodName = methodName;
		String loop_invariant_prefix_name = methodName + "_" + LOOP_INVARIANT;
		this.loop_invariant_methods = getLoopInvariantMethods(className,
				loop_invariant_prefix_name);

		if (this.loop_invariant_methods.isEmpty()) {
			throw new NoSuchMethodException(
					"No method was found with prefix name: "
							+ loop_invariant_prefix_name);
		}
	}

	private static List<LoopInvariantMethod> getLoopInvariantMethods(
			String className, String loop_invariant_method_prefix)
			throws ClassNotFoundException {
		Class<?> clazz = ClassLoader.getSystemClassLoader()
				.loadClass(className);
		Method[] methods = clazz.getDeclaredMethods();
		List<LoopInvariantMethod> loop_invariant_methods = new LinkedList<LoopInvariantMethod>();
		for (int i = 0; i < methods.length; i++) {
			Method method = methods[i];
			if (method.getName().startsWith(loop_invariant_method_prefix)) {
				LoopInvariantMethod loop_invariant_method = new LoopInvariantMethod(
						method);
				loop_invariant_methods.add(loop_invariant_method);
			}
		}
		return loop_invariant_methods;
	}

	public List<Invariant> parseMutations(String ginpinkFilename)
			throws IOException, RecognitionException {
		MutationsFileReader reader = new MutationsFileReader(className,
				methodName);
		reader.readFile(ginpinkFilename);
		List<Invariant> mutations = reader.getExpressions(className + "."
				+ methodName);
		parsed_invariants.addAll(mutations);
		return parsed_invariants;
	}

	public void checkTyping() {
		if (this.parsed_invariants == null) {
			throw new IllegalStateException(
					"No new parsed expressions to typecheck!");
		}

		for (LoopInvariantMethod loop_invariant_method : loop_invariant_methods) {
			Method reflectMethod = loop_invariant_method.getReflectMethod();
			ExprTypeChecker type_checker = new ExprTypeChecker(reflectMethod);
			ExprVisibilityChecker visibility_checker = new ExprVisibilityChecker(
					reflectMethod);

			for (Invariant parsed_invariant : parsed_invariants) {
				JExpr parsed_expr = parsed_invariant.getJavaExpr();
				JExpr type_checked_expr = parsed_expr.accept(type_checker);
				if (type_checked_expr.isTypeChecked()) {
					boolean visibility_check = type_checked_expr
							.accept(visibility_checker);
					if (visibility_check) {
						String java_string = parsed_invariant.getJavaString();
						String jml_string = parsed_invariant.getJmlString();
						Invariant typechecked_invariant = new Invariant(
								type_checked_expr, java_string, jml_string);
						loop_invariant_method
								.addInvariant(typechecked_invariant);
					}
				}
			}

		}

		this.parsed_invariants = null;
	}

	private String pptSelectPattern;
	private String pptOmitPattern;

	public void checkInvariants(String dtraceFilename)
			throws IllegalStateException, IOException {
		GinPink ginPink = new GinPink();
		GinPinkProcessor processor = new GinPinkProcessor();

		if (pptOmitPattern != null) {
			processor.setPptOmitPattern(pptOmitPattern);
		}
		if (pptSelectPattern != null) {
			processor.setPptSelectPattern(pptSelectPattern);
		}

		JavaHeapEvaluator javaHeapEvaluator = new JavaHeapEvaluator(
				loop_invariant_methods);
		processor.registerObserver(javaHeapEvaluator);
		ginPink.setProcessor(processor);
		ginPink.process_data_trace_file(dtraceFilename);
	}

	public Set<String> getLoopInvariantMethodNames() {
		HashSet<String> methodNames = new HashSet<String>();
		for (LoopInvariantMethod loopInvariantMethod : this.loop_invariant_methods) {
			String methodName = loopInvariantMethod.getMethodName();
			methodNames.add(methodName);
		}
		return methodNames;
	}

	public List<Invariant> getInvariants(String loopInvariantMethodName) {
		for (LoopInvariantMethod loopInvariantMethod : this.loop_invariant_methods) {
			if (loopInvariantMethod.getMethodName().equals(
					loopInvariantMethodName)) {
				return loopInvariantMethod.getInvariants();
			}
		}
		return null;
	}

	public void writeInvariants(String invFilename) throws IOException,
			RecognitionException {
		InvariantFileWriter writer = new InvariantFileWriter(
				this.loop_invariant_methods);
		writer.writeFile(invFilename);
	}

	public void writeDiscardedInvariants(String discardedInvFilename)
			throws IOException, RecognitionException {
		InvariantFileWriter writer = new InvariantFileWriter(
				this.loop_invariant_methods);
		writer.writeFile(discardedInvFilename, true);
	}

	public void readInvariants(String invFilename) throws IOException,
			RecognitionException {
		InvariantFileReader reader = new InvariantFileReader(className,
				methodName);
		List<LoopInvariantMethod> new_loop_invariant_methods = reader
				.readFile(invFilename);
		this.loop_invariant_methods.clear();
		this.loop_invariant_methods.addAll(new_loop_invariant_methods);
	}

	public String getPptSelectPattern() {
		return pptSelectPattern;
	}

	public void setPptSelectPattern(String regex) {
		this.pptSelectPattern = regex;
	}

	public String getPptOmitPattern() {
		return pptOmitPattern;
	}

	public void setPptOmitPattern(String regex) {
		this.pptOmitPattern = regex;
	}



}
