package org.ginpink.daikon;

import org.ginpink.GinPinkRuntime;

import daikon.ValueTuple;
import daikon.VarInfo;
import daikon.VarInfo.VarKind;
import daikon.util.Intern;

public final class SnapshotBuilder {

	public Snapshot buildSnapshot(String className, String methodName, final VarInfo[] var_infos, ValueTuple vt) {
		Snapshot snapshot = new Snapshot(className, methodName);
		for (int i = 0; i < var_infos.length; i++) {
			VarInfo varInfo = var_infos[i];
			Object value = vt.getValue(varInfo);

			if (varInfo.isDerived()) {
				// we are not interested in size(..) and array[index] expressions
				continue;
			}

			boolean isMissing = vt.isMissing(varInfo);
			if (isMissing) {
				// nonSensical value
				continue;
			}

			if (varInfo.type.isArray() && varInfo.arr_dims == 1) {
				process_multi_value_sample(snapshot, varInfo, value);
			} else if (!varInfo.type.isArray() || varInfo.arr_dims == 0) {
				process_single_value_sample(snapshot, varInfo, value);
			} else {
				throw new UnsupportedOperationException(
						"Unsupported sample format!");
			}
		}
		
		return snapshot;
	}

	private void process_single_value_sample(Snapshot snapshot,
			VarInfo varInfo, Object value) {
		String varName = varInfo.name();
		String type_base = varInfo.type.base();

		/*
		 * Get SnapShotValue
		 */
		SnapshotValue snapshotValue = null;
		if (varInfo.type.isPrimitive()) {

			if (type_base.equals(int.class.getName())) {
				int intValue = new Integer(String.valueOf(value)).intValue();
				snapshotValue = snapshot.getIntValue(intValue);
			} else if (type_base.equals(long.class.getName())) {
				long longValue = new Long(String.valueOf(value)).longValue();
				snapshotValue = snapshot.getLongValue(longValue);
				
			} else if (type_base.equals(boolean.class.getCanonicalName())) {
				String value_str = value.toString();
				boolean booleanValue = new Boolean(
						Boolean.parseBoolean(value_str)).booleanValue();
				snapshotValue = snapshot.getBooleanValue(booleanValue);

			}

		} else if (varInfo.isArray()) {
			assert (varInfo.arr_dims == 0);

			String class_name = varInfo.type.toString();
			Long longValue = (Long) value;
			if (longValue == Intern.intern(0L)) {
				// assuming null 
				snapshotValue = snapshot.getNullReference();
			} else {
				int hashCodeValue = new Integer(String.valueOf(value));
				SnapshotNonNullReference snapshotReference = snapshot
						.getNonNullReference(hashCodeValue);
				snapshotReference.setCanonicalClassName(class_name);
				snapshotReference
						.setCanonicalBaseClassName(varInfo.type.base());
				snapshotReference.isArray(true);
				snapshotValue = snapshotReference;
			}

		} else {

			if (type_base.equals(Class.class.getName())) {

				String prefixVariable = varName.replace(".getClass()", "");
				snapshotValue = snapshot.evaluateDotExpression(prefixVariable);
				String className = (String) value;
				if (snapshotValue instanceof SnapshotNonNullReference) {
					SnapshotNonNullReference non_null_ref = (SnapshotNonNullReference) snapshotValue;
					non_null_ref.setCanonicalClassName(className);
				}
				return;
			}

			String class_name = type_base;
			Long longValue = (Long) value;
			if (longValue == Intern.intern(0L)) {
				// assuming null 
				snapshotValue = snapshot.getNullReference();
			} else {
				int hashCodeValue = new Integer(String.valueOf(value));
				SnapshotNonNullReference snapshotReference = snapshot
						.getNonNullReference(hashCodeValue);
				snapshotReference.setCanonicalClassName(class_name);
				snapshotValue = snapshotReference;
			}
		}

		/*
		 * Store SnapShotValue
		 */
		if (varInfo.isParam()) {
			snapshot.setVariable(varName, snapshotValue);

		} else {

			if (varInfo.is_field()) {
				boolean isStatic = isStatic(varName);
				if (isStatic) {
					// CLASS field
					snapshot.setClassField(varName, snapshotValue);
				} else {
					// INSTANCE field
					String[] navigation_path = varName.split("\\.");
					String rootVariableName = navigation_path[0];
					String fieldName = navigation_path[navigation_path.length - 1];
					SnapshotNonNullReference curr = snapshot
							.getReference(rootVariableName);

					for (int j = 1; j < navigation_path.length - 1; j++) {
						String currFieldName = navigation_path[j];
						SnapshotValue instanceFieldValue = snapshot
								.getInstanceField(curr, currFieldName);
						curr = (SnapshotNonNullReference) instanceFieldValue;
					}
					snapshot.setInstanceField(fieldName, curr, snapshotValue);
				}
			}

		}
	}

	private void process_multi_value_sample(Snapshot snapshot, VarInfo varInfo,
			Object value) {

		String varName = varInfo.name();
		String type_base = varInfo.type.base();
		String relative_name = varInfo.relative_name;
		if (relative_name != null) {

			//myArrayList.elementData[..].getClass()
			//myArrayList.elementData[..].intValue
			//zipArray[..].zap
			//zipArray[..].zap.zip

			String[] array_navigation_pair = varName.split("\\[\\.\\.\\]\\.");
			String arrayName = array_navigation_pair[0];
			String navigation_str = array_navigation_pair[1];
			SnapshotValue array_value = snapshot
					.evaluateDotExpression(arrayName);
			SnapshotNonNullReference array_ref = (SnapshotNonNullReference) array_value;

			// get SnapshotValue[] from [..].relative_name
			int array_size = array_ref.getArraySize();
			SnapshotValue[] snapshotRefs = new SnapshotValue[array_size];
			String[] navigation_path = navigation_str.split("\\.");
			for (int i = 0; i < array_size; i++) {
				snapshotRefs[i] = array_ref.getItem(i);
				for (int j = 0; j < navigation_path.length - 1; j++) {
					String fieldName = navigation_path[j];
					if (snapshotRefs[i] != null) {
						if (snapshotRefs[i] instanceof SnapshotNonNullReference) {
							SnapshotNonNullReference snapshot_ref = (SnapshotNonNullReference) snapshotRefs[i];
							SnapshotValue new_snapshot_ref = snapshot_ref
									.getFieldValue(fieldName);
							snapshotRefs[i] = new_snapshot_ref;
						} else {
							//nonsensical value
							snapshotRefs[i] = null;
						}
					} else {
						// null reachead
						break;
					}
				}
			}

			String update_field = navigation_path[navigation_path.length - 1];

			assert array_ref.isArray();
			if (value instanceof long[]) {

				long[] longValues = (long[]) value;

				if (varInfo.type.baseIsPrimitive()) {
					if (varInfo.type.base().equals(int.class.getName())) {
						for (int i = 0; i < snapshotRefs.length; i++) {
							if (snapshotRefs[i] != null
									&& snapshotRefs[i] instanceof SnapshotNonNullReference) {
								SnapshotNonNullReference snapshot_ref = (SnapshotNonNullReference) snapshotRefs[i];
								SnapshotIntValue snapshotValue = snapshot
										.getIntValue((int) longValues[i]);
								snapshot.setInstanceField(update_field,
										snapshot_ref, snapshotValue);
							}
						}
					} else {
						throw new UnsupportedOperationException(
								"No primitive support for this type "
										+ varInfo.type.base());
					}
				} else {

					for (int i = 0; i < snapshotRefs.length; i++) {
						if (snapshotRefs[i] != null
								&& snapshotRefs[i] instanceof SnapshotNonNullReference) {
							SnapshotNonNullReference snapshot_ref = (SnapshotNonNullReference) snapshotRefs[i];
							String class_name = varInfo.type.base();
							SnapshotValue snapshotValue;
							if (longValues[i] == Intern.intern(0L)) {
								// assuming null 
								snapshotValue = snapshot.getNullReference();
							} else {
								int hashCodeValue = new Integer(
										String.valueOf(longValues[i]));
								SnapshotNonNullReference snapshotReference = snapshot
										.getNonNullReference(hashCodeValue);
								snapshotReference
										.setCanonicalClassName(class_name);
								snapshotValue = snapshotReference;
							}
							snapshot.setInstanceField(update_field,
									snapshot_ref, snapshotValue);

						}
					}

				}

			} else if (value instanceof Long[]) {

				Long[] longValues = (Long[]) value;

				if (varInfo.type.baseIsPrimitive()) {
					if (varInfo.type.base().equals(boolean.class.getName())) {
						for (int i = 0; i < snapshotRefs.length; i++) {
							if (snapshotRefs[i] != null
									&& snapshotRefs[i] instanceof SnapshotNonNullReference) {
								SnapshotNonNullReference snapshot_ref = (SnapshotNonNullReference) snapshotRefs[i];
								boolean booleanValue;
								if (longValues[i].intValue() == Intern.intern(0L)) {
									booleanValue = false;
								} else {
									booleanValue = true;
								}
								SnapshotBooleanValue snapshotValue = snapshot
										.getBooleanValue(booleanValue);
								snapshot.setInstanceField(update_field,
										snapshot_ref, snapshotValue);
							}
						}
					} else if (varInfo.type.base().equals(int.class.getName())) {
						for (int i = 0; i < snapshotRefs.length; i++) {
							if (snapshotRefs[i] != null
									&& snapshotRefs[i] instanceof SnapshotNonNullReference) {
								SnapshotNonNullReference snapshot_ref = (SnapshotNonNullReference) snapshotRefs[i];
								SnapshotIntValue snapshotValue = snapshot
										.getIntValue(longValues[i].intValue());
								snapshot.setInstanceField(update_field,
										snapshot_ref, snapshotValue);
							}
						}
					} else {
						throw new UnsupportedOperationException(
								"No primitive support for this type "
										+ varInfo.type.base());
					}
				} else {

					for (int i = 0; i < snapshotRefs.length; i++) {
						if (snapshotRefs[i] != null
								&& snapshotRefs[i] instanceof SnapshotNonNullReference) {
							SnapshotNonNullReference snapshot_ref = (SnapshotNonNullReference) snapshotRefs[i];
							String class_name = varInfo.type.base();
							SnapshotValue snapshotValue;
							if (longValues[i] == Intern.intern(0L)) {
								// assuming null 
								snapshotValue = snapshot.getNullReference();
							} else {
								int hashCodeValue = new Integer(
										String.valueOf(longValues[i]));
								SnapshotNonNullReference snapshotReference = snapshot
										.getNonNullReference(hashCodeValue);
								snapshotReference
										.setCanonicalClassName(class_name);
								snapshotValue = snapshotReference;
							}
							snapshot.setInstanceField(update_field,
									snapshot_ref, snapshotValue);

						}
					}

				}

			} else if (value instanceof String[]) {

				String[] stringValues = (String[]) value;
				for (int i = 0; i < snapshotRefs.length; i++) {
					if (snapshotRefs[i] != null) {
						if (snapshotRefs[i] instanceof SnapshotNonNullReference) {
							SnapshotNonNullReference ref = (SnapshotNonNullReference) snapshotRefs[i];
							ref.setCanonicalClassName(stringValues[i]);
						} else if (snapshotRefs[i] instanceof SnapshotNullReference) {
							assert stringValues[i].equals("null");
						}

					}
				}
			} else {

			}

			return;
		}

		if (!varName.endsWith("[..]")) {
			throw new UnsupportedOperationException();
		}

		String varNamePrefix = varName.replace("[..]", "");

		SnapshotNonNullReference array_ref = (SnapshotNonNullReference) snapshot
				.evaluateDotExpression(varNamePrefix);

		String base_class_name = type_base;

		long[] longValueArray = (long[]) value;

		array_ref.setArraySize(longValueArray.length);

		for (int j = 0; j < longValueArray.length; j++) {
			Long longValue = (Long) longValueArray[j];

			SnapshotValue arrayItem;

			if (base_class_name.equals(long.class.getCanonicalName())) {
				arrayItem = snapshot.getLongValue(longValue.longValue());
			} else if (base_class_name.equals(int.class.getCanonicalName())) {
				int intValue = longValue.intValue();
				arrayItem = snapshot.getIntValue(intValue);
			} else {

				if (longValue == Intern.intern(0L)) {
					// assuming null 
					arrayItem = snapshot.getNullReference();
				} else {
					int hashCodeValue = new Integer(String.valueOf(longValue));
					SnapshotNonNullReference snapshotReference = snapshot
							.getNonNullReference(hashCodeValue);
					snapshotReference.setCanonicalClassName(base_class_name);
					arrayItem = snapshotReference;
				}
			}
			array_ref.setItemAt(j, arrayItem);
		}

	}

	private boolean isStatic(String varName) {
		ClassLoader classLoader = GinPinkRuntime.getClassLoader();
		String[] navigation = varName.split("\\.");
		for (int i = navigation.length - 1; i >= 0; i--) {
			StringBuffer buff = new StringBuffer();
			for (int j = 0; j < i; j++) {
				if (buff.length() > 0)
					buff.append(".");
				buff.append(navigation[j]);
			}
			String className = buff.toString();
			try {
				classLoader.loadClass(className);
				return true;
			} catch (ClassNotFoundException e) {
				continue;
			}
		}
		return false;
	}

}
