package org.ginpink.expr;

import java.util.LinkedList;
import java.util.List;

public final class JExprPrinter implements JExprVisitor<String> {

	@Override
	public String visit(JBinaryOperation e) {
		String left_str = e.getLeft().accept(this);
		String right_str = e.getRight().accept(this);
		return String.format("%s %s %s", left_str, e.getOperator(), right_str);
	}

	@Override
	public String visit(JIntegerConstant e) {
		return String.valueOf(e.intValue());
	}

	@Override
	public String visit(JVar e) {
		return e.getVarName();
	}

	@Override
	public String visit(JUnaryOperation e) {
		String operator = e.getOperator().toString();
		String expr = e.getExpr().accept(this);
		return String.format("%s %s", operator, expr);
	}

	@Override
	public String visit(JBooleanConstant e) {
		return String.valueOf(e.booleanValue());
	}

	@Override
	public String visit(JInstanceField e) {
		String receiver = e.getInstanceExpr().accept(this);
		String fieldName = e.getFieldName();
		return String.format("%s.%s", receiver, fieldName);
	}

	@Override
	public String visit(JConditionalExpr e) {
		String pred_str = e.getPredExpr().accept(this);
		String true_str = e.getTrueExpr().accept(this);
		String false_str = e.getFalseExpr().accept(this);
		return String.format("%s ? %s : %s", pred_str, true_str, false_str);
	}

	@Override
	public String visit(JNullExpr jNullExpr) {
		return "null";
	}

	@Override
	public String visit(JArrayAccess e) {
		String array_str = e.getArrayExpr().accept(this);
		String index_str = e.getIndexExpr().accept(this);
		return String.format("%s[%s]", array_str, index_str);
	}

	@Override
	public String visit(JIdentifier e) {
		return e.getIdentifier();
	}

	@Override
	public String visit(JDot e) {
		String leftStr = e.getLeft().accept(this);
		String rightStr = e.getRight().accept(this);
		return leftStr + "." + rightStr;
	}

	@Override
	public String visit(JClassName e) {
		return e.getClassName();
	}

	@Override
	public String visit(JStaticMethodCall e) {
		List<String> arg_strs = new LinkedList<String>();
		for (JExpr arg : e.getArguments()) {
			String arg_str = arg.accept(this);
			arg_strs.add(arg_str);
		}
		String className = e.getClassName();
		String methodName = e.getMethodName();

		StringBuffer buff = new StringBuffer();
		for (String arg_str : arg_strs) {
			if (buff.length() > 0) {
				buff.append(",");
			}
			buff.append(arg_str);
		}

		return String.format("%s.%s(%s)", className, methodName,
				buff.toString());
	}

	@Override
	public String visit(JInstanceMethodCall e) {
		String receiver = e.getReceiver().accept(this);
		List<String> arg_strs = new LinkedList<String>();
		for (JExpr arg : e.getArguments()) {
			String arg_str = arg.accept(this);
			arg_strs.add(arg_str);
		}
		String methodName = e.getMethodName();

		StringBuffer buff = new StringBuffer();
		for (String arg_str : arg_strs) {
			if (buff.length() > 0) {
				buff.append(",");
			}
			buff.append(arg_str);
		}

		return String
				.format("%s.%s(%s)", receiver, methodName, buff.toString());
	}

	@Override
	public String visit(JMethodCall e) {
		String expr = e.getExpression().accept(this);
		List<String> arg_strs = new LinkedList<String>();
		for (JExpr arg : e.getArguments()) {
			String arg_str = arg.accept(this);
			arg_strs.add(arg_str);
		}

		StringBuffer buff = new StringBuffer();
		for (String arg_str : arg_strs) {
			if (buff.length() > 0) {
				buff.append(",");
			}
			buff.append(arg_str);
		}

		return String.format("%s(%s)", expr, buff.toString());
	}

	@Override
	public String visit(JStringLiteral e) {
		String stringLiteral = e.getStringLiteral();
		return String.format("\"%s\"", stringLiteral);
	}

	@Override
	public String visit(JLongConstant e) {
		return String.valueOf(e.longValue());
	}

}
