package org.ginpink.parser;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeAdaptor;
import org.antlr.runtime.tree.TreeVisitor;
import org.ginpink.Invariant;
import org.ginpink.expr.JExpr;
import org.ginpink.parser.JavaParser.expression_return;

public final class MutationsFileReader {

	private static final String EXEC_JML = "<--exec|JML-->";
	private static final String RIGHT_ARROW = "----->";
	private static final String STARTING_AT = "Starting at:";
	private static final String PROCESSING_WITH = "Processing with:";
	private static final String METHOD = "METHOD:";
	private static final String NEW_METHOD_MARKER = "*********************************************************";

	public MutationsFileReader() {
		this(null, null);
	}

	private final String className;
	private final String methodName;

	public MutationsFileReader(String className, String methodName) {
		this.className = className;
		this.methodName = methodName;
	}

	private HashMap<String, List<Invariant>> parsed_invariants = new HashMap<String, List<Invariant>>();
	private HashMap<String, Integer> ignored_invariants = new HashMap<String, Integer>();

	public void readFile(String filename) throws IOException,
			RecognitionException {
		BufferedReader in = new BufferedReader(new FileReader(filename));

		String class_name_dot_method_name = null;
		String current_class_name = null;
		String current_method_name = null;

		List<Invariant> current_invariants = new LinkedList<Invariant>();
		HashSet<String> current_input_strings = new HashSet<String>();
		int ignored_expressions_count = 0;

		while (in.ready()) {
			String str = in.readLine();
			str = str.trim();
			if (str.equals(NEW_METHOD_MARKER)) {
				// end of current method
				if (class_name_dot_method_name != null) {
					parsed_invariants.put(class_name_dot_method_name,
							current_invariants);
					ignored_invariants.put(class_name_dot_method_name,
							ignored_expressions_count);
				}

				// new method (close previous/ open new)
				class_name_dot_method_name = null;
				current_invariants = new LinkedList<Invariant>();
				current_input_strings = new HashSet<String>();
				ignored_expressions_count = 0;
			} else if (str.startsWith(METHOD)) {
				// get new method name
				class_name_dot_method_name = str.replace(METHOD, "").trim();
				int last_dot = class_name_dot_method_name.lastIndexOf(".");
				current_class_name = class_name_dot_method_name.substring(0,
						last_dot);
				current_method_name = class_name_dot_method_name.substring(
						last_dot + 1, class_name_dot_method_name.length());

			} else if (str.startsWith(PROCESSING_WITH)) {
				//discard
			} else if (str.startsWith(STARTING_AT)) {
				//discard
			} else if (str.startsWith(RIGHT_ARROW)) {
				//discard
			} else if (str.equals("")) {
				//discard
			} else if (str.startsWith("Iteration #")) {
				//discard
			} else {
				// mutation
				if (className != null && methodName != null
						&& !className.equals(current_class_name)
						&& !methodName.equals(current_method_name)) {
					//skip mutant 
					continue;
				}

				if (str.contains("\\result")) {
					//ignore mutant 
					continue;
				}

				if (!str.contains(EXEC_JML)) {
					//ignore mutant, must be a syntactic invariant
					continue;

				}
				// contains jml_format and java_format
				String java_string = str.split(EXEC_JML.replace("|", "\\|"))[0];
				String jml_string = str.split(EXEC_JML.replace("|", "\\|"))[1];

				if (java_string.contains("++") || java_string.contains("--")) {
					// discard non-pure mutant
					continue;
				}
				
				JExpr java_format = parseExpr(java_string);
				current_input_strings.add(java_string);
				Invariant invariant = new Invariant(java_format, java_string, jml_string);
				current_invariants.add(invariant);

			}
		}

		// end of current method
		if (class_name_dot_method_name != null) {
			parsed_invariants.put(class_name_dot_method_name,
					current_invariants);
			ignored_invariants.put(class_name_dot_method_name,
					ignored_expressions_count);
		}

		in.close();
	}

	private JExpr parseExpr(String javaString) throws RecognitionException {
		ANTLRStringStream charStream = new ANTLRStringStream(javaString);
		JavaLexer javalexer = new JavaLexer(charStream);
		CommonTokenStream tokens = new CommonTokenStream();
		tokens.setTokenSource(javalexer);
		JavaParser javaparser = new JavaParser(tokens);

		expression_return expr_return = javaparser.expression();

		CommonTree common_tree = expr_return.getTree();

		TreeVisitor v = new TreeVisitor(new CommonTreeAdaptor());
		ExprVisitorAction actions = new ExprVisitorAction();
		v.visit(common_tree, actions);
		return actions.getExpression();
	}

	public Set<String> methodSet() {
		return this.parsed_invariants.keySet();
	}

	public List<Invariant> getExpressions(String methodName) {
		return this.parsed_invariants.get(methodName);
	}

	public int getIgnoredExpressionsCount(String methodName) {
		return this.ignored_invariants.get(methodName);
	}
}
