package org.ginpink.typecheck;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Map;

import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.classfile.ClassParser;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.LocalVariable;
import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.generic.BasicType;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.Type;
import org.ginpink.GinPinkRuntime;
import org.ginpink.expr.JArrayAccess;
import org.ginpink.expr.JBinaryOperation;
import org.ginpink.expr.JBooleanConstant;
import org.ginpink.expr.JClassName;
import org.ginpink.expr.JConditionalExpr;
import org.ginpink.expr.JDot;
import org.ginpink.expr.JExpr;
import org.ginpink.expr.JExprPrinter;
import org.ginpink.expr.JExprVisitor;
import org.ginpink.expr.JIdentifier;
import org.ginpink.expr.JInstanceField;
import org.ginpink.expr.JInstanceMethodCall;
import org.ginpink.expr.JIntegerConstant;
import org.ginpink.expr.JLongConstant;
import org.ginpink.expr.JMethodCall;
import org.ginpink.expr.JNullExpr;
import org.ginpink.expr.JStaticMethodCall;
import org.ginpink.expr.JStringLiteral;
import org.ginpink.expr.JType;
import org.ginpink.expr.JUnaryOperation;
import org.ginpink.expr.JVar;
import org.ginpink.util.Reflection;

public final class ExprTypeChecker implements JExprVisitor<JExpr> {

	private static final String LENGTH = "length";
	private final Method reflect_method;
	private final org.apache.bcel.classfile.Method bcel_method;
	private final Map<String, Class<?>> local_variable_signatures = new HashMap<String, Class<?>>();

	public ExprTypeChecker(Method reflect_method) {
		this.reflect_method = reflect_method;

		Class<?> declaringClass = reflect_method.getDeclaringClass();
		String declaringClassName = declaringClass.getCanonicalName();
		String resourceName = declaringClassName.replace(".", "/") + ".class";
		InputStream is = declaringClass.getClassLoader().getResourceAsStream(
				resourceName);
		ClassParser class_parser = new ClassParser(is, null);
		try {
			JavaClass java_class = class_parser.parse();

			bcel_method = java_class.getMethod(reflect_method);

			if (bcel_method == null) {
				throw new RuntimeException(
						"BCEL method could not be found for "
								+ reflect_method.getName());
			}

			for (LocalVariable localVariable : bcel_method
					.getLocalVariableTable().getLocalVariableTable()) {

				Type bcel_type = Type.getType(localVariable.getSignature());
				local_variable_signatures.put(localVariable.getName(),
						getClassOf(bcel_type));
			}

		} catch (ClassFormatException e) {
			throw new RuntimeException("Error while parsing resource "
					+ resourceName, e);
		} catch (IOException e) {
			throw new RuntimeException("Error while parsing resource "
					+ resourceName, e);
		}

	}

	private Class<?> getClassOf(Type bcel_type) {
		if (bcel_type.equals(BasicType.INT)) {
			return int.class;
		}
		if (bcel_type.equals(BasicType.LONG)) {
			return long.class;
		}

		if (bcel_type.equals(BasicType.BOOLEAN)) {
			return boolean.class;
		}
		if (bcel_type instanceof ObjectType) {
			ObjectType objectType = (ObjectType) bcel_type;
			String className = objectType.getClassName();
			try {
				Class<?> clazz = GinPinkRuntime.getClassLoader().loadClass(
						className);
				return clazz;
			} catch (ClassNotFoundException e) {
				throw new RuntimeException(e);
			}
		}
		if (bcel_type instanceof ArrayType) {
			ArrayType arrayType = (ArrayType) bcel_type;
			int dimensions = arrayType.getDimensions();
			Class<?> basic_type = getClassOf(arrayType.getBasicType());
			Class<?> array_class = Array.newInstance(basic_type, dimensions)
					.getClass();
			return array_class;
		}

		throw new RuntimeException("Could not find class for signature "
				+ bcel_type);
	}

	@Override
	public JExpr visit(JBinaryOperation e) {
		JExpr left = e.getLeft().accept(this);
		JExpr right = e.getRight().accept(this);

		if (left.getType() == null || right.getType() == null) {
			return e;
		}

		switch (e.getOperator()) {
		case ADD:
		case SUB:
		case MULT:
		case DIV:
		case REM: {
			if (!left.getType().isInt()) {
				return e;
			}
			if (!right.getType().isInt()) {
				return e;
			}

			JBinaryOperation type_checked_op = new JBinaryOperation(left,
					e.getOperator(), right);
			type_checked_op.setType(new JType(int.class));
			return type_checked_op;
		}
		case LESS_THAN:
		case GREATER_THAN:
		case GREATER_THAN_EQUAL:
		case LESS_THAN_EQUAL: {

			if (!left.getType().isInt()) {
				return e;
			}
			if (!right.getType().isInt()) {
				return e;
			}

			JBinaryOperation type_checked_op = new JBinaryOperation(left,
					e.getOperator(), right);
			type_checked_op.setType(new JType(boolean.class));
			return type_checked_op;

		}

		case EQUAL:
		case NOT_EQUAL: {

			JType right_type = right.getType();
			JType left_type = left.getType();

			if (left_type.isNullType() && right_type.isPrimitive()) {
				return e;
			}

			if (right_type.isNullType() && left_type.isPrimitive()) {
				return e;
			}

			if (!left_type.isNullType() && !right_type.isNullType()
					&& !left_type.equals(right_type)) {
				return e;
			}

			JBinaryOperation type_checked_op = new JBinaryOperation(left,
					e.getOperator(), right);
			type_checked_op.setType(new JType(boolean.class));
			return type_checked_op;

		}
		case LOGICAL_AND:
		case LOGICAL_OR: {
			if (!left.getType().isBoolean() || !right.getType().isBoolean()) {
				return e;
			}
			JBinaryOperation binary_op = new JBinaryOperation(left,
					e.getOperator(), right);
			binary_op.setType(new JType(boolean.class));
			return binary_op;
		}
		case LOGICAL_NOT:
		case UNARY_MINUS:
		case UNARY_PLUS:
			throw new IllegalArgumentException(
					"The operator is not a binary operator " + e.getOperator());
		default:
			throw new RuntimeException("Unhandled operator " + e.getOperator());
		}
	}

	@Override
	public JExpr visit(JIntegerConstant e) {
		e.setType(new JType(int.class));
		return e;
	}

	@Override
	public JExpr visit(JLongConstant e) {
		e.setType(new JType(long.class));
		return e;
	}

	@Override
	public JExpr visit(JUnaryOperation e) {
		JExpr expr = e.getExpr().accept(this);
		if (expr.getType() == null) {
			return e;
		}

		switch (e.getOperator()) {
		case LOGICAL_NOT: {
			if (!expr.getType().isBoolean()) {
				return e;
			}
			JUnaryOperation unary_op = new JUnaryOperation(e.getOperator(),
					expr);
			unary_op.setType(new JType(boolean.class));
			return unary_op;
		}
		case UNARY_MINUS:
		case UNARY_PLUS: {
			if (!expr.getType().isInt()) {
				return e;
			}
			JUnaryOperation unary_op = new JUnaryOperation(e.getOperator(),
					expr);
			unary_op.setType(new JType(int.class));
			return unary_op;
		}

		case ADD:
		case SUB:
		case MULT:
		case DIV:
		case REM:
		case LESS_THAN:
		case GREATER_THAN:
		case GREATER_THAN_EQUAL:
		case LESS_THAN_EQUAL:
		case EQUAL:
		case NOT_EQUAL:
		case LOGICAL_AND:
		case LOGICAL_OR:
			throw new IllegalArgumentException(
					"The operator is not a unary operator " + e.getOperator());
		default:
			throw new RuntimeException("Unhandled operator " + e.getOperator());
		}

	}

	@Override
	public JExpr visit(JBooleanConstant e) {
		e.setType(new JType(boolean.class));
		return e;
	}

	@Override
	public JExpr visit(JConditionalExpr e) {
		JExpr pred_expr = e.getPredExpr().accept(this);
		JExpr true_expr = e.getTrueExpr().accept(this);
		JExpr false_expr = e.getFalseExpr().accept(this);

		if (!pred_expr.isTypeChecked() || !true_expr.isTypeChecked()
				|| !false_expr.isTypeChecked()) {
			return e;
		}

		if (!pred_expr.getType().isBoolean()) {
			return e;
		}

		if (!true_expr.getType().equals(false_expr.getType())) {
			return e;
		}

		JConditionalExpr condExpr = new JConditionalExpr(pred_expr, true_expr,
				false_expr);
		JType condExpr_type = true_expr.getType();
		condExpr.setType(condExpr_type);

		return condExpr;
	}

	@Override
	public JExpr visit(JNullExpr e) {
		e.setType(JType.NULL_TYPE);
		return e;
	}

	@Override
	public JExpr visit(JIdentifier e) {
		String ident = e.getIdentifier();
		if (local_variable_signatures.containsKey(ident)) {
			JVar var = new JVar(ident);
			JType type = new JType(local_variable_signatures.get(ident));
			var.setType(type);
			return var;
		}

		String canonicalName = reflect_method.getDeclaringClass()
				.getCanonicalName();
		String className;
		if (canonicalName.contains(".")) {
			String[] parts = canonicalName.split("\\.");
			String[] new_parts = java.util.Arrays.copyOfRange(parts, 0,
					parts.length - 1);
			String packageName = join(new_parts);
			className = packageName + "." + ident;
		} else {
			className = ident;
		}
		try {
			Class<?> clazz = GinPinkRuntime.getClassLoader().loadClass(
					className);

			JClassName exprClassName = new JClassName(className);
			exprClassName.setType(new JType(clazz));
			return exprClassName;

		} catch (ClassNotFoundException e1) {
			return e;
		} catch (NoClassDefFoundError e2) {
			return e;
		}

	}

	private static String join(String[] parts) {
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < parts.length; i++) {
			if (buf.length() > 0)
				buf.append(".");
			buf.append(parts[i]);
		}
		return buf.toString();
	}

	@Override
	public JExpr visit(JDot e) {
		JExpr leftExpr = e.getLeft().accept(this);
		JExpr rightExpr = e.getRight().accept(this);

		if (rightExpr instanceof JClassName) {
			//discard JClassName if a dot is found (preference over method's package)
			rightExpr = e.getRight();
		} else if (!(rightExpr instanceof JIdentifier)) {
			return e;
		}
		JIdentifier rightIdentifier = (JIdentifier) rightExpr;
		String ident_str = rightIdentifier.getIdentifier();

		if (leftExpr.isTypeChecked() && !leftExpr.isPrimitive()) {
			return dotToField(e, leftExpr, ident_str);
		}

		if (!leftExpr.isTypeChecked()) {
			JExprPrinter printer = new JExprPrinter();
			String packageName = leftExpr.accept(printer);
			String simpleClassName = ident_str;

			String className = packageName + "." + simpleClassName;
			try {
				Class<?> clazz = GinPinkRuntime.getClassLoader().loadClass(
						className);
				JClassName classNameExpr = new JClassName(className);
				classNameExpr.setType(new JType(clazz));
				return classNameExpr;
			} catch (ClassNotFoundException e1) {
				return e;
			}
		}
		return e;
	}

	private static JExpr dotToField(JDot e, JExpr leftExpr, String ident_str) {
		Class<?> clazz = leftExpr.getType().getJavaClass();

		Class<?> fieldType;
		if (clazz.isArray() && ident_str.equals(LENGTH)) {
			fieldType = int.class;
		} else {
			try {
				Field field = clazz.getDeclaredField(ident_str);
				fieldType = field.getType();

			} catch (SecurityException e1) {
				return e;
			} catch (NoSuchFieldException e1) {
				return e;
			}
		}

		JInstanceField fieldExpr = new JInstanceField(leftExpr, ident_str);
		fieldExpr.setType(new JType(fieldType));
		return fieldExpr;
	}

	@Override
	public JExpr visit(JMethodCall e) {

		JExpr[] type_checked_args = new JExpr[e.getArguments().size()];
		Class<?>[] arg_types = new Class<?>[e.getArguments().size()];

		for (int i = 0; i < e.getArguments().size(); i++) {
			JExpr argExpr = e.getArguments().get(i);
			JExpr type_checked_arg = argExpr.accept(this);
			if (!type_checked_arg.isTypeChecked()) {
				//failed to typecheck
				return e;
			}
			type_checked_args[i] = type_checked_arg;
			arg_types[i] = type_checked_arg.getType().getJavaClass();
		}

		JExpr expr = e.getExpression();
		String methodName;
		if (expr instanceof JDot) {
			JDot dot = (JDot) expr;
			JExpr dot_left = dot.getLeft().accept(this);
			JExpr dot_right = dot.getRight().accept(this);

			if (dot_right instanceof JIdentifier) {
				JIdentifier ident = (JIdentifier) dot_right;
				methodName = ident.getIdentifier();
			} else if (dot_right instanceof JVar) {
				JVar var = (JVar) dot_right;
				String varName = var.getVarName();
				methodName = varName;
			} else {
				//fail to typecheck
				return e;
			}

			if (!dot_left.isTypeChecked()) {
				return e;
			}

			Class<?> clazz = dot_left.getType().getJavaClass();
			try {
				Method method = Reflection.getDeclaredMethod(clazz, methodName,
						arg_types);
				boolean isStatic = Modifier.isStatic(method.getModifiers());

				JExpr methodCall;
				String className = clazz.getCanonicalName();
				if (isStatic) {
					methodCall = new JStaticMethodCall(className, methodName,
							type_checked_args);

				} else {

					JExpr receiver = dot_left;
					methodCall = new JInstanceMethodCall(receiver, methodName,
							type_checked_args);

				}
				methodCall.setType(new JType(method.getReturnType()));
				return methodCall;

			} catch (SecurityException e1) {
				return e;
			} catch (NoSuchMethodException e1) {
				return e;
			}

		} else if (expr instanceof JIdentifier) {
			JIdentifier ident = (JIdentifier) expr;
			methodName = ident.getIdentifier();

			throw new RuntimeException("not yet implemented!");

		} else {
			// fail to typecheck
			return e;
		}
	}

	public static boolean checkArgumentTypes(
			Class<?>[] concrete_parameter_types,
			Class<?>[] formal_parameter_types) {

		if (concrete_parameter_types.length != formal_parameter_types.length) {
			return false;
		}

		for (int i = 0; i < formal_parameter_types.length; i++) {
			Class<?> formal_parameter_type = formal_parameter_types[i];
			Class<?> concrete_parameter_type = concrete_parameter_types[i];

			if (concrete_parameter_type == null) {
				// handle null type as reference type
				if (!Object.class.isAssignableFrom(formal_parameter_type)) {
					return false;
				} else {
					// null value can be casted to this type
				}
			} else if (!formal_parameter_type
					.isAssignableFrom(concrete_parameter_type)) {
				return false;
			}
		}
		return true;
	}

	@Override
	public JExpr visit(JArrayAccess e) {
		JExpr arrayExpr = e.getArrayExpr().accept(this);
		JExpr indexExpr = e.getIndexExpr().accept(this);

		if (!indexExpr.isTypeChecked() || !arrayExpr.isTypeChecked()) {
			return e;
		}
		if (!indexExpr.getType().isInt()) {
			return e;
		}
		if (!arrayExpr.getType().isArray()) {
			return e;
		}

		JArrayAccess array_access = new JArrayAccess(arrayExpr, indexExpr);
		Class<?> componentType = arrayExpr.getType().getComponentType();
		JType array_type = new JType(componentType);
		array_access.setType(array_type);
		return array_access;
	}

	Method getReflectMethod() {
		return reflect_method;
	}

	org.apache.bcel.classfile.Method getBcelMethod() {
		return bcel_method;
	}

	@Override
	public JExpr visit(JInstanceField e) {
		throw new IllegalArgumentException();
	}

	@Override
	public JExpr visit(JVar e) {
		throw new IllegalArgumentException();
	}

	@Override
	public JExpr visit(JClassName e) {
		throw new IllegalArgumentException();
	}

	@Override
	public JExpr visit(JInstanceMethodCall e) {
		throw new IllegalArgumentException();
	}

	@Override
	public JExpr visit(JStaticMethodCall e) {
		throw new IllegalArgumentException();
	}

	@Override
	public JExpr visit(JStringLiteral e) {
		e.setType(new JType(String.class));
		return e;
	}

}
